--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Multi-class management of crowd control
	
	All code here is inside functions. Nothing runs when the .TOC loads, other than the
	high-level variable definitions. Execution beings in LeaderBot.lua:OnLoad()

--]]

local CCL_Debug = 0;
local version = "1.02 BETA";

CrowdControllerFrame = nil;

local CROWDCONTROL_WARLOCK = "Warlock"
local CROWDCONTROL_DRUID = "Druid"
local CROWDCONTROL_MAGE = "Mage"
local CROWDCONTROL_PRIEST = "Priest"
local CROWDCONTROL_ROGUE = "Rogue"
local CROWDCONTROL_HUNTER = "Hunter"

local CCTYPES = {
                ["Warlock"] = CROWDCONTROL_WARLOCK,
                ["Druid"]   = CROWDCONTROL_DRUID,
                ["Hunter"]  = CROWDCONTROL_HUNTER,
                ["Mage"]    = CROWDCONTROL_MAGE,
                ["Priest"]  = CROWDCONTROL_PRIEST,
                ["Rogue"]   = CROWDCONTROL_ROGUE,
            };

--Global variables to be saved

CrowdController_Assignments = {};
for class, assignmentType in next, CCTYPES do
    CrowdController_Assignments[assignmentType] = {};
end 

--[[
CrowdController_Assignments[CROWDCONTROL_WARLOCK]   = {};
CrowdController_Assignments[CROWDCONTROL_DRUID]     = {};
CrowdController_Assignments[CROWDCONTROL_MAGE ]     = {};
CrowdController_Assignments[CROWDCONTROL_PRIEST]    = {};
CrowdController_Assignments[CROWDCONTROL_ROGUE]     = {};
CrowdController_Assignments[CROWDCONTROL_HUNTER]    = {};
--]]


local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local CrowdController_MsgChannel ={};

local columnWidth = 90;
local columnWidthInt = 150;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local variablesLoaded = false;

local druidDropDownName = "CrowdControllerDruidDropDown";
local hunterDropDownName = "CrowdControllerHunterDropDown";
local mageDropDownName = "CrowdControllerMageDropDown";
local priestDropDownName = "CrowdControllerPriestDropDown";
local rogueDropDownName = "CrowdControllerRogueDropDown";
local shamanDropDownName = "CrowdControllerShamanDropDown";
local warlockDropDownName = "CrowdControllerWarlockDropDown";
local warriorDropDownName = "CrowdControllerWarriorDropDown";

local colorCode = "|cff11aa11 ";
local whiteColorCode = "|cffffffff";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};
	
local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function CCLmsg(msg, override)
	if (CCL_Debug == 1 or override ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function CrowdControllerFrame_OnLoad(parentFrame, teamMembers, crowdControllerCount)
	CCLmsg("Function: CrowdControllerFrame_OnLoad()");

	--CrowdControllerFrame = CreateFrame("Frame", "CrowdControllerFrame", parentFrame);  -- wow classic
	CrowdControllerFrame = CreateFrame("Frame", "CrowdControllerFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate");  -- wow TBC

	-- UI Commands
	SLASH_HunterLeader1 = "/ccl";
	SLASH_HunterLeader2 = "/CCL";

	SlashCmdList["CrowdController"] = CrowdController_SlashHandler;

	ChatFrame1:AddMessage(colorCode .. "CrowdController Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's hunterLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	CrowdController_InitMainFrame(teamMembers, crowdControllerCount);

	return (CrowdControllerFrame);
end;

-- *************************************************************************************
function CrowdController_InitMainFrame(teamMembers, crowdControllerCount)
	CCLmsg("Function: CrowdController_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove hunters

	CrowdControllerFrame:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 2);	
	CrowdController_AdjustMainPanelSize(crowdControllerCount);	

	CrowdControllerFrame:SetBackdrop(mainBackdrop);
	CrowdControllerFrame:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	CrowdControllerFrame:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	CrowdControllerFrame:EnableMouse(true);
	CrowdControllerFrame:SetScript("OnUpdate", CrowdControllerFrame_OnUpdate);
	CrowdControllerFrame:SetScript("OnEvent", CrowdControllerFrame_OnEvent);

	CrowdControllerFrame:SetScript("OnShow", function(self)
						CCLmsg("Function: CrowdControllerFrame_OnShow() ");
						CCLmsg("current panel = " .. LeaderBotOptions["currentPanel"] );

						if (LeaderBotOptions["currentPanel"] == "crowdcontroller") then 
							local parentFrame = CrowdControllerFrame:GetParent();
							parentFrame:SetWidth(CrowdControllerFrame:GetWidth() );
						end					
				end);

    local appLogo = LeaderBot_GetTextForSpellID(710);
	local appTitle = CrowdControllerFrame:CreateFontString("CrowdControllerAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", CrowdControllerFrame, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Crowd Controller " .. version);

	local infoText = CrowdControllerFrame:CreateFontString("CrowdControllerInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	--infoText:SetText(colorCode .. " ");

	columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, crowdControllerCount, "crowdcontroller", CrowdControllerFrame, "Crowd Controllers");
	columnFrames[2], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[1], 
							columnWidth, "CrowdControllerColumn2", "Warlock", 
							"LeaderBotcrowdcontrollerMember", warlockDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_WARLOCK, 
							"CrowdControllerAnnounce2", CrowdController_AnnounceToChat, "Leader Bot: Warlock Crowd Control Assignments",
							CrowdController_WarlockDropDownOnClick,
							LeaderBotSingleSelect);	
	columnFrames[3], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[2], 
							columnWidth, "CrowdControllerColumn3", "Druid", 
							"LeaderBotcrowdcontrollerMember", druidDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_DRUID, 
							"CrowdControllerAnnounce3", CrowdController_AnnounceToChat, "Leader Bot: Druid Crowd Control Assignments",
							CrowdController_DruidDropDownOnClick,
							LeaderBotSingleSelect);	
    columnFrames[4], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[3], 
							columnWidth, "CrowdControllerColumn4", "Hunter", 
							"LeaderBotcrowdcontrollerMember", hunterDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_HUNTER, 
							"CrowdControllerAnnounce4", CrowdController_AnnounceToChat, "Leader Bot: Hunter Crowd Control Assignments",
							CrowdController_HunterDropDownOnClick,
							LeaderBotSingleSelect);	  
    columnFrames[5], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[4], 
							columnWidth, "CrowdControllerColumn5", "Mage", 
							"LeaderBotcrowdcontrollerMember", mageDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_MAGE, 
							"CrowdControllerAnnounce5", CrowdController_AnnounceToChat, "Leader Bot: Mage Crowd Control Assignments",
							CrowdController_MageDropDownOnClick,
							LeaderBotSingleSelect);	    
    columnFrames[6], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[5], 
							columnWidth, "CrowdControllerColumn6", "Priest", 
							"LeaderBotcrowdcontrollerMember", priestDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_PRIEST, 
							"CrowdControllerAnnounce6", CrowdController_AnnounceToChat, "Leader Bot: Priest Crowd Control Assignments",
							CrowdController_PriestDropDownOnClick,
							LeaderBotSingleSelect);	  
    columnFrames[7], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[6], 
							columnWidth, "CrowdControllerColumn7", "Rogue", 
							"LeaderBotcrowdcontrollerMember", rogueDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_ROGUE, 
							"CrowdControllerAnnounce7", CrowdController_AnnounceToChat, "Leader Bot: Rogue Crowd Control Assignments",
							CrowdController_RogueDropDownOnClick,
							LeaderBotSingleSelect);	                                                                                                      

	if (crowdControllerCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 1, columnFrames[1], "LeaderBotcrowdcontrollerMember");
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 2, columnFrames[2], warlockDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 3, columnFrames[3], druidDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 4, columnFrames[4], hunterDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 5, columnFrames[5], mageDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 6, columnFrames[6], priestDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 7, columnFrames[7], rogueDropDownName);

		LeaderBot_SetupScrollBar("crowdcontroller");
	end 


	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","CrowdControllerChatChannelDropDown", CrowdControllerFrame, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );
	dd:SetPoint("BottomLeft", CrowdControllerFrame, "BottomLeft", -15, 0);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelList, CrowdController_ChannelDropDownOnClick ) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();
	UIDropDownMenu_SetText(dd, "<Select channel>");
	
	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end
	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
	
	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end


end

-- ********************************************************************************
function CrowdController_WarlockDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_WarlockDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Warlock")
end 

-- ********************************************************************************
function CrowdController_DruidDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_DruidDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Druid")
end 

-- ********************************************************************************
function CrowdController_HunterDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_HunterDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Hunter")
end 

-- ********************************************************************************
function CrowdController_MageDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_MageDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Mage")
end 

-- ********************************************************************************
function CrowdController_PriestDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_PriestDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Priest")
end 

-- ********************************************************************************
function CrowdController_RogueDropDownOnClick(dropDownButton, arg1, arg2, checked)
    CCLmsg("Function: CrowdController_RogueDropDownOnClick")
    CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, "Rogue")
end 

-- ********************************************************************************
function CrowdController_GenericDropDownOnClick(dropDownButton, arg1, arg2, checked, class)
    CCLmsg("Function: CrowdController_GenericDropDownOnClick(" .. class ..")" );

    local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock. I prefer not to rename arg1, arg2 so I
    --in the function definition so I have documentation on *where* those arguments come from.
	local playerName = arg1;
    local targetIcon = arg2;

    CrowdController_Assignments[ CCTYPES[class] ][playerName] = targetIcon;

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.playerName = playerName;
	dropDownFrame.target = dropDownButton.value;

	dropDownFrame.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "crowdcontroller", CCTYPES[class], playerName, targetIcon);
end 

-- ******************************************************************************
--adjust the height as players join or leave
function CrowdController_AdjustMainPanelSize(numberOfRows)
	CCLmsg("Function: CrowdController_AdjustMainPanelSize(count = " .. numberOfRows .. ")" );

	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidthInt
                        + columnWidth * 6
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
                        + LeaderBotplayerHeight -- extra space for the assignmentType pulldown
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	   
	CrowdControllerFrame:SetSize(math.max(requiredWidth, LeaderBotMinWidth), requiredHeight);
end


-- ************************************************************
function CrowdController_DataMigration()
    CCLmsg("Function: CrowdController_DataMigration() ");
    
    if (LeaderBotOptions["CrowdControllerDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		CrowdController_Assignments = {};

        for class, assignmentType in next, CCTYPES do
            CrowdController_Assignments[assignmentType] = {};
        end 

		LeaderBotOptions["CrowdControllerDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for class, assignmentType in next, CCTYPES do
		if (CrowdController_Assignments[assignmentType] == nil) then
			CrowdController_Assignments[assignmentType] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, CrowdController_Assignments[assignmentType] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then  	
					CrowdController_Assignments[assignmentType][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************************
function CrowdController_RosterChanged(teamMembers, crowdControllerCount)
    CCLmsg("Function: CrowdController_RosterChanged() ");

	CrowdController_AdjustMainPanelSize(crowdControllerCount);

    columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, crowdControllerCount, "crowdcontroller", CrowdControllerFrame, "Crowd Controllers");
	columnFrames[2], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[1], 
							columnWidth, "CrowdControllerColumn2", "Warlock", 
							"LeaderBotcrowdcontrollerMember", warlockDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_WARLOCK, 
							"CrowdControllerAnnounce2", CrowdController_AnnounceToChat, "Leader Bot: Warlock Crowd Control Assignments",
							CrowdController_WarlockDropDownOnClick,
							LeaderBotSingleSelect);	
	columnFrames[3], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[2], 
							columnWidth, "CrowdControllerColumn3", "Druid", 
							"LeaderBotcrowdcontrollerMember", druidDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_DRUID, 
							"CrowdControllerAnnounce3", CrowdController_AnnounceToChat, "Leader Bot: Druid Crowd Control Assignments",
							CrowdController_DruidDropDownOnClick,
							LeaderBotSingleSelect);	
    columnFrames[4], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[3], 
							columnWidth, "CrowdControllerColumn4", "Hunter", 
							"LeaderBotcrowdcontrollerMember", hunterDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_HUNTER, 
							"CrowdControllerAnnounce4", CrowdController_AnnounceToChat, "Leader Bot: Hunter Crowd Control Assignments",
							CrowdController_HunterDropDownOnClick,
							LeaderBotSingleSelect);	  
    columnFrames[5], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[4], 
							columnWidth, "CrowdControllerColumn5", "Mage", 
							"LeaderBotcrowdcontrollerMember", mageDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_MAGE, 
							"CrowdControllerAnnounce5", CrowdController_AnnounceToChat, "Leader Bot: Mage Crowd Control Assignments",
							CrowdController_MageDropDownOnClick,
							LeaderBotSingleSelect);	    
    columnFrames[6], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[5], 
							columnWidth, "CrowdControllerColumn6", "Priest", 
							"LeaderBotcrowdcontrollerMember", priestDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_PRIEST, 
							"CrowdControllerAnnounce6", CrowdController_AnnounceToChat, "Leader Bot: Priest Crowd Control Assignments",
							CrowdController_PriestDropDownOnClick,
							LeaderBotSingleSelect);	  
    columnFrames[7], CrowdController_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, crowdControllerCount, true, 
							CrowdControllerFrame, columnFrames[6], 
							columnWidth, "CrowdControllerColumn7", "Rogue", 
							"LeaderBotcrowdcontrollerMember", rogueDropDownName, 
							CrowdController_Assignments, CROWDCONTROL_ROGUE, 
							"CrowdControllerAnnounce7", CrowdController_AnnounceToChat, "Leader Bot: Rogue Crowd Control Assignments",
							CrowdController_RogueDropDownOnClick,
							LeaderBotSingleSelect);	                                                                                                      

	if (crowdControllerCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 1, columnFrames[1], "LeaderBotcrowdcontrollerMember");
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 2, columnFrames[2], warlockDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 3, columnFrames[3], druidDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 4, columnFrames[4], hunterDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 5, columnFrames[5], mageDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 6, columnFrames[6], priestDropDownName);
		LeaderBot_RegisterColumnForScrolling(7, "crowdcontroller", 7, columnFrames[7], rogueDropDownName);

		LeaderBot_SetupScrollBar("crowdcontroller");
	end 

    local dd = _G["CrowdControllerChatChannelDropDown"];
	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
        CrowdController_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	end

end

-- ************************************************************
function CrowdController_AnnounceButtonEnable(state)
    CCLmsg("Function: CrowdController_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["CrowdControllerAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["CrowdControllerAnnounce" .. i];
	end 
end 

-- ************************************************************
function CrowdController_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	HUNLmsg("HunterLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	CrowdController_MsgChannel[LOCKCHANTYPE] = arg1;
	CrowdController_MsgChannel[LOCKCHANLANG] = nil;
	CrowdController_MsgChannel[LOCKCHANNUMB] = arg2;

	CrowdController_AnnounceButtonEnable(1);
	
end
-- ***************************************************************
function CrowdController_AnnounceToChat(title, assignmentIndex)
	HLmsg("Function: CrowdController_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	local chanType = CrowdController_MsgChannel[LOCKCHANTYPE];
	local chanLang = CrowdController_MsgChannel[LOCKCHANLANG];
	local chanNumber = CrowdController_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot - Crowd Controller: No chat channel selected");
		return;
	end

	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

    --we don't presently have any multiselect options for Crowd Controller. This block is left in
    --as reference for future implementation.
	if (assignmentIndex == FORTITUDE_MULTISELECT or assignmentIndex == HEALS_MULTISELECT) then 

		for playerName, playerArray in next, HealerLeader_Assignments[assignmentIndex] do

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(playerName);
			if (isInGroup) then 

				local msg = "";
				local comma = "";

				for assignmentName, assignmentValue in next, playerArray do 

					if (assignmentName ~= LeaderBotNone and assignmentValue == true) then 
						msg = assignmentName .. comma .. msg;
						comma = ", ";
					end

				end 

				if (msg ~= "") then 
					msg = playerName .. ": " .. msg;
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end

				end
			end 

		end 


	else 

		for assignmentKey, assignmentValue in next, CrowdController_Assignments[assignmentIndex] do
			local playerName, target;
			local msg = nil;

			playerName = assignmentKey;
			targetName = assignmentValue;

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(playerName);
			if (isInGroup) then 

				--skip players assigned to LeaderBotNone, they're probably healing classes not playing
				--a healing role.
				if (targetName ~= LeaderBotNone ) then 
					msg = "     " .. playerName .. ": " .. targetName .. " " ;

					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end
				end
				
				--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
				--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
				--sync everything being announced.
				if (string.find(assignmentIndex, "MultiSelect") ) then 
					for key_TargetName, val_CheckMarkStatus in next, CrowdController_Assignments[assignmentIndex][playerName] do
						LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "crowdcontroller", assignmentIndex, playerName, key_TargetName, val_CheckMarkStatus);
					end
				else 
					LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "crowdcontroller", assignmentIndex, playerName, CrowdController_Assignments[assignmentIndex][playerName]);
				end  

			end

		end -- end for

	end -- if assignmentIndex

	if (chanNumber == nil) then
		--Party, raid
		SendChatMessage("-------------------------------------------", chanType);
	else
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
end 

