--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
	All code here is inside functions. Nothing runs when the .TOC loads, other than the
	high-level variable definitions. Execution beings in LeaderBot.lua:OnLoad()

--]]

local classicOrRetail;
local HL_Debug = 0;
local HL_loopcontrol = 0;
local version = "1.39";

HealerLeaderFrame = nil;

local HEALS_MULTISELECT = "Heals_MultiSelect"
local FORTITUDE_MULTISELECT = "Fortitude_MultiSelect";
local SHACKLE = "Shackle";
local HEALERASSIGNMENTCOUNT = 3;

--Global variables to be saved
--healer leader assignments must be global to allow saving
HealerLeader_Assignments = {};
HealerLeader_Assignments[HEALS_MULTISELECT] = {};
HealerLeader_Assignments[FORTITUDE_MULTISELECT] = {};
HealerLeader_Assignments[SHACKLE] = {};

local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local HealerLeader_MsgChannel ={};

local columnWidth = 150;
local columnWidthHeal = 160;
local columnWidthFort = 150;
local columnWidthSheep = 100;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CurseOfAgonyID = 11713;
local CurseOfDoomID = 603;
local CurseOfTonguesID = 11719;
local CurseOfRecklessnessID = 7658;
local CurseOfWeaknessID = 7646;
local CurseOfElementsID = 11721;
local CurseOfShadowID = 17862;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local variablesLoaded = false;
local addHealerIndex = 0;
local addHealerTargetIndex = 0;
local addHealerFortitudeIndex = 0;

local healerDropDownName = "HealerLeaderHealDropDown";
local fortitudeDropDownName = "HealerLeaderFortitudeDropDown";
local shackleDropDownName = "HealerLeaderShackleDropDown";

local healerDropDowns = {
						[HEALS_MULTISELECT] = healerDropDownName,
						[FORTITUDE_MULTISELECT] = fortitudeDropDownName,
						[SHACKLE] = shackleDropDownName,
						};

local healerRaid = "Raid";
local healerMelee = "Melee";
local healerRanged = "Ranged";

local group1 = "Group 1";
local group2 = "Group 2";
local group3 = "Group 3";
local group4 = "Group 4";
local group5 = "Group 5";
local group6 = "Group 6";
local group7 = "Group 7";
local group8 = "Group 8";

local colorCode = "|cff5566ff ";
local whiteColorCode = "|cffffffff";
local greyColorCode = "|cff444444";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function HLmsg(msg)
	if (HL_Debug == 1 ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function HealerLeaderFrame_OnLoad(parentFrame, teamMembers, healerCount)
	HLmsg("Function: HealerLeaderFrame_OnLoad()");

	--HealerLeaderFrame = CreateFrame("Frame", "HealerLeaderFrame", parentFrame); -- wow classic
	HealerLeaderFrame = CreateFrame("Frame", "HealerLeaderFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate"); -- wow TBC

	-- UI Commands
	SLASH_HealerLeader1 = "/hl";
	SLASH_HealerLeader2 = "/HL";

	SlashCmdList["HealerLeader"] = HealerLeader_SlashHandler;

	--HealerLeaderFrame:RegisterEvent("VARIABLES_LOADED");
	--HealerLeaderFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	--HealerLeaderFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
     --

	ChatFrame1:AddMessage(colorCode .. "HealerLeader Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's healerLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	HealerLeader_InitMainFrame(teamMembers, healerCount);

	return (HealerLeaderFrame);
end;

-- *************************************************************************************
function HealerLeader_InitMainFrame(teamMembers, healerCount)
	HLmsg("Function: HealerLeader_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove healers

	HealerLeaderFrame:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 2);	
	HealerLeader_AdjustMainPanelSize(healerCount);	

	HealerLeaderFrame:SetBackdrop(mainBackdrop);
	HealerLeaderFrame:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	HealerLeaderFrame:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	HealerLeaderFrame:EnableMouse(true);
	HealerLeaderFrame:SetScript("OnUpdate", HealerLeaderFrame_OnUpdate);
	HealerLeaderFrame:SetScript("OnEvent", HealerLeaderFrame_OnEvent);

	HealerLeaderFrame:SetScript("OnShow", function(self)
						HLmsg("Function: HealerLeaderFrame_OnShow() ");
						HLmsg("current panel = " .. LeaderBotOptions["currentPanel"] );

						if (LeaderBotOptions["currentPanel"] == "healer") then 
							local parentFrame = HealerLeaderFrame:GetParent();
							parentFrame:SetWidth(HealerLeaderFrame:GetWidth() );
						end					
				end);

	local appLogo = LeaderBot_GetTextForClass("Priest");
	local appTitle = HealerLeaderFrame:CreateFontString("HealerLeaderAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", HealerLeaderFrame, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Healer Leader " .. version);

	local infoText = HealerLeaderFrame:CreateFontString("HealerLeaderInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	infoText:SetText(colorCode .. "Showing all classes that could be healers. Assign Non-healers to \"-None-\"\nto exclude them from assignment announcements.");

	columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, healerCount, "healer", HealerLeaderFrame, "Healers");
	columnFrames[2] = HealerLeader_AddColumnHealTargets(teamMembers, healerCount, true);
	columnFrames[3] = HealerLeader_AddColumnFortitudeTargets(teamMembers, healerCount, true);
	columnFrames[4] , HealerLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, healerCount, true,
							HealerLeaderFrame, columnFrames[3],
							columnWidthSheep, "HealerLeaderColumn4", "Shackle",
							"LeaderBothealerMember", shackleDropDownName,
							HealerLeader_Assignments, SHACKLE,
							"HealerLeaderAnnounce4", HealerLeader_AnnounceToChat, "Leader Bot: Shackle Targets",
							HealerLeader_ShackleDropDownOnClick,
							LeaderBotSingleSelect);

	--setup scrolling, if needed
	if (healerCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(4, "healer", 1, columnFrames[1], "LeaderBothealerMember");
		LeaderBot_RegisterColumnForScrolling(4, "healer", 2, columnFrames[2], healerDropDownName);		
		LeaderBot_RegisterColumnForScrolling(4, "healer", 3, columnFrames[3], fortitudeDropDownName);
		LeaderBot_RegisterColumnForScrolling(4, "healer", 4, columnFrames[4], shackleDropDownName);

		LeaderBot_SetupScrollBar("healer");
	end

	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","HealerLeaderChatChannelDropDown", HealerLeaderFrame, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );

	dd:SetPoint("BottomLeft", HealerLeaderFrame, "BottomLeft", -15, 0);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelList, HealerLeader_ChannelDropDownOnClick) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();
	UIDropDownMenu_SetText(dd, "<Select channel>");
	
	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end
	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);

	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end

end


-- ******************************************************************************
--adjust the height as players join or leave
function HealerLeader_AdjustMainPanelSize(healerCount)
	HLmsg("Function: HealerLeader_AdjustMainPanelSize(count = " .. healerCount .. ")" );

	local numberOfRows = healerCount;
	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidth 
						+ columnWidthHeal
						+ columnWidthFort
						+ columnWidthSheep
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	
	HealerLeaderFrame:SetSize(math.max(requiredWidth, LeaderBotMinWidth), requiredHeight);
end

-- ******************************************************************************
function HealerLeader_AddColumnHealTargets(teamMembers, healerCount, forceDisable)
	HLmsg("Function: HealerLeader_AddColumnHealTargets()" );
	
	local columnFrame = _G["HealerLeaderColumn2"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "HealerLeaderColumn2", HealerLeaderFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[1], "TopLeft", columnWidth + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, healerCount, columnWidthHeal)

	local colTitle = _G["HealerLeaderColumn2Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("HealerLeaderColumn2Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Healing");

	local healerTeam = {};
	addHealerTargetIndex = 0;
	for class, healer in next, teamMembers do
		if (LeaderBotFrame_IsMemberOfGroup(class, "healer") ) then
			HealerLeader_AddHealTargetsToColumn(columnFrame, class, teamMembers[class]);
		end
	end

	HLmsg("     addHealerTargetIndex=" .. addHealerTargetIndex);

	--there may be leftover icons if we've had players leave
	addHealerTargetIndex = addHealerTargetIndex + 1;
	local dd = _G[healerDropDownName .. addHealerTargetIndex];
	while (dd ~= nil) do
		dd:Hide();

		addHealerTargetIndex = addHealerTargetIndex+1;
		dd = _G[healerDropDownName .. addHealerTargetIndex];
	end
		
	local button = _G["HealerLeaderAnnounce2"];
	if (button == nil) then
		button = CreateFrame("Button", "HealerLeaderAnnounce2", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce Healing assignments to selected chat channels";
	button:SetSize(75,30);

	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											HLmsg("Hander: HealerLeaderAnnounce2.OnClick()");
											HealerLeader_AnnounceToChat("Leader Bot: Healing assignments", HEALS_MULTISELECT);
								end);

	if (healerCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame;

end
 -- *****************************************************************************
 function HealerLeader_AddHealTargetsToColumn(columnFrame, class, healTeamMembers)
	HLmsg("Function: HealerLeader_AddHealTargetsToColumn(" .. class .. ")" )

	for healerName, data in next, healTeamMembers do
		HLmsg("     adding dropdown for healer: " .. healerName);

		addHealerTargetIndex = addHealerTargetIndex + 1;

		-- make the drop down menu for heal targets
		local dd = _G[healerDropDownName .. addHealerTargetIndex];
		if (dd == nil) then
			dd = CreateFrame("Frame", healerDropDownName..addHealerTargetIndex, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd, columnWidthHeal - 35);
		UIDropDownMenu_EnableDropDown(dd);
		dd:Show();

		local topBuffer = 2;
		if (addHealerTargetIndex > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (addHealerTargetIndex -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() HealerLeader_HealDropDownInit(healerName, dd) end;
		dd.healerName = healerName;

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[healerDropDownName .. addHealerTargetIndex .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", healerDropDownName .. addHealerTargetIndex .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[healerDropDownName .. addHealerTargetIndex .."label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(healerDropDownName .. addHealerTargetIndex .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		LeaderBot_InitNewMultiSelectPlayer(healerName, HealerLeader_Assignments, HEALS_MULTISELECT);
		local labelText = LeaderBot_UpdateMultiselectDropDownLabel(HealerLeader_Assignments, HEALS_MULTISELECT, healerName);
		ddMyLabel:SetText(whiteColorCode .. labelText);

		--UIDropDownMenu_SetSelectedValue(dd, dd.healerTarget);
	end

 end

-- ******************************************************************************
function HealerLeader_HealDropDownInit(healerName, dropDownMenu)
	HLmsg("Function: HealerLeader_HealDropDownInit(" .. healerName .. ")" );

	--Todo: get all the raid members that are designated main tanks or assists.


	local classIcons =
	{
		["WARRIOR"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:0:64|t",
		["MAGE"]=		"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:0:64|t",
		["ROGUE"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:0:64|t",
		["DRUID"]=	 	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:0:64|t",
		["HUNTER"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:64:128|t",
		["SHAMAN"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:64:128|t",
		["PRIEST"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:64:128|t",
		["WARLOCK"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:64:128|t",
		["PALADIN"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:128:196|t",
	 }

	--create the dropdown row
	local rowInfo = UIDropDownMenu_CreateInfo();

	rowInfo.text = "Multi-Select:";
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;
	rowInfo.isTitle = true;		
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo = UIDropDownMenu_CreateInfo();

	if (LeaderBotIsAssistant) then 
		rowInfo.isTitle = false;
	else
		rowInfo.isTitle = true;
	end 

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][LeaderBotNone] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][LeaderBotNone] = true;
	end	

	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][LeaderBotNone];
	rowInfo.text = LeaderBotNone;
	rowInfo.value = LeaderBotNone;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRaid] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRaid] = false;
	end	

	rowInfo.isNotRadio = true;	
	rowInfo.keepShownOnClick = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRaid];
	rowInfo.text = healerRaid;
	rowInfo.value = healerRaid;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerMelee] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerMelee] = false;
	end	

	rowInfo.isNotRadio = true;	
	rowInfo.keepShownOnClick = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerMelee];
	rowInfo.text = healerMelee;
	rowInfo.value = healerMelee;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRanged] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRanged] = false;
	end	

	rowInfo.isNotRadio = true;	
	rowInfo.keepShownOnClick = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][healerRanged];
	rowInfo.text = healerRanged;
	rowInfo.value = healerRanged;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.isNotRadio = true;	
	rowInfo.text ="   ----    ";
	rowInfo.isTitle = true;
	rowInfo.checked = false;
	UIDropDownMenu_AddButton(rowInfo);

	--it appears that once you set isTitle to true, it always stays true until you recreate the 
	--row info, regardless of what you set it to. ><
	rowInfo = UIDropDownMenu_CreateInfo();
	if (LeaderBotIsAssistant) then 
		rowInfo.isTitle = false;
	else
		rowInfo.isTitle = true;
	end 

	if (IsInRaid() ) then
		local raidSize = GetNumGroupMembers();

		for raidIndex = 1, raidSize do
			local name, rank, subgroup, level, class, fileName, 
  					zone, online, isDead, role, isML, combatRole = GetRaidRosterInfo(raidIndex);

			if (role) then

				if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][name] == nil) then 
					HealerLeader_Assignments[HEALS_MULTISELECT][healerName][name] = false;
				end	

				--create the dropdown row
				rowInfo.keepShownOnClick = true;					
				rowInfo.isNotRadio = true;		
				rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][name];
				rowInfo.text = name;
				rowInfo.value = name;
				rowInfo.owner  = dropDownMenu;

				--TODO: Fix the icons of the target
				--rowInfo.icon ="Interface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES";
				--rowInfo.tCoordLeft, rowInfo.tCoordRight, rowInfo.tCoordTop, rowInfo.tCoordBottom  = unpack(CLASS_ICON_TCOORDS[displayClassCaps]);
				rowInfo.arg1 = healerName;								-- the click handler needs to know which checkbox this dropdown pertains to.
				rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
				UIDropDownMenu_AddButton(rowInfo);		
			end

		end
					
	end

	rowInfo.isNotRadio = true;	
	rowInfo.text ="   ----    ";
	rowInfo.isTitle = true;
	rowInfo.checked = false;
	UIDropDownMenu_AddButton(rowInfo);

	--it appears that once you set isTitle to true, it always stays true until you recreate the 
	--row info, regardless of what you set it to. ><
	rowInfo = UIDropDownMenu_CreateInfo();
	if (LeaderBotIsAssistant) then 
		rowInfo.isTitle = false;
	else
		rowInfo.isTitle = true;
	end 

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group1] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group1] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group1];
	rowInfo.text = group1;
	rowInfo.value = group1;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group2] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group2] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group2];
	rowInfo.text = group2;
	rowInfo.value = group2;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group3] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group3] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group3];
	rowInfo.text = group3;
	rowInfo.value = group3;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group4] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group4] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group4];
	rowInfo.text = group4;
	rowInfo.value = group4;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group5] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group5] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;	
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group5];
	rowInfo.text = group5;
	rowInfo.value = group5;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group6] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group6] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group6];
	rowInfo.text = group6;
	rowInfo.value = group6;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group7] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group7] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group7];
	rowInfo.text = group7;
	rowInfo.value = group7;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	if (HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group8] == nil) then 
		HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group8] = false;
	end	

	rowInfo.keepShownOnClick = true;	
	rowInfo.isNotRadio = true;
	rowInfo.checked = HealerLeader_Assignments[HEALS_MULTISELECT][healerName][group8];
	rowInfo.text = group8;
	rowInfo.value = group8;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_HealDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function HealerLeader_HealDropDownOnClick(dropDownButton, healerName, arg2, checked)
	HLmsg("Function: HealerLeader_HealDropDownOnClick(" .. healerName .. ")" );

	local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list. DO NOT CALL THIS ON MULTISELECT
	--UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock
	HealerLeader_Assignments[HEALS_MULTISELECT][healerName][dropDownButton.value] = UIDropDownMenuButton_GetChecked(dropDownButton);

	if (dropDownButton.value == LeaderBotNone and UIDropDownMenuButton_GetChecked(dropDownButton) ) then 
		-- they selected "None" so uncheck everybody else	

		--TODO: well, now what? Can't find a function to uncheck this one button without resetting all the other selections.

	elseif (UIDropDownMenuButton_GetChecked(dropDownButton)) then 
		-- they checked any button, so uncheck None
		local ddNone = _G[healerDropDownName.. "1"];
		
		--TODO: well, now what? Can't find a function to uncheck this one button without resetting all the other selections.
	end 

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.healerName = healerName;
	dropDownFrame.healerTarget = dropDownButton.value;

	--dropDownFrame.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);
	local labelText = LeaderBot_UpdateMultiselectDropDownLabel(HealerLeader_Assignments, HEALS_MULTISELECT, healerName);
	dropDownFrame.ddMyLabel:SetText(whiteColorCode .. labelText);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "healer", HEALS_MULTISELECT, healerName, dropDownButton.value, UIDropDownMenuButton_GetChecked(dropDownButton))

end

-- ******************************************************************************
function HealerLeader_AddColumnFortitudeTargets(teamMembers, healerCount, forceDisable)
	HLmsg("Function: HealerLeader_AddColumnFortitudeTargets()" );

	local columnFrame = _G["HealerLeaderColumn3"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "HealerLeaderColumn3", HealerLeaderFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[2], "TopLeft", columnWidthHeal + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, healerCount, columnWidthFort)

	local colTitle = _G["HealerLeaderColumn3Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("HealerLeaderColumn3Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Fortitude");

	local healerTeam = {};
	addHealerFortitudeIndex = 0;
	for class, healer in next, teamMembers do
		if (LeaderBotFrame_IsMemberOfGroup(class, "healer") ) then
			HealerLeader_AddFortitudeTargetsToColumn(columnFrame, class, teamMembers[class]);
		end
	end

	HLmsg("     addHealerFortitudeIndex=" .. addHealerFortitudeIndex);

	--there may be leftover icons if we've had players leave
	addHealerFortitudeIndex = addHealerFortitudeIndex + 1;
	local dd = _G[fortitudeDropDownName .. addHealerFortitudeIndex];
	while (dd ~= nil) do
		dd:Hide();

		addHealerFortitudeIndex = addHealerFortitudeIndex+1;
		dd = _G[fortitudeDropDownName .. addHealerFortitudeIndex];

	end
		
	local button = _G["HealerLeaderAnnounce3"];
	if (button == nil) then
		button = CreateFrame("Button", "HealerLeaderAnnounce3", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce Fortitude assignments to selected chat channels";
	button:SetSize(75,30);

	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											HLmsg("Hander: HealerLeaderAnnounce3.OnClick()");
											HealerLeader_AnnounceToChat("Leader Bot: Fortitude assignments", FORTITUDE_MULTISELECT);
								end);

	if (healerCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame;

end
 -- *****************************************************************************
 function HealerLeader_AddFortitudeTargetsToColumn(columnFrame, class, healTeamMembers)
	HLmsg("Function: HealerLeader_AddFortitudeTargetsToColumn(" .. class .. ")" )

	for healerName, data in next, healTeamMembers do
		HLmsg("     adding dropdown for healer: " .. healerName);

		addHealerFortitudeIndex = addHealerFortitudeIndex + 1;

		-- make the drop down menu for heal targets
		local dd = _G[fortitudeDropDownName .. addHealerFortitudeIndex];
		if (dd == nil) then
			dd = CreateFrame("Frame", fortitudeDropDownName .. addHealerFortitudeIndex, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd, columnWidthFort - 35);
		UIDropDownMenu_EnableDropDown(dd);
		dd:Show();

		local topBuffer = 2;
		if (addHealerFortitudeIndex > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (addHealerFortitudeIndex -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() HealerLeader_FortitudeDropDownInit(healerName, dd) end;
		dd.healerName = healerName;

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[fortitudeDropDownName .. addHealerFortitudeIndex .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", fortitudeDropDownName .. addHealerFortitudeIndex .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[fortitudeDropDownName .. addHealerFortitudeIndex .."label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(fortitudeDropDownName .. addHealerFortitudeIndex .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );
		
		LeaderBot_InitNewMultiSelectPlayer(healerName, HealerLeader_Assignments, FORTITUDE_MULTISELECT);
		local labelText = LeaderBot_UpdateMultiselectDropDownLabel(HealerLeader_Assignments, FORTITUDE_MULTISELECT, healerName);
		ddMyLabel:SetText(whiteColorCode .. labelText);

	 	if (UnitClass(healerName) ~= "Priest") then
			--no fortitude from this guy
			UIDropDownMenu_DisableDropDown(dd);
			ddMyLabel:SetText(greyColorCode .. "N/A");
		end 

		--DO NOT CALL THIS FOR MULTISELECT DROPDOWNS
		--UIDropDownMenu_SetSelectedValue(dd, dd.fortitudeTarget);

	end

 end

-- ******************************************************************************
function HealerLeader_FortitudeDropDownInit(healerName, dropDownMenu)
	HLmsg("Function: HealerLeader_FortitudeDropDownInit(" .. healerName .. ")" );

	--create the dropdown row
	local rowInfo = UIDropDownMenu_CreateInfo();

	rowInfo.text = "Multi-Select:";
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;
	rowInfo.isTitle = true;		
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo = UIDropDownMenu_CreateInfo();

	if (LeaderBotIsAssistant) then 
		rowInfo.isTitle = false;
	else
		rowInfo.isTitle = true;
	end 

	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][LeaderBotNone] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][LeaderBotNone] = true;
	end	

	rowInfo.text = LeaderBotNone;
	rowInfo.value = LeaderBotNone;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][LeaderBotNone];
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group1] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group1] = false;
	end	

	rowInfo.text = group1;
	rowInfo.value = group1;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group1];	
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group2] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group2] = false;
	end	

	rowInfo.text = group2;
	rowInfo.value = group2;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group2];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group3] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group3] = false;
	end	

	rowInfo.text = group3;
	rowInfo.value = group3;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group3];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group4] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group4] = false;
	end	

	rowInfo.text = group4;
	rowInfo.value = group4;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group4];	
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group5] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group5] = false;
	end	

	rowInfo.text = group5;
	rowInfo.value = group5;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group5];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group6] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group6] = false;
	end	

	rowInfo.text = group6;
	rowInfo.value = group6;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group6];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group7] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group7] = false;
	end	

	rowInfo.text = group7;
	rowInfo.value = group7;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group7];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group8] == nil) then 
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group8] = false;
	end	

	rowInfo.text = group8;
	rowInfo.value = group8;
	rowInfo.checked = HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][group8];		
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;		
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.arg1 = healerName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HealerLeader_FortitudeDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function HealerLeader_FortitudeDropDownOnClick(dropDownButton, healerName, arg2, checked)
	HLmsg("Function: HealerLeader_FortitudeDropDownOnClick(" .. healerName .. ")" );

	local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list. DO NOT CALL THIS ON MULTISELECT Dropdowns
	--UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock
	HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName][dropDownButton.value] = UIDropDownMenuButton_GetChecked(dropDownButton);

	
	if (dropDownButton.value == LeaderBotNone and UIDropDownMenuButton_GetChecked(dropDownButton) ) then 
		-- they selected "None" so uncheck everybody else	

		--TODO: well, now what? Can't find a function to uncheck this one button without resetting all the other selections.

	elseif (UIDropDownMenuButton_GetChecked(dropDownButton)) then 
		-- they checked any button, so uncheck None
		local ddNone = _G[fortitudeDropDownName.. "1"];
		
		--TODO: well, now what? Can't find a function to uncheck this one button without resetting all the other selections.
	end 

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.healerName = healerName;
	dropDownFrame.fortitudeTarget = dropDownButton.value;

	--dropDownFrame.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);
	local labelText = LeaderBot_UpdateMultiselectDropDownLabel(HealerLeader_Assignments, FORTITUDE_MULTISELECT, healerName);
	dropDownFrame.ddMyLabel:SetText(whiteColorCode .. labelText);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "healer", FORTITUDE_MULTISELECT, healerName, dropDownButton.value, UIDropDownMenuButton_GetChecked(dropDownButton) );

	--CloseDropDownMenus();

end


-- *********************************************************************************
function HealerLeader_ShackleDropDownOnClick(dropDownButton, arg1, arg2, checked)
	HLmsg("Function: HealerLeader_ShackleDropDownOnClick(" .. arg1 .. ")" );

		--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	local healerName = arg1;
	local shackleTargetIcon = arg2;
	HealerLeader_Assignments[SHACKLE][healerName] = shackleTargetIcon;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);


	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "healer", SHACKLE, healerName, shackleTargetIcon);
	
end


--******************************************************************************************************
-- Command line parser
function HealerLeader_SlashHandler(arg1)
	HLmsg("Function: HealerLeader_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. "HealerLeader by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. "Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (HL_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. "HealerLeader debug messages disabled");
			HL_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. "HealerLeader debug messages enabled");
			HL_Debug = 1;
		end
	end

end

--******************************************************************************************************
-- EventHandler
function HealerLeaderFrame_OnEvent(self, event, ...)
	local arg1 = select(1, ...);
	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	HLmsg("Function: HealerLeaderFrame_OnEvent(Event: " .. event .. ", " .. arg1 .. ")" );


	-- *****************************************************************************************
	if (event == "VARIABLES_LOADED") then
			HLmsg ("     VARIABLES LOADED HANDLER");

			--used to prevent the OnShow() handler from trying to render a list before we have names
			--loaded.
			variablesLoaded = true;

			HLmsg("     preparing to run Update Display()");

	
	-- *****************************************************************************************
	-- lookup details of the event. The lookup method varies based on the even type
	elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		local timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, b_or_d = select(1, ...);
		HLmsg("event = " .. timestamp);


	-- ******************************************************************************************
	elseif (ttype == "SPELL_AURA_REMOVED" and dstGUID == UnitGUID("player") ) then
		--NightLightFrame_RemoveTimer(spellid)
	end
end

-- ************************************************
function HealerLeader_DataMigration()
	HLmsg("Function: HealerLeader_DataMigration()" );

	if (LeaderBotOptions["HealerLeaderDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		HealerLeader_Assignments = {};
		HealerLeader_Assignments[HEALS_MULTISELECT] = {};
		HealerLeader_Assignments[FORTITUDE_MULTISELECT] = {};
		HealerLeader_Assignments[SHACKLE] = {};

		LeaderBotOptions["HealerLeaderDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for assignment = 1, HEALERASSIGNMENTCOUNT do
		if (HealerLeader_Assignments[assignment] == nil) then
			HealerLeader_Assignments[assignment] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, HealerLeader_Assignments[assignment] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then  	
					HealerLeader_Assignments[assignment][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************
function HealerLeader_RosterChanged(newTeamMembers, newHealerCount)
	HLmsg("Function: HealerLeader_RosterChanged()");

--[[
	--TODO: At first, I thought this was a good idea to wipe out assignments when a player leaves the group. However,
	--I found that upon logging in, events fire 4-6 times that call _RosterChanged(). and sometimes those events fire
	--before the group is "formed" in Blizzard's mind. So, I'd get here, see group members "were gone" and remove them. 
	--then 3 events later find them again and add them to the table with blank assignments. So I'm taking this chunk out
	--for now in an attempt to preserve assignments after re-logging.

	-- walk each assignment type (HEAL, FORTITUDE, ETC)
	for assignmentType in next, HealerLeader_Assignments do
	
		-- walk each healer in that assignment type
		for healerNameKey, targetValue in next, HealerLeader_Assignments[assignmentType] do

			--local found = LeaderBotFrame_IsMemberOfTeam(healerNameKey, newTeamMembers);
			local playerClass = UnitClass(healerNameKey);
			if (playerClass == nil) then
				--this healer has left the building
				--remove their assignments
				HealerLeader_Assignments[assignmentType][healerNameKey] = nil;
			end 

		end --for healerNameKey

	end --for assignmentType
--]]

	HealerLeader_AdjustMainPanelSize(newHealerCount);
	columnFrames[1] = LeaderBotFrame_AddColumn1(newTeamMembers, newHealerCount, "healer", HealerLeaderFrame, "Healers");
	columnFrames[2] = HealerLeader_AddColumnHealTargets(newTeamMembers, newHealerCount, false);
	columnFrames[3] = HealerLeader_AddColumnFortitudeTargets(newTeamMembers, newHealerCount, false);
	columnFrames[4] , HealerLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newHealerCount, true,
							HealerLeaderFrame, columnFrames[3],
							columnWidthSheep, "HealerLeaderColumn4", "Shackle",
							"LeaderBothealerMember", shackleDropDownName,
							HealerLeader_Assignments, SHACKLE,
							"HealerLeaderAnnounce4", HealerLeader_AnnounceToChat, "Leader Bot: Shackle Targets",
							HealerLeader_ShackleDropDownOnClick,
							LeaderBotSingleSelect);

	if (newHealerCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(4, "healer", 1, columnFrames[1], "LeaderBothealerMember");
		LeaderBot_RegisterColumnForScrolling(4, "healer", 2, columnFrames[2], healerDropDownName);		
		LeaderBot_RegisterColumnForScrolling(4, "healer", 3, columnFrames[3], fortitudeDropDownName);
		LeaderBot_RegisterColumnForScrolling(4, "healer", 4, columnFrames[4], shackleDropDownName);

		LeaderBot_SetupScrollBar("healer");
	end

	local dd = _G["HealerLeaderChatChannelDropDown"];
	if (LeaderBotIsAssistant == false) then
		HealerLeader_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	else
		UIDropDownMenu_EnableDropDown(dd);
	end 

end

-- ******************************************************************************
function HealerLeader_AnnounceButtonEnable(state)
	HLmsg("FUnction: HealerLeader_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["HealerLeaderAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["HealerLeaderAnnounce" .. i];
	end 

end

-- ******************************************************************************
function isPlayerATank(playerName)
	HLmsg("Function: HealerLeader_AnnounceToChat(" .. playerName .. ")" );
				
	local raidSize = GetNumGroupMembers();
	for raidIndex = 1, raidSize do
		local name, _, _, _, _, _, _, _, _, role = GetRaidRosterInfo(raidIndex);
		if (name == playerName) then
			-- we found this player, return their role assignment: nil, MAINTANK or MAINASSIST
			return role;
		end 
	end 

	--indicates this player is no longer in raid.
	return nil;

end 


-- ******************************************************************************
function HealerLeader_AnnounceToChat(title, assignmentIndex)
	HLmsg("Function: HealerLeader_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	local chanType = HealerLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = HealerLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = HealerLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot - Healer Leader: No chat channel selected");
		return;
	end

	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

	if (assignmentIndex == FORTITUDE_MULTISELECT or assignmentIndex == HEALS_MULTISELECT) then 

		for playerName, playerArray in next, HealerLeader_Assignments[assignmentIndex] do

			--leader assignments save assignments for previous players, so only announce assignments for casters in the raid
			local isInGroup = UnitClass(playerName);
			if (isInGroup) then 

				local msg = "";
				local comma = "";

				for assignmentName, assignmentValue in next, playerArray do 

					--leader assignments save assignments for previous players, so only announce assignments for targets in the raid
					local isInGroup = UnitClass(assignmentName);

					--The assignee may no longer be a tank, even if they are in raid
					local isTank;
					if (isInGroup)  then
						isTank = isPlayerATank(assignmentName);
					end

					if ( (isInGroup and isTank) 																--player is in raid and a tank (heal target)
						or assignmentName == "Raid" or assignmentName == "Melee" or assignmentName == "Ranged"  --a generic heal assignment
						or string.match(assignmentName, "Group ") 												--a group assignment for fortitude
						) then
							if (assignmentName ~= LeaderBotNone and assignmentValue == true) then 
								msg = assignmentName .. comma .. msg;
								comma = ", ";
							end

					end

				end 

				if (msg ~= "") then 
					msg = playerName .. ": " .. msg;
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end		

					--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
					--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
					--sync everything being announced.
					if (string.find(assignmentIndex, "MultiSelect") ) then 
						for key_TargetName, val_CheckMarkStatus in next, HealerLeader_Assignments[assignmentIndex][playerName] do
							LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "healer", assignmentIndex, playerName, key_TargetName, val_CheckMarkStatus);
						end
					else 
						LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "healer", assignmentIndex, playerName, HealerLeader_Assignments[assignmentIndex][playerName]);
					end 

				end
			end 

		end 


	else 

		for assignmentKey, assignmentValue in next, HealerLeader_Assignments[assignmentIndex] do
			local name, target;
			local msg = nil;

			name = assignmentKey;
			targetName = assignmentValue;

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(name);
			if (isInGroup) then 

				--skip players assigned to LeaderBotNone, they're probably healing classes not playing
				--a healing role.
				if (targetName ~= LeaderBotNone ) then 
					msg = "     " .. name .. ": " .. targetName .. " " ;

					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end
				end 

			end

		end -- end for

	end -- if assignmentIndex

	if (chanNumber == nil) then
		--Party, raid
		SendChatMessage("-------------------------------------------", chanType);
	else
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
								
end


-- ******************************************************************************
function HealerLeader_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	HLmsg("HealerLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	HealerLeader_MsgChannel[LOCKCHANTYPE] = arg1;
	HealerLeader_MsgChannel[LOCKCHANLANG] = nil;
	HealerLeader_MsgChannel[LOCKCHANNUMB] = arg2;

	HealerLeader_AnnounceButtonEnable(1);
	
end


-- ***************************************************************************
function HealerLeader_InitNewMultiSelectHealer(healerName)
 
	--Somewhere is resetting HealerLeader_Assignments, but MULTISELECT hasn't been added there yet.
	--I can't find where, so am forcing the issue here.
	if (not HealerLeader_Assignments[FORTITUDE_MULTISELECT]) then
		HealerLeader_Assignments[FORTITUDE_MULTISELECT] = {}
	end 

	--on first click for this healer, the array won't exist yet. Init it here.
	if (not HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName]) then
		HealerLeader_Assignments[FORTITUDE_MULTISELECT][healerName] = {};
	end 
	
end 

