--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages hunter assignments during raids
	
	All code here is inside functions. Nothing runs when the .TOC loads, other than the
	high-level variable definitions. Execution beings in LeaderBot.lua:OnLoad()

--]]

local classicOrRetail;
local HUNL_Debug = 0;
local version = "1.12";

HunterLeaderFrame = nil;

local HUNTERPULL = "Pull"
local HUNTERTRANQ = "Tranq Shot"
local HUNTERASSIGNMENTCOUNT = 2;

--Global variables to be saved
--hunter leader assignments must be global to allow saving
HunterLeader_Assignments = {};
HunterLeader_Assignments[HUNTERPULL] = {};
HunterLeader_Assignments[HUNTERTRANQ] = {};

local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local HunterLeader_MsgChannel ={};

local columnWidth = 150;
local columnWidthInt = 150;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CurseOfAgonyID = 11713;
local CurseOfDoomID = 603;
local CurseOfTonguesID = 11719;
local CurseOfRecklessnessID = 7658;
local CurseOfWeaknessID = 7646;
local CurseOfElementsID = 11721;
local CurseOfShadowID = 17862;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local variablesLoaded = false;
local addHealerIndex = 0;
local addHealerTargetIndex = 0;

local pullDropDownName = "HunterLeaderPullDropDown";
local tranqDropDownName = "HunterLeaderTranqDropDown"
local hunterDropDowns = {
						[HUNTERPULL] = pullDropDownName,
						[HUNTERTRANQ] = tranqDropDownName,
						};

local hunterRaid = "Raid";

local group1 = "Group 1";
local group2 = "Group 2";
local group3 = "Group 3";
local group4 = "Group 4";
local group5 = "Group 5";
local group6 = "Group 6";
local group7 = "Group 7";
local group8 = "Group 8";
local group12 = "Groups 1 & 2";
local group34 = "Groups 3 & 4";
local group56 = "Groups 5 & 6";
local group78 = "Groups 7 & 8";

local colorCode = "|cffeeaa11 ";
local whiteColorCode = "|cffffffff";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};
	
local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function HUNLmsg(msg, override)
	if (HUNL_Debug == 1 or override ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function HunterLeaderFrame_OnLoad(parentFrame, teamMembers, hunterCount)
	HUNLmsg("Function: HunterLeaderFrame_OnLoad()");

	--HunterLeaderFrame = CreateFrame("Frame", "HunterLeaderFrame", parentFrame);	--wow classic
	HunterLeaderFrame = CreateFrame("Frame", "HunterLeaderFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate"); -- wow TBC

	-- UI Commands
	SLASH_HunterLeader1 = "/huntl";
	SLASH_HunterLeader2 = "/HUNTL";

	SlashCmdList["HunterLeader"] = HunterLeader_SlashHandler;

	--HunterLeaderFrame:RegisterEvent("VARIABLES_LOADED");
	--HunterLeaderFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	--HunterLeaderFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
     --

	ChatFrame1:AddMessage(colorCode .. "HunterLeader Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's hunterLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	HunterLeader_InitMainFrame(teamMembers, hunterCount);

	return (HunterLeaderFrame);
end;

-- *************************************************************************************
function HunterLeader_InitMainFrame(teamMembers, hunterCount)
	HUNLmsg("Function: HunterLeader_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove hunters

	HunterLeaderFrame:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 2);	
	HunterLeader_AdjustMainPanelSize(hunterCount);	

	HunterLeaderFrame:SetBackdrop(mainBackdrop);
	HunterLeaderFrame:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	HunterLeaderFrame:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	HunterLeaderFrame:EnableMouse(true);
	HunterLeaderFrame:SetScript("OnUpdate", HunterLeaderFrame_OnUpdate);
	HunterLeaderFrame:SetScript("OnEvent", HunterLeaderFrame_OnEvent);

	HunterLeaderFrame:SetScript("OnShow", function(self)
						HUNLmsg("Function: HunterLeaderFrame_OnShow() ");
						HUNLmsg("current panel = " .. LeaderBotOptions["currentPanel"] );

						if (LeaderBotOptions["currentPanel"] == "Hunter") then 
							local parentFrame = HunterLeaderFrame:GetParent();
							parentFrame:SetWidth(HunterLeaderFrame:GetWidth() );
						end					
				end);

	local appLogo = LeaderBot_GetTextForClass("Hunter");
	local appTitle = HunterLeaderFrame:CreateFontString("HunterLeaderAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", HunterLeaderFrame, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Hunter Leader " .. version);

	local infoText = HunterLeaderFrame:CreateFontString("HunterLeaderInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	--infoText:SetText(colorCode .. " ");

	columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, hunterCount, "Hunter", HunterLeaderFrame, "Hunters");
	columnFrames[2], HunterLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, hunterCount, true, 
							HunterLeaderFrame, columnFrames[1], 
							columnWidthInt, "HunterLeaderColumn3", "Pull Target", 
							"LeaderBotHunterMember", pullDropDownName, 
							HunterLeader_Assignments, HUNTERPULL, 
							"HunterLeaderAnnounce3", HunterLeader_AnnounceToChat, "Leader Bot: Hunter Pull Targets",
							HunterLeader_PullDropDownOnClick,
							LeaderBotSingleSelect);	
	columnFrames[3] = HunterLeader_AddColumnTranqOrder(teamMembers, hunterCount, true);

	if (hunterCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 1, columnFrames[1], "LeaderBotHunterMember");
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 2, columnFrames[2], pullDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 3, columnFrames[3], tranqDropDownName);

		LeaderBot_SetupScrollBar("Hunter");
	end 

	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","HunterLeaderChatChannelDropDown", HunterLeaderFrame, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );
	dd:SetPoint("BottomLeft", HunterLeaderFrame, "BottomLeft", -15, 0);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelList, HunterLeader_ChannelDropDownOnClick) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();
	UIDropDownMenu_SetText(dd, "<Select channel>");
	
	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end
	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
	
	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end
end

-- ******************************************************************************
--adjust the height as players join or leave
function HunterLeader_AdjustMainPanelSize(hunterCount)
	HUNLmsg("Function: HunterLeader_AdjustMainPanelSize(count = " .. hunterCount .. ")" );

	local numberOfRows = hunterCount;
	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidth 
						+ columnWidth  --padding to make the frame big enough when only one column is shown.
						+ columnWidthInt
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	
	HunterLeaderFrame:SetSize(math.max(requiredWidth, LeaderBotMinWidth), requiredHeight);
end

-- ******************************************************************************
function HunterLeader_AddColumnTranqOrder(teamMembers, hunterCount, forceDisable)
	HUNLmsg("Function: HunterLeader_AddColumnTranqOrder()" );

	local columnFrame = _G["HunterLeaderColumn2"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "HunterLeaderColumn2", HunterLeaderFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[2], "TopLeft", columnWidthInt + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, hunterCount, columnWidthInt)

	local colTitle = _G["HunterLeaderColumn2Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("HunterLeaderColumn2Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Tranq Order");

	local addHunterTranqIndex = HunterLeader_AddTranqOrderToColumn(columnFrame, "Hunter", teamMembers["Hunter"]);

	--there may be leftover icons if we've had players leave
	addHunterTranqIndex = addHunterTranqIndex + 1;
	local dd = _G[tranqDropDownName .. addHunterTranqIndex];
	while (dd ~= nil) do
		dd:Hide();

		addHunterTranqIndex = addHunterTranqIndex+1;
		dd = _G[tranqDropDownName .. addHunterTranqIndex];
	end
		
	local button = _G["HunterLeaderAnnounce2"];
	if (button == nil) then
		button = CreateFrame("Button", "HunterLeaderAnnounce2", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce Tranq Shot order assignments to selected chat channels";
	button:SetSize(75,30);

	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											HUNLmsg("Hander: HunterLeaderAnnounce2.OnClick()");
											HunterLeader_AnnounceToChat("Leader Bot: Hunter Tranq Shot order assignments", HUNTERTRANQ);
								end);

	if (hunterCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame;

end
 -- *****************************************************************************
 function HunterLeader_AddTranqOrderToColumn(columnFrame, class, hunterTeamMembers)
	HUNLmsg("Function: HunterLeader_AddTranqOrderToColumn(" .. class .. ")" )

	local addIndex = 0;
	for hunterName, data in next, hunterTeamMembers do
		HUNLmsg("     adding tranq dropdown for hunter: " .. hunterName);

		addIndex = addIndex + 1;

		-- make the drop down menu for heal targets
		local dd = _G[tranqDropDownName .. addIndex];
		if (dd == nil) then
			dd = CreateFrame("Frame", tranqDropDownName .. addIndex, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd, columnWidthInt - 35);

		UIDropDownMenu_EnableDropDown(dd);
		if (LeaderBotIsAssistant == false) then 
			UIDropDownMenu_DisableDropDown(dd);
		end 

		local topBuffer = 2;
		if (addIndex > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (addIndex -1) * -LeaderBotplayerHeight  -topBuffer);
		dd.initialize = function() HunterLeader_TranqDropDownInit(hunterName, dd) end;
		dd.hunterName = hunterName;

		--did the lock previously have an assignment? If so, re-apply it.
		local targetName = HunterLeader_Assignments[HUNTERTRANQ][hunterName];
		if (targetName) then
			--they previously had an assignment; re-apply it.
			dd.tranqTarget = targetName;
		else 
			--they have not had an assignment before, so assign them to RESERVE
			dd.tranqTarget = LeaderBotNone;
			HunterLeader_Assignments[HUNTERTRANQ][hunterName] = LeaderBotNone;			
		end 

		--Icons may have been hidden in a prior run of this column
		dd:Show();

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[tranqDropDownName .. addIndex .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", tranqDropDownName .. addIndex .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[tranqDropDownName .. addIndex .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(tranqDropDownName .. addIndex .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		ddMyLabel:SetText(whiteColorCode .. dd.tranqTarget);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		UIDropDownMenu_SetSelectedValue(dd, dd.tranqTarget);

	end

	return addIndex;

 end

-- ******************************************************************************
function HunterLeader_TranqDropDownInit(hunterName, dropDownMenu)
	HUNLmsg("Function: HunterLeader_TranqDropDownInit(" .. hunterName .. ")" );

	--create the dropdown row
	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = LeaderBotNone;
	rowInfo.value = LeaderBotNone;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = "1st";
	rowInfo.value = "1st";
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = "2nd";
	rowInfo.value = "2nd";
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = "3rd";
	rowInfo.value = "3rd";
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = "4th";
	rowInfo.value = "4th";
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = "5th";
	rowInfo.value = "5th";
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = hunterName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = HunterLeader_TranqDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function HunterLeader_TranqDropDownOnClick(dropDownButton, arg1, arg2, checked)
	HUNLmsg("Function: HunterLeader_TranqDropDownOnClick(" .. arg1 .. ")" );

	local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock
	local hunterName = arg1;
	HunterLeader_Assignments[HUNTERTRANQ][hunterName] = dropDownButton.value;

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.hunterName = hunterName;
	dropDownFrame.tranqTarget = dropDownButton.value;

	dropDownFrame.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Hunter", HUNTERTRANQ, hunterName, dropDownButton.value);
end

-- *********************************************************************************
function HunterLeader_PullDropDownOnClick(dropDownButton, arg1, pullTargetIcon, checked)
	HUNLmsg("Function: HunterLeader_PullDropDownOnClick(" .. arg1 .. ")" );

		--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock. I prefer not to rename arg1, arg2 in the
	--function definition so I have documentation as to where those values come from
	local hunterName = arg1;

	HunterLeader_Assignments[HUNTERPULL][hunterName] = pullTargetIcon;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Hunter", HUNTERPULL, hunterName, pullTargetIcon);

end

--******************************************************************************************************
-- Command line parser
function HunterLeader_SlashHandler(arg1)
	HUNLmsg("Function: HunterLeader_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. "HunterLeader by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. "Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (HUNL_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. "HunterLeader debug messages disabled");
			HUNL_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. "HunterLeader debug messages enabled");
			HUNL_Debug = 1;
		end
	end

end

--******************************************************************************************************
-- EventHandler
function HunterLeaderFrame_OnEvent(self, event, ...)
	local arg1 = select(1, ...);
	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	HUNLmsg("Function: HunterLeaderFrame_OnEvent(Event: " .. event .. ", " .. arg1 .. ")" );

	-- *****************************************************************************************
	if (event == "VARIABLES_LOADED") then
			HUNLmsg ("     VARIABLES LOADED HANDLER");

			--used to prevent the OnShow() handler from trying to render a list before we have names
			--loaded.
			variablesLoaded = true;

			HUNLmsg("     preparing to run Update Display()");

	
	-- *****************************************************************************************
	-- lookup details of the event. The lookup method varies based on the even type
	elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		local timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, b_or_d = select(1, ...);
		HUNLmsg("event = " .. timestamp);


	-- ******************************************************************************************
	elseif (ttype == "SPELL_AURA_REMOVED" and dstGUID == UnitGUID("player") ) then
		--NightLightFrame_RemoveTimer(spellid)
	end
end

-- ************************************************
function HunterLeader_DataMigration()
	HUNLmsg("Function: HunterLeader_DataMigration()" );

	if (LeaderBotOptions["HunterLeaderDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		HunterLeader_Assignments = {};

		HunterLeader_Assignments[HUNTERPULL] = {};
		HunterLeader_Assignments[HUNTERTRANQ] = {};

		LeaderBotOptions["HunterLeaderDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for assignment = 1, HUNTERASSIGNMENTCOUNT do
		if (HunterLeader_Assignments[assignment] == nil) then
			HunterLeader_Assignments[assignment] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, HunterLeader_Assignments[assignment] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then  	
					HunterLeader_Assignments[assignment][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************
function HunterLeader_RosterChanged(newTeamMembers, newHunterCount)
	HUNLmsg("Function: HunterLeader_RosterChanged()");

--[[
	--TODO: At first, I thought this was a good idea to wipe out assignments when a player leaves the group. However,
	--I found that upon logging in, events fire 4-6 times that call _RosterChanged(). and sometimes those events fire
	--before the group is "formed" in Blizzard's mind. So, I'd get here, see group members "were gone" and remove them. 
	--then 3 events later find them again and add them to the table with blank assignments. So I'm taking this chunk out
	--for now in an attempt to preserve assignments after re-logging.
	-- walk each assignment type (HEAL, FORTITUDE, ETC)
	for assignmentType in next, HunterLeader_Assignments do
	
		-- walk each hunter in that assignment type
		for hunterNameKey, targetValue in next, HunterLeader_Assignments[assignmentType] do

			--local found = LeaderBotFrame_IsMemberOfTeam(hunterNameKey, newTeamMembers);
			local playerClass = UnitClass(hunterNameKey);
			if (playerClass == nil) then
				--this hunter has left the building
				--remove their assignments
				HunterLeader_Assignments[assignmentType][hunterNameKey] = nil;		
			end 

		end --for hunterNameKey

	end --for assignmentType
--]]

	HunterLeader_AdjustMainPanelSize(newHunterCount);
	columnFrames[1] = LeaderBotFrame_AddColumn1(newTeamMembers, newHunterCount, "Hunter", HunterLeaderFrame, "Hunters");
	columnFrames[2], HunterLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newHunterCount, true, 
							HunterLeaderFrame, columnFrames[1], 
							columnWidthInt, "HunterLeaderColumn3", "Pull Target", 
							"LeaderBotHunterMember", pullDropDownName, 
							HunterLeader_Assignments, HUNTERPULL, 
							"HunterLeaderAnnounce3", HunterLeader_AnnounceToChat, "Leader Bot: Hunter Pull Targets",
							HunterLeader_PullDropDownOnClick,
							LeaderBotSingleSelect);	
	columnFrames[3] = HunterLeader_AddColumnTranqOrder(newTeamMembers, newHunterCount, false);

	if (newHunterCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 1, columnFrames[1], "LeaderBotHunterMember");
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 2, columnFrames[2], pullDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "Hunter", 3, columnFrames[3], tranqDropDownName);

		LeaderBot_SetupScrollBar("Hunter");
	end 

	local dd = _G["HunterLeaderChatChannelDropDown"];
	if (LeaderBotIsAssistant == false) then
		HunterLeader_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	else
		UIDropDownMenu_EnableDropDown(dd);
	end 


end

-- ******************************************************************************
function HunterLeader_AnnounceButtonEnable(state)
	HUNLmsg("FUnction: HunterLeader_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["HunterLeaderAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["HunterLeaderAnnounce" .. i];
	end 

end

-- ******************************************************************************
function HunterLeader_AnnounceToChat(title, assignmentIndex)
	HUNLmsg("Function: HunterLeader_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	local chanType = HunterLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = HunterLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = HunterLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot -Hunter Leader: No chat channel selected");
		return;
	end

	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

	for assignmentKey, assignmentValue in next, HunterLeader_Assignments[assignmentIndex] do
		local playerName, target;
		local msg = nil;

		playerName = assignmentKey;
		targetName = assignmentValue;

		--skip players assigned to LeaderBotNone
		if (targetName ~= LeaderBotNone ) then 
			msg = "     " .. playerName .. ": " .. targetName .. " " ;

			--if (chanType == "RAID") then
			if (chanNumber == nil) then
				--Party, raid
				SendChatMessage(msg, chanType);
			else
				SendChatMessage(msg, chanType, chanLang, chanNumber);
			end

			--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
			--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
			--sync everything being announced.
			if (string.find(assignmentIndex, "MultiSelect") ) then 
				for key_TargetName, val_CheckMarkStatus in next, HunterLeader_Assignments[assignmentIndex][playerName] do
					LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Hunter", assignmentIndex, playerName, key_TargetName, val_CheckMarkStatus);
				end
			else 
				LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Hunter", assignmentIndex, playerName, HunterLeader_Assignments[assignmentIndex][playerName]);
			end 

		end 

	end

	if (chanNumber == nil) then
		--Party, raid
		SendChatMessage("-------------------------------------------", chanType);
	else
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
								
end

-- ******************************************************************************
function HunterLeader_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	HUNLmsg("HunterLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	HunterLeader_MsgChannel[LOCKCHANTYPE] = arg1;
	HunterLeader_MsgChannel[LOCKCHANLANG] = nil;
	HunterLeader_MsgChannel[LOCKCHANNUMB] = arg2;

	HunterLeader_AnnounceButtonEnable(1);
	
end

