--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
--]]

local version = "1.67";
local LB_Debug = 0;
LeaderBotCurrentDataModel = 1.38;
LeaderBotCurrentOptionsModel = 1.1;

local LeaderBotSmall = 0;
local LeadeBotSmallFont = "GameFontWhiteSmall";

local LeaderBotLarge = 1;
local LeaderBotColumnTitleFontLarge = "GameFontNormal";
local LeaderBotPlayerNameFontLarge = "GameFontNormal";
local LeaderBotDropDownLabelFontLarge = "GameFontNormalSmall";
local LeaderBotPlayerHeightLarge = 35;
local LeaderBotHeaderHeightLarge = 80;
local LeaderBotFooterHeightLarge = 40;
local LeaderBotBorderPaddingLarge = 2;
local LeaderBotInfoPaddingLarge = -25;

--all global variables so they can a) be saved and b) be seen by the other modules.
LeaderBotColumnTitleFont = LeaderBotColumnTitleFontLarge;
LeaderBotPlayerNameFont = LeaderBotPlayerNameFontLarge;
LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontLarge;	
LeaderBotplayerHeight = LeaderBotPlayerHeightLarge;
LeaderBotheaderHeight = LeaderBotHeaderHeightLarge;
LeaderBotfooterHeight = LeaderBotFooterHeightLarge;
LeaderBotborderPadding = LeaderBotBorderPaddingLarge;
LeaderBotInfoPadding = LeaderBotInfoPaddingLarge;

LeaderBotSenderName = "";
LeaderBotMinWidth = 550;

local COUNTERAMA = 0;

local MAXIMIZED = "yes";
local MINIMIZED = "no";

local LBLoopControl = 60;
local LBGetTeamRunning = 0;
local LBRosterChangedQueued = 0;

--globals for cross-file access, and or saved variables
LeaderBotPanels = {};	
LeaderBotNoticePanels = {};
LeaderBotNone = "-None-";
LeaderBotLegacyNone = "-none-";	--used only to account for lowercase none saved in WTF variables
LeaderBotScrollSlots = 10;

local leaderBotWidgetArray = {};
leaderBotWidgetArray["Warlock"] = {}; 
leaderBotWidgetArray["healer"] = {};
leaderBotWidgetArray["tank"] = {};
leaderBotWidgetArray["Mage"] = {};
leaderBotWidgetArray["Hunter"] = {};
leaderBotWidgetArray["crowdcontroller"] = {};

LeaderBotOptions["LockLeader"] = true;
LeaderBotOptions["HealerLeader"] = true;
LeaderBotOptions["TankLeader"] = true;
LeaderBotOptions["MageLeader"] = true;
LeaderBotOptions["HunterLeader"] = true;

LeaderBotIsAssistant = true; 

LeaderBotSingleSelect = false;
LeaderBotMultiSelect = true;

local showOfflineAlert = false;
local offlineList = {};
local offlineListDisplayNames = "";

local classicOrRetail;

local classCounts;
local teamMembers;
local throttled;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local needAssistNote;

local playerClass = {"Priest", "Paladin", "Shaman", "Warrior", "Druid", "Hunter", "Mage", "Rogue", "Warlock"};

local GroupTypes = 	{
						["healer"] =		{"Priest", "Paladin", "Shaman", "Druid"},
						["tank"]  =			{"Warrior", "Druid", "Paladin"},
						["crowdcontroller"] =	{"Druid", "Hunter", "Mage", "Priest", "Rogue", "Warlock"},
					};

--[[
-- example code
	for indexNumber, group in next, GroupTypes do
		print ("group: " .. indexNumber);
		for indexNumber2, class in next, group do 
			print("     index = " .. indexNumber2 .. ", class = " .. class);
		end 
	end
--]]

local alpha = { "a", "b", "c", "d", "e"};

local colorCode = "|cffff9900 ";
local greyColorCode = "|cff444444";
local whiteColorCode = "|cffffffff"; 

local column1Width = 150;


local shadowFormSpellID = 15473;
local bearFormSpellID = 19179;

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

local buttonBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2,	
	insets = 
	{
		left = 4,
		right = 4,
		top = 4,
		bottom = 4,
	},
};

local LeaderBotIncomingMsgPrefix = "!lb"
local LeaderBotOutgoingMsgPrefix = "(Leader Bot)";

 -- *****************************************************************************************************
function LBmsg(msg, override)
	if (LB_Debug == 1 or override) then
		ChatFrame1:AddMessage( colorCode .. msg);
	end
end

-- *******************************************************************************************************
-- Main code. We come straight here because there's no lua to send us to OnLoad()
function LeaderBotFrame_OnLoad()
	LBmsg("Function: LeaderBotFrame_OnLoad()");

	-- UI Commands
	SLASH_LeaderBot1 = "/lb";
	SLASH_LeaderBot2 = "/LB";

	SlashCmdList["LeaderBot"] =LeaderBot_SlashHandler;

	LL_VARNOTLOADED = true;
 
	-- custom tooltip for scanning item link info
	--CreateFrame("GameTooltip", "NLTooltip", UIparent, "GameTooltipTemplate");
	--NLTooltip:SetOwner(UIParent, "ANCHOR_NONE");

	ChatFrame1:AddMessage(colorCode .. "LeaderBot has been Loaded");
 
	classicOrRetail = "Classic";

	LeaderBotFrame:RegisterEvent("VARIABLES_LOADED");
	LeaderBotFrame:RegisterEvent("CHANNEL_UI_UPDATE");
	LeaderBotFrame:RegisterEvent("CHAT_MSG_WHISPER");
	
	LeaderBotFrame:RegisterEvent("CHAT_MSG_SYSTEM"); --arg 1 has "joins the group", "has joined the raid", "has left the raid"
	LeaderBotFrame:RegisterEvent("GROUP_ROSTER_UPDATE"); --fires 3x before join, 2x after join, 2x after leave. Fires when party member goes offline, but not raid member
	LeaderBotFrame:RegisterEvent("UNIT_CONNECTION"); --fires 1x for group, and again 1x for raid. Fires when joining a raid, as well as coming online.
	--LeaderBotFrame:RegisterEvent("UNIT_LEVEL"); --fires 1x after join group. fires 2x afer join raid (arg1 says party and raid)
	--LeaderBotFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");

	-- suppresses incoming whisper responses to limit spam
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER", function(self, event, msg, ...)								
														if (strfind(msg, LeaderBotIncomingMsgPrefix) ) then
															return true;
														end
		
													end);

	-- suppresses outgoing whisper responses to limit spam
	ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", function(self, event, msg, ...)

														if (strfind(msg, LeaderBotOutgoingMsgPrefix ) ) then
															return true;
														end 
													end);		

	--Now we wait for VARIABLES_LOADED to fire. See OnEvent() for next steps.

end

-- ****************************************************************************************************
function LeaderBot_InitMainFrame()
	LBmsg("Function: LeaderBot_InitMainFrame()");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove warlocks

	if (LeaderBotOptions["currentPanel"] == nil) then
		LeaderBotOptions["currentPanel"] = "Warlock";
	end

	LeaderBotFrame:SetSize(50,25);
	LeaderBotFrame:SetPoint("Center", UIParent, "Center", 0, 0 );

	LeaderBotFrame:EnableMouse(true);
	LeaderBotFrame:SetMovable(true);
	LeaderBotFrame:SetScript("OnMouseDown",function(self) self:StartMoving() end);
	LeaderBotFrame:SetScript("OnMouseUp",function(self)self:StopMovingOrSizing() end);
	
	LeaderBotFrame:SetBackdrop(mainBackdrop);
	LeaderBotFrame:SetBackdropColor(0.0, 0.0, 0.0, 1);
	LeaderBotFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

	LeaderBotFrameNoticeFrame = CreateFrame("Frame", "LeaderBotFrameNoticeFrame", LeaderBotFrame, BackdropTemplateMixin and "BackdropTemplate");
	LeaderBotFrameNoticeFrame:SetPoint("BottomRight", LeaderBotFrame, "TopLeft", 15, -15);
	LeaderBotFrameNoticeFrame:SetSize(20,20);
	LeaderBotFrameNoticeFrame:SetBackdrop(mainBackdrop);
	LeaderBotFrameNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
	LeaderBotFrameNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

	LeaderBotFrameNoticeIcon = LeaderBotFrameNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "ARTWORK", "QuestFont_Shadow_Huge");
	LeaderBotFrameNoticeIcon:SetPoint("Center", LeaderBotFrameNoticeFrame, "Center", -3, 0);
	LeaderBotFrameNoticeIcon:SetText(colorCode .. "!");
	
	LeaderBotFrameNoticeFrame:Hide();

	--this is pure magic, but takes the icon path and somehow sticks it into a text string. way easier than managing texture maps!!
	--|T and |t apparently signal the start and end of a texture defintion.
	--only the last four numbers seem to have significance. They specify the X range from start to finish and the Y range from start to finish
	--in this case, the icon we want occupies 196x to 256x and 64y to 128y.
	--in a 4x4 grid of icons, that means warlock is the last one on the middle row
	--see also https://github.com/Gethe/wow-ui-textures/blob/live/WorldStateFrame/ICONS-CLASSES.PNG
	local appLogo = LeaderBotFrame:CreateFontString("LeaderBotAppIcon", "ARTWORK", "QuestFont_Shadow_Huge");
	appLogo:SetPoint("TopLeft", LeaderBotFrame, "TopLeft", 5, -5);
	appLogo:SetText("|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:64:128|t"); 

	local maxButton = CreateFrame("Button", "LeaderBotMiniMaximizeButton", LeaderBotFrame, "OptionsButtonTemplate");
	maxButton:SetText("+");
	maxButton.tooltipText = "LeaderBot " .. classicOrRetail .. " by Quel.\nRemember kids: Quel rhymes with Fuel";
	maxButton:SetPoint("TopRight", LeaderBotFrame, "TopRight", -4, -4);
	maxButton:SetSize(20,20);

	maxButton:SetScript("OnClick",
									function()
											--maximize the selected panel, minimize us
											LeaderBotFrame:SetShown(false);
											LeaderBotFrameMax:SetShown(true);

											--show the last selected panel
											local showPanel;
											for panelName, panelFrame in next, LeaderBotPanels do 
												if (LeaderBotOptions["currentPanel"] == panelName) then
													--make sure this panel shows last for rezing
													showPanel = panelFrame;
												else
													panelFrame:Hide();
												end 
											end 
											showPanel:Show();

											LeaderBotOptions["LeaderBotShowMax"] = MAXIMIZED;
											LeaderBotFrameNoticeFrame:Hide();
									end);

	-- ********************************************************
	--Create a frame of class tab buttons that will be the anchor point the for class leader frames
	LeaderBotFrameMax = CreateFrame("Frame", "LeaderBotFrameMax", UIParent, BackdropTemplateMixin and "BackdropTemplate");
	LeaderBotFrameMax:SetSize(625, 25);
	LeaderBotFrameMax:SetPoint("Center", UIParent, "Center", 0, 0);

	LeaderBotFrameMax:SetBackdrop(mainBackdrop);
	LeaderBotFrameMax:SetBackdropColor(0.1, 0.1, 0.1, 0.85);
	LeaderBotFrameMax:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	LeaderBotFrameMax:EnableMouse(true);
	LeaderBotFrameMax:SetMovable(true);
	LeaderBotFrameMax:SetScript("OnMouseDown",function(self) self:StartMoving() end);
	LeaderBotFrameMax:SetScript("OnMouseUp",function(self)self:StopMovingOrSizing() end);
	LeaderBotFrameMax:SetScript("OnUpdate",LeaderBotFrame_OnUpdate);

	--local appLogo = LeaderBotFrameMax:CreateFontString("LeaderBotAppIcon", "ARTWORK", "QuestFont_Shadow_Huge");
	--appLogo:SetPoint("TopLeft", LeaderBotFrameMax, "TopLeft", 5, -5);
	--appLogo:SetText("|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:64:128|t"); 
								
	local appTitle = LeaderBotFrameMax:CreateFontString("LeaderBotAppTitle", "ARTWORK", "QuestFont_Shadow_Huge");
	appTitle:SetPoint("Top", LeaderBotFrameMax, "Top", 0, -2);
	appTitle:SetText("LeaderBot " .. classicOrRetail .. " " .. version );	

	needAssistNote = LeaderBotFrameMax:CreateFontString("LeaderBotNeedAssist", "OVERLAY", "GameFontWhiteSmall");
	needAssistNote:SetPoint("Top", appTitle, "Bottom", 0, -4);
	needAssistNote:SetText("You are not a raid leader or assistant. This page is READ ONLY.");

	-- MinMax button
	local minButton = CreateFrame("Button", "LeaderBotMinimizeButton", LeaderBotFrameMax, "OptionsButtonTemplate");
	minButton:SetText("-");
	minButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	minButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
	minButton.tooltipText = "LeaderBot " .. classicOrRetail .. " by Quel-Myzrael\nRemember kids: \"Quel\" rhymes with \"Fuel\" ";
	minButton:SetPoint("TopRight", LeaderBotFrameMax, "TopRight", -4, -2);
	minButton:SetSize(20,20);
	minButton:SetScript("OnClick",
									function()
											--maximize the selected panel, minimize us

											LeaderBotFrame:SetShown(true);
											LeaderBotFrameMax:SetShown(false);

											LeaderBotOptions["LeaderBotShowMax"] = MINIMIZED;

									end);


	--add the config button
	local configButton = CreateFrame("Button", "LeaderBotConfigButton", LeaderBotFrameMax, "OptionsButtonTemplate");
	configButton:SetNormalTexture("Interface\\Addons\\LeaderBot\\textures\\icon-config");
	configButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");
--	configButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-Button-Highlight.PNG");
	configButton:SetPoint("TopRight", LeaderBotFrameMax, "TopRight", -30, -2);
	configButton:SetSize(20,20);
	configButton.tooltipText = "Options";
	configButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	configButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
	configButton:SetScript("OnClick", function() 
										--if we run this once, it opens the interface menu. If we run it twice, it actually opens Leaderbot options. ><
										InterfaceOptionsFrame_OpenToCategory(LeaderBotOptionsPanel);
										InterfaceOptionsFrame_OpenToCategory(LeaderBotOptionsPanel);	--yes, run twice ><
									end);

	--add the sync request button
	local syncRequestButton = CreateFrame("Button", "LeaderBotSyncRequestButton", LeaderBotFrameMax, "OptionsButtonTemplate");
	syncRequestButton:SetText("<>");
	--syncRequestButton:SetNormalTexture("Interface\\Addons\\LeaderBot\\textures\\icon-config");
	--syncRequestButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");
	----syncRequestButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-Button-Highlight.PNG");
	syncRequestButton:SetPoint("Right", configButton, "Left", -5, 0);
	syncRequestButton:SetSize(20,20);
	syncRequestButton.tooltipText = "Request a Full Sync from the raid leader";
	syncRequestButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	syncRequestButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
	syncRequestButton:SetScript("OnClick", function() 
										LeaderBotSyncMaster_RequestFullRefresh();
									end);




		local warlockTabButton = CreateFrame("Button", "LeaderBotWarlockTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		warlockTabButton:SetNormalTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		warlockTabButton:GetNormalTexture():SetTexCoord(0.765, 1.0, 0.25, 0.5);
		warlockTabButton:SetDisabledTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		warlockTabButton:GetDisabledTexture():SetTexCoord(0.765, 1.0, 0.25, 0.5);		
		warlockTabButton.disabledtexture = warlockTabButton:GetDisabledTexture();	--get the actual texture object
		warlockTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out
		warlockTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		warlockTabButton:SetPoint("TopLeft", LeaderBotFrameMax, "TopLeft", 10, -4);
		warlockTabButton:SetSize(20, 20);
		warlockTabButton:SetBackdrop(buttonBackdrop);
		warlockTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		warlockTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		warlockTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		warlockTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		warlockTabButton.tooltipText = "Warlock Assignments";
		warlockTabButton.panelName = "Warlock";
		warlockTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);

		LeaderBotFrameWarlockNoticeFrame = CreateFrame("Frame", "LeaderBotFrameWarlockNoticeFrame", warlockTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameWarlockNoticeFrame:SetPoint("BottomRight", warlockTabButton, "TopLeft", 10, -10);
		LeaderBotFrameWarlockNoticeFrame:SetSize(15,15);
		LeaderBotFrameWarlockNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameWarlockNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameWarlockNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		local noticeFrameIcon = LeaderBotFrameWarlockNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameWarlockNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameWarlockNoticeFrame:Hide();

	if (not LeaderBotOptions["LockLeader"]) then 
		warlockTabButton:Disable();
	end


		local healerTabButton = CreateFrame("Button", "LeaderBotHealerTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		healerTabButton:SetNormalTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		healerTabButton:GetNormalTexture():SetTexCoord(0.5, 0.75, 0.25, 0.5);
		healerTabButton:SetDisabledTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		healerTabButton:GetDisabledTexture():SetTexCoord(0.5, 0.75, 0.25, 0.5);		
		healerTabButton.disabledtexture = healerTabButton:GetDisabledTexture();	--get the actual texture object
		healerTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out		
		healerTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		healerTabButton:SetPoint("Left", warlockTabButton, "Right", 7, 0);		
		healerTabButton:SetSize(20, 20);
		healerTabButton:SetBackdrop(buttonBackdrop);
		healerTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		healerTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		healerTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		healerTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		healerTabButton.tooltipText = "Healer Assignments";
		healerTabButton.panelName = "healer";
		healerTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);

		LeaderBotFrameHealerNoticeFrame = CreateFrame("Frame", "LeaderBotFrameHealerNoticeFrame", healerTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameHealerNoticeFrame:SetPoint("BottomRight", healerTabButton, "TopLeft", 10, -10);
		LeaderBotFrameHealerNoticeFrame:SetSize(15,15);
		LeaderBotFrameHealerNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameHealerNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameHealerNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		noticeFrameIcon = LeaderBotFrameHealerNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameHealerNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameHealerNoticeFrame:Hide();

	if (not LeaderBotOptions["HealerLeader"]) then 
		healerTabButton:Disable();
	end 


		local tankTabButton = CreateFrame("Button", "LeaderBotTankTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		tankTabButton:SetNormalTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		tankTabButton:GetNormalTexture():SetTexCoord(0.0, 0.25, 0.0, 0.25);
		tankTabButton:SetDisabledTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		tankTabButton:GetDisabledTexture():SetTexCoord(0.0, 0.25, 0.0, 0.25);		
		tankTabButton.disabledtexture = tankTabButton:GetDisabledTexture();	--get the actual texture object
		tankTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out			
		tankTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		tankTabButton:SetPoint("Left", healerTabButton, "Right", 7, 0);		
		tankTabButton:SetSize(20, 20);
		tankTabButton:SetBackdrop(buttonBackdrop);
		tankTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		tankTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		tankTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		tankTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		tankTabButton.tooltipText = "Tank Assignments";
		tankTabButton.panelName = "tank";
		tankTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);

		LeaderBotFrameTankNoticeFrame = CreateFrame("Frame", "LeaderBotFrameTankNoticeFrame", tankTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameTankNoticeFrame:SetPoint("BottomRight", tankTabButton, "TopLeft", 10, -10);
		LeaderBotFrameTankNoticeFrame:SetSize(15,15);
		LeaderBotFrameTankNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameTankNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameTankNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		noticeFrameIcon = LeaderBotFrameTankNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameTankNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameTankNoticeFrame:Hide();

	if (not LeaderBotOptions["TankLeader"]) then 			
		tankTabButton:Disable();
	end


		local mageTabButton = CreateFrame("Button", "LeaderBotMageTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		mageTabButton:SetNormalTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		mageTabButton:GetNormalTexture():SetTexCoord(0.25, 0.50, 0.0, 0.25);
		mageTabButton:SetDisabledTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		mageTabButton:GetDisabledTexture():SetTexCoord(0.25, 0.50, 0.0, 0.25);		
		mageTabButton.disabledtexture = mageTabButton:GetDisabledTexture();	--get the actual texture object
		mageTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out		
		mageTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		mageTabButton:SetPoint("Left", tankTabButton, "Right", 7, 0);	
		mageTabButton:SetSize(20, 20);
		mageTabButton:SetBackdrop(buttonBackdrop);
		mageTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		mageTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		mageTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		mageTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		mageTabButton.tooltipText = "Mage Assignments";
		mageTabButton.panelName = "Mage";
		mageTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);
	
		LeaderBotFrameMageNoticeFrame = CreateFrame("Frame", "LeaderBotFrameMageNoticeFrame", mageTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameMageNoticeFrame:SetPoint("BottomRight", mageTabButton, "TopLeft", 10, -10);
		LeaderBotFrameMageNoticeFrame:SetSize(15,15);
		LeaderBotFrameMageNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameMageNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameMageNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		noticeFrameIcon = LeaderBotFrameMageNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameMageNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameMageNoticeFrame:Hide();

	if (not LeaderBotOptions["MageLeader"]) then 
		mageTabButton:Disable();
	end 


		local hunterTabButton = CreateFrame("Button", "LeaderBotHunterTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		hunterTabButton:SetNormalTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		hunterTabButton:GetNormalTexture():SetTexCoord(0.0, 0.25, 0.25, 0.50);
		hunterTabButton:SetDisabledTexture("Interface\\WorldStateFrame\\ICONS-CLASSES");
		hunterTabButton:GetDisabledTexture():SetTexCoord(0.0, 0.25, 0.25, 0.50);		
		hunterTabButton.disabledtexture = hunterTabButton:GetDisabledTexture();	--get the actual texture object
		hunterTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out		
		hunterTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		hunterTabButton:SetPoint("Left", mageTabButton, "Right", 7, 0);
		hunterTabButton:SetSize(20, 20);
		hunterTabButton:SetBackdrop(buttonBackdrop);
		hunterTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		hunterTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		hunterTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		hunterTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		hunterTabButton.tooltipText = "Hunter Assignments";
		hunterTabButton.panelName = "Hunter";
		hunterTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);
	
		LeaderBotFrameHunterNoticeFrame = CreateFrame("Frame", "LeaderBotFrameHunterNoticeFrame", hunterTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameHunterNoticeFrame:SetPoint("BottomRight", hunterTabButton, "TopLeft", 10, -10);
		LeaderBotFrameHunterNoticeFrame:SetSize(15,15);
		LeaderBotFrameHunterNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameHunterNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameHunterNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		noticeFrameIcon = LeaderBotFrameHunterNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameHunterNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameHunterNoticeFrame:Hide();

	if (not LeaderBotOptions["HunterLeader"]) then 
		hunterTabButton:Disable();		
	end 

		local crowdTabButton = CreateFrame("Button", "LeaderBotCrowdTabButton", LeaderBotFrameMax, "OptionsButtonTemplate" and BackdropTemplateMixin and "BackdropTemplate");
		crowdTabButton:SetNormalTexture( GetSpellTexture(710) );
		crowdTabButton:SetDisabledTexture( GetSpellTexture(710) );
		crowdTabButton.disabledtexture = crowdTabButton:GetDisabledTexture();	--get the actual texture object
		crowdTabButton.disabledtexture:SetVertexColor(0.3, 0.2, 0.2);	--and grey it out	
		crowdTabButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp");	
		crowdTabButton:SetPoint("Left", hunterTabButton, "Right", 7, 0);
		crowdTabButton:SetSize(20, 20);
		crowdTabButton:SetBackdrop(buttonBackdrop);
		crowdTabButton:SetBackdropColor(0.4, 0.4, 0.4, 0.85);
		crowdTabButton:SetBackdropBorderColor(0.0, 1.0, 0.0, 1.0);
		crowdTabButton:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
		crowdTabButton:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);
		crowdTabButton.tooltipText = "Crowd Control Assignments";
		crowdTabButton.panelName = "crowdcontroller";
		crowdTabButton:SetScript("OnClick", LeaderBotFrame_OnClickPanelButton);
	
		LeaderBotFrameCrowdControllerNoticeFrame = CreateFrame("Frame", "LeaderBotFrameCrowdControllerNoticeFrame", hunterTabButton, BackdropTemplateMixin and "BackdropTemplate");
		LeaderBotFrameCrowdControllerNoticeFrame:SetPoint("BottomRight", crowdTabButton, "TopLeft", 10, -10);
		LeaderBotFrameCrowdControllerNoticeFrame:SetSize(15,15);
		LeaderBotFrameCrowdControllerNoticeFrame:SetBackdrop(labelBackdrop);
		LeaderBotFrameCrowdControllerNoticeFrame:SetBackdropColor(0.6, 0.1, 0.1, 1);
		LeaderBotFrameCrowdControllerNoticeFrame:SetBackdropBorderColor(0.1, 0.1, 0.1, 1);

		noticeFrameIcon = LeaderBotFrameCrowdControllerNoticeFrame:CreateFontString("LeaderBotFrameNoticeIcon", "OVERLAY", "QuestFont_Shadow_Huge");
		noticeFrameIcon:SetPoint("Center", LeaderBotFrameCrowdControllerNoticeFrame, "Center", -3, 0);
		noticeFrameIcon:SetText(colorCode .. "!");

		LeaderBotFrameCrowdControllerNoticeFrame:Hide();

	if (not LeaderBotOptions["CrowdController"]) then 
		crowdTabButton:Disable();		
	end 

	--tell LeaderBot that we're the warlock pane provider. In essence, we're basically passing a
	--pointer through this global array so other .lua files can properly talk to each other.

	throttled, teamMembers, classCounts = LeaderBotFrame_GetTeamMembers();

	if (LeaderBotOptions["LockLeader"]) then
		LeaderBotPanels["Warlock"] = LockLeaderFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["Warlock"]);
		LeaderBotNoticePanels["WarlockNotice"] = LeaderBotFrameWarlockNoticeFrame;
	end 

	if (LeaderBotOptions["TankLeader"]) then
		LeaderBotPanels["tank"] = TankLeaderFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["tank"] );
		LeaderBotNoticePanels["tankNotice"] = LeaderBotFrameTankNoticeFrame;
	end 

	if (LeaderBotOptions["HealerLeader"]) then
		LeaderBotPanels["healer"] = HealerLeaderFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["healer"]);
		LeaderBotNoticePanels["healerNotice"] = LeaderBotFrameHealerNoticeFrame;		
	end

	if (LeaderBotOptions["MageLeader"]) then
		LeaderBotPanels["Mage"] = MageLeaderFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["Mage"]);
		LeaderBotNoticePanels["MageNotice"] = LeaderBotFrameMageNoticeFrame;		
	end

	if (LeaderBotOptions["HunterLeader"]) then
		LeaderBotPanels["Hunter"] = HunterLeaderFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["Hunter"]);
		LeaderBotNoticePanels["HunterNotice"] = LeaderBotFrameHunterNoticeFrame;		
	end 

	if (LeaderBotOptions["CrowdController"]) then
		LeaderBotPanels["crowdcontroller"] = CrowdControllerFrame_OnLoad(LeaderBotFrameMax, teamMembers, classCounts["crowdcontroller"]);
		LeaderBotNoticePanels["crowdcontrollerNotice"] = LeaderBotFrameCrowdControllerNoticeFrame;		
	end 

	LeaderBotSenderName = UnitName("player");
end

-- ****************************************************
function LeaderBot_EnableFullSyncButton(enable)
	if (enable) then 
		LBmsg("Function: LeaderBot_EnableFullSyncButton(TRUE)");
	else 
		LBmsg("Function: LeaderBot_EnableFullSyncButton(FALSE)");
	end

	local button = _G["LeaderBotSyncRequestButton"];
	if (enable) then
		button:Enable()
	else
		button:Disable();
	end

end 

-- *******************************************************************************************
--determines which panel to show based on which class tab has been clicked. Each panel registers
--itself in the LeaderBotPanels{} array on creation. This onclick handler walks the array
--hiding unclicked panels and showing the clicked panel
function LeaderBotFrame_OnClickPanelButton(self)
	LBmsg("Function: LeaderBotFrame_OnClickPanelButton(" .. self.panelName .. ")" )

	--save the currently selected panel
	LeaderBotOptions["currentPanel"] = self.panelName;

	playerAssistStatus = UnitIsGroupAssistant("player");
	playerLeaderStatus = UnitIsGroupLeader("player");
	if (not playerAssistStatus and not playerLeaderStatus and LeaderBotOptions["noLead"]) then 
        StaticPopupDialogs["LEADERBOT_NOT_ASSISTANT"] = 
		{
			preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
			timeout = 7,
            whileDead = true,
            hideOnEscape = true,
        	text = "You are not the raid leader or an assistant. This page is read-only.\n\nYou can disable this popup in the options.",
                                                                                                    
            button1 = "OK",
            OnAccept = function()
                                                                                                        							
            			end,
			
        };

        StaticPopup_Show("LEADERBOT_NOT_ASSISTANT");

	end 


	--find the associated panelFrame in the registered panels
	for panelName, panelFrame in next, LeaderBotPanels do 

		if (LeaderBotOptions["currentPanel"] == panelName) then
			--this will resize the pane based on the current group roster,
			--then adjust LeaderBotFrameMax to the corerct width
			panelFrame:Show();

			local noticeFrame = LeaderBotNoticePanels[panelName .. "Notice"];
			noticeFrame:Hide();

			--scroll bar will be nil until some pane actually needs to scroll and creates it
			--via SetupScrolLBar, so check validity before attempting to hide it.
			if (LeaderBotScrollFrame) then
				LeaderBotScrollFrame:Hide() -- hide until proven that we need it
			end 

			if (classCounts[LeaderBotOptions["currentPanel"]] > LeaderBotScrollSlots) then 	
				LeaderBot_SetupScrollBar(LeaderBotOptions["currentPanel"]);
				LeaderBotScrollFrame:Show();
			end
		else
			panelFrame:Hide();
		end 
	end 

end

-- *******************************************************************************************
--this moves the tooltip for the header buttons to appear next to the button itself instead
--of the lower right corner of the screen
function LeaderBotFrame_OnEnterButton(self)
	GameTooltip:SetOwner(self);
	GameTooltip:SetText(self.tooltipText, nil, nil, nil, nil, true);
	GameTooltip:Show()
end

-- ********************************************************************************************
--closes tooltip when you mouse away from a header button
function LeaderBotFrame_OnLeaveButton(self)
	GameTooltip:Hide();
end


-- ******************************************************************************************************
function LeaderBotFrame_OnEvent(self, event, ...)

	local arg1 = select(1, ...);
	local arg2 = select(2, ...);

	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	LBmsg("Function: LeaderBot_OnEvent(" .. event .. ", " .. arg1 .. ")");

	-- *******************************************************************
	if (event == "VARIABLES_LOADED") then

		if (LeaderBotOptions["CrowdController"]) then 
			CrowdController_DataMigration();
		end 

		if (LeaderBotOptions["HunterLeader"]) then 
			HunterLeader_DataMigration();
		end 

		if (LeaderBotOptions["TankLeader"]) then 
			TankLeader_DataMigration();
		end

		if (LeaderBotOptions["MageLeader"]) then 
			MageLeader_DataMigration();
		end 

		if (LeaderBotOptions["HealerLeader"]) then 
			HealerLeader_DataMigration();
		end

		if (LeaderBotOptions["LockLeader"]) then 
			LockLeader_DataMigration();
		end 

		if (LeaderBotOptions["OptionsModelVersion"] ~= LeaderBotCurrentOptionsModel) then 
			LeaderBotOptions = {};

			StaticPopupDialogs["LEADERBOT_NEW_OPTIONS"] = {
				preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
				timeout = 0,
				whileDead = true,
				hideOnEscape = true,
				text = "New Leader Bot options are available. Options have been reset to defaults",
							
				button1 = "Options",
				OnAccept = function()
								--if we run this once, it opens the interface menu. If we run it twice, it actually opens Leaderbot options. ><
								InterfaceOptionsFrame_OpenToCategory(LeaderBotOptionsPanel);
								InterfaceOptionsFrame_OpenToCategory(LeaderBotOptionsPanel);
								LeaderBotOptions["OptionsModelVersion"] = LeaderBotCurrentOptionsModel;								
							end,

				button2 = "Cancel",
				OnCancel = function()
								LeaderBotOptions["OptionsModelVersion"] = LeaderBotCurrentOptionsModel;				
							end,				
				};

			StaticPopup_Show("LEADERBOT_NEW_OPTIONS");

		end 

		if (LeaderBotOptions["LeaderBotShowMax"] == nil) then
			LeaderBotOptions["LeaderBotShowMax"] = MAXIMIZED;
		end

		if (LeaderBotOptions["groupSize"] == nil) then 
			LeaderBotOptions["groupSize"] = GetNumGroupMembers();
		end 

		--initialize new options as of 3/8/20
		if (LeaderBotOptions["noLead"] == nil) then
			LeaderBotOptions["noLead"] = true;
		end 

		--new options 1/5/2020. If it's not present in existing options, then make it present.
		--always change the "if" to check the last added tab to ensure this check runs and picks up the new tab
		if (LeaderBotOptions["CrowdController"] == nil) then 
			LeaderBotOptions["LockLeader"] = true;		
			LeaderBotOptions["HealerLeader"] = true;
			LeaderBotOptions["TankLeader"] = true;
			LeaderBotOptions["MageLeader"] = true;
			LeaderBotOptions["HunterLeader"] = true;
			LeaderBotOptions["CrowdController"] = true;
		end 

		if (LeaderBotOptions["LockOfflinePopup"] == nil) then 
			LeaderBotOptions["LockOfflinePopup"] = true;
		end 

		if (LeaderBotOptions["LockSSPopup"] == nil) then
			LeaderBotOptions["LockSSPopup"] = true;
		end 

		if (LeaderBotOptions["LockTargetPopup"] == nil) then 
			LeaderBotOptions["LockTargetPopup"] = true;
		end 

		if (LeaderBotOptions["HealerLeftPopup"] == nil) then 
			LeaderBotOptions["HealerLeftPopup"] = true;
		end

		if(LeaderBotOptions["TankLeftPopup"] == nil) then 
			LeaderBotOptions["TankLeftPopup"] = true;
		end

		if (LeaderBotOptions["5rows"] == nil) then
			LeaderBotOptions["5rows"] = true;
		end 

		if (LeaderBotOptions["SyncMode"] == nil) then 
			LeaderBotOptions["SyncMode"] = true;
		end

		if (LeaderBotOptions["SyncModeDebug"] == nil) then 
			LeaderBotOptions["SyncModeDebug"] = false;
		elseif (LeaderBotOptions["SyncModeDebug"] == true) then 
			LeaderBotSyncMaster_SetDebug(true);
		end 

		LeaderBotOptionsPanel_Init();
		LeaderBot_InitMainFrame();
		LeaderBot_RosterChanged();

		LeaderBotSyncMaster_OnInit();
		LeaderBotHealthManager_OnInit();

		LeaderBot_EnableFullSyncButton(LeaderBotOptions["SyncMode"]);

		if (LeaderBotOptions["LeaderBotShowMax"] == MINIMIZED) then
			LeaderBotFrame:Show();
			LeaderBotFrameMax:Hide();
		else 
			LeaderBotFrame:Hide();
			LeaderBotFrameMax:Show();

			--show the last selected panel
			if (LeaderBotOptions["currentPanel"] == nil) then
				LeaderBotOptions["currentPanel"] = "Warlock";
			end

			LBmsg("     LeaderBotOptions[currentPanel] = " .. LeaderBotOptions["currentPanel"] );
			for panelName, panelFrame in next, LeaderBotPanels do 
				LBmsg("          compare panel " .. panelName .. " to savedPanel " .. LeaderBotOptions["currentPanel"]);

				if (LeaderBotOptions["currentPanel"] == panelName) then
					LBmsg("              Showing panel " .. panelName);
					panelFrame:Show();

					--reset LeaderBotFrameMax's width to match the currently selected panel
					local width = panelFrame:GetWidth();
					LeaderBotFrameMax:SetWidth(width);

				else
					LBmsg("               Hiding Panel " .. panelName);
					panelFrame:Hide();
				end 
			end 
 
			if (LeaderBotOptions["5rows"] == true) then 
				LeaderBotScrollSlots = 5;
			else 
				LeaderBotScrollSlots = 10;
			end 
			
		end


	-- *********************************************************************
	elseif (event == "CHANNEL_UI_UPDATE") then
		channelCount, channelList = LeaderBotFrame_GetChatChannels();




	-- *********************************************************************
	elseif( event == "CHAT_MSG_WHISPER") then 
		if (string.match(arg1, LeaderBotIncomingMsgPrefix) ) then 

			local msg, senderRealmName, languageName, channelName, senderName, specialFlags, zoneChannelID, channelIndex, 
				channelBaseName, unused, lineID, senderGUID, bnSenderID, isMobile, isSubtitle, hideSender, suppressRaidIcons = select(1, ...);

			LeaderBot_AssignmentRequestWhisper(msg, senderRealmName, senderName, suppressRaidIcons );
		end 

	
	-- *******************************************************************
	elseif( event == "CHAT_MSG_ADDON") then 




	-- *******************************************************************
	-- other GROUP_ROSTER_CHANGED fires way too many times. 3x when an invite goes out,
	-- 2x when the party joins. plus when loot mode changes and many other conditions. The
	-- stacked calls to onEvent were stepping on each other and causing thread lock.

	-- CHAT_MSG_SYSTEM, though, gives a reliable, 1-occurence-per-event for players
	-- joining and leaving teams. The only annoyance is that the message tense varies so 
	-- much.

	-- TODO: It's not reliable on Group Disbanded. The message fires well before the last
	-- "leaver" is gone. need to delay RosterChanged() call when we see "disbanded"

	-- TODO: This doesn't detect when you join someone else's group. The only chat msg
	-- says "has invited you to join a group" but nothing fires after you actually join

	elseif (event == "CHAT_MSG_SYSTEM") then 

			if (
				string.match(arg1, "joins the party")
				or string.match(arg1, "leaves the party")
				or string.match(arg1, "has joined the raid group")
				or string.match(arg1, "has left the raid group") 
				or string.match(arg1, "Party converted to Raid")
				or string.match(arg1, "Raid converted to Party")

				or string.match(arg1, "You have joined a raid group")

				or string.match(arg1, "Your group has been disbanded")			--fires SOMETIMES before the unit is removed from the group ><
				or string.match(arg1, "You leave the group") 					--fires SOMETIMES before the unit is removed from the group ><
				or string.match(arg1, "You have been removed from the group")	--fires SOMETIMES before the unit is removed from the group ><

			) then

				--Since some of these events SOMETIMES fire before and SOMETIMES after the group membership change, 
				--we need to delay a bit before we call RosterChanged(). We set this flag and wait for onUpdate to pick it up. 
				--Because of OnUpdate()'s 1-second throttle, this seems to inject enough delay that the group membership
				--is finally updated before we call RosterChanged()
				LeaderBot_QueueRosterUpdate();
				
			end
	

	--needed for Group Invite Accepted, Raid Member coming online, group member disconnecting
	--needed for when players go offline (1 fire BUT before the player goes offline >< ) or comes online (3 fires, 2 before they show online. 3rd time they appear as online ><)
	--Otherwise, it fires so much that's almost spam. CHAT_MSG_SYSTEM is more usable for basic join/leave actions
	elseif (event == "GROUP_ROSTER_UPDATE") then

		if (IsInGroup() or IsInRaid() ) then 			
				--Since some of these events SOMETIMES fire before and SOMETIMES after the group membership change, 
				--we need to delay a bit before we call RosterChanged(). We set this flag and wait for onUpdate to pick it up. 
				--Because of OnUpdate()'s 1-second throttle, this seems to inject enough delay that the group membership
				--is finally updated before we call RosterChanged()
				LeaderBot_QueueRosterUpdate();	
		end
	

	--fires when raid member goes online/offline (twice, once for groupID once for raidID)
	--fires when player joins group (twice, once for groupID once for raidID)
	elseif (event == "UNIT_CONNECTION") then 
		local unitID = arg1;
		local isConnected = arg2;
		LeaderBot_QueueRosterUpdate();
	end

end;

-- **************************************************************************
--used primarily by SyncMaster
function LeaderBot_QueueRosterUpdate()
	LBmsg("Function: LeaderBot_QueueRosterUpdate()")
	if (LBRosterChangedQueued > 1) then
		--already queued.
		return;
	end 

	LBRosterChangedQueued = LBRosterChangedQueued + 1;
end 

-- **************************************************************************
function LeaderBot_RosterChanged()
	LBmsg("Function: LeaderBot_RosterChanged()");

	throttled, newTeamMembers, classCounts = LeaderBotFrame_GetTeamMembers();			

	if (throttled == true) then 
		--our call was throttled, probably during raid formation or exit. Lots of events fire and many threads
		--stack. So we need to throttle the requests. Tell OnUpdate() that we want to try again when possible.
		LeaderBot_QueueRosterUpdate();

	else
		--our call was successfully processed.
		

		--it's extremely fucking difficult to get WoW to tell us if we've simply joined or left a group. So
		--we have to concot this bullshit to try and track our state and detect when teh number of group members
		--changes. However, WoW doesn't reliably return the number of group members at a consistent time, either.
		--So logging in while in a group, reloading UI while in a group, and actually, you know, joining a group
		--share the same footprint. ><

		--as it stands, this will fire 
			--1. once on VARIABLES_LOADED during login. This is because I have no idea how long you've been offline
			--   and you need to resync.
			--2. Triggered via the queue when events fire that suggest you have joined a group.
			--3. Triggered via the queue when events fire that suggest you have left a group.

		local currentGroupMemberCount = GetNumGroupMembers();
		
		if (LeaderBotOptions["groupSize"] == 0 and currentGroupMemberCount > 0) then
			--it appears that we joined a group.
			--print("Leader Bot believes you have joined a group");
			LeaderBotSyncMaster_RequestFullRefresh();
		elseif (LeaderBotOptions["groupSize"] > 0 and currentGroupMemberCount == 0) then
			--it appears we have left a group. Nothing to do here except debugging print()
			--print ("Leader Bot believes you have left a group");
		end 
		LeaderBotOptions["groupSize"] = currentGroupMemberCount;


		--only allow raid leaders, group leaders, assistants or solo players to make changes
		if (UnitIsGroupAssistant("player") == true 
				or UnitIsGroupLeader("player") == true
				or (not IsInGroup() and not IsInRaid() )
			) then 
				needAssistNote:Hide();
				LeaderBotFrameMax:SetHeight(25);
				LeaderBotIsAssistant = true;
		else
				needAssistNote:Show();
				LeaderBotFrameMax:SetHeight(40);							
				LeaderBotIsAssistant = false;
		end 

		--showOfflineAlert will be updated when (re)adding Column1, as appropriate
		showOfflineAlert = false;	

		--tell each module to update their rosters
		if (LeaderBotOptions["LockLeader"]) then 
			LockLeader_RosterChanged(newTeamMembers, classCounts["Warlock"]);
		end 

		if (LeaderBotOptions["HealerLeader"]) then 
			HealerLeader_RosterChanged(newTeamMembers, classCounts["healer"]);
		end 

		if (LeaderBotOptions["TankLeader"]) then
			TankLeader_RosterChanged(newTeamMembers, classCounts["tank"]);
		end 

		if (LeaderBotOptions["MageLeader"]) then
			MageLeader_RosterChanged(newTeamMembers, classCounts["Mage"]);
		end 

		if (LeaderBotOptions["HunterLeader"]) then
			HunterLeader_RosterChanged(newTeamMembers, classCounts["Hunter"]);
		end

		if (LeaderBotOptions["CrowdController"]) then
			CrowdController_RosterChanged(newTeamMembers, classCounts["crowdcontroller"])
		end 

		if (showOfflineAlert) then
			showOfflineAlert = false;
			

			local msg = colorCode .. "LEADER BOT WARNING:\n\n" .. whiteColorCode;
			offlineListDisplayNames = "";

			for name, string in next, offlineList do
				if (string ~= "alreadyDisplayed") then 
					msg = msg .. string;
					offlineList[name] = "alreadyDisplayed";
					offlineListDisplayNames = offlineListDisplayNames .. " " .. name;
				end 
			end 

			if (LeaderBotIsAssistant) then 
				StaticPopupDialogs["LEADERBOT_OFFLINE_ALERT"] = {
							preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
							timeout = 10,
							whileDead = true,
							hideOnEscape = true,
							text = msg,
							
							button1 = "Panic!",
							OnAccept = function()
											SendChatMessage("Leader Bot:" .. offlineListDisplayNames .. " disconnected! WE'RE ALL GOING TO DIE!!", "RAID");
										end,

							button2 = "Nobody Cares",
							OnCancel = function()
											
										end,				
							};
			else 
				StaticPopupDialogs["LEADERBOT_OFFLINE_ALERT"] = {
							preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
							timeout = 10,
							whileDead = true,
							hideOnEscape = true,
							text = msg,
							
							button1 = "OK",
							OnAccept = function()

										end,
				
							};
			end 

			StaticPopup_Show("LEADERBOT_OFFLINE_ALERT");

		end 

	end 

end

-- **********************************************************************************
function LeaderBotFrame_OnUpdate(self, elapsed)
	--ChatFrame1:AddMessage("----------------------------");
	--ChatFrame1:AddMessage("LeaderBotFrame_OnUpdate(" .. elapsed .. ")" );

	--DO NOT SHORTEN THIS THROTTLE INTERVAL. Several chat messages regarding players coming/going fire
	--well before the group roster is updated. We have to delay a full 2 seconds to ensure rosters
	--are fully updated before calling RosterChanged().
	if (LBLoopControl <  2.5) then
		LBLoopControl = LBLoopControl + elapsed;
		return;
	end
	LBLoopControl = 0;

	--update Lock Leader's Soulstone display
	if (LeaderBotOptions["LockLeader"]) then 
		LockLeaderFrame_UpdateSoulStoneIcon(elapsed);
	end 

	--safety check to prevent getting stuck in a loop.
	if (LBRosterChangedQueued > 10) then
		-- this is too many queued requests. clear the queue but don't attempt to call RosterChanged().
		print(colorCode .. "     LeaderBot:OnUpdate(): **WARNING** too many attempts (" .. LBRosterChangedQueued .. ") to update roster. Aborting update loop." );
		LBRosterChangedQueued = 0;

	elseif (LBRosterChangedQueued > 0) then
		--we have pending, stacked requests to execute RosterChanged, which we have throttled.
		LBRosterChangedQueued = 0; 
		LeaderBot_RosterChanged();
	end 

end 


-- ******************************************************************************
--The first call to GetTeamMembers gets to run, and sets the LBGetTeamRunning semaphore to true. Anyone 
--that calls on top of this sets LBNeedsGetTeam, which will be picked up by OnUpdate(). This should throttle the
--calls to GetTeamMembers when lots of people are joining or leaving a raid.
function LeaderBotFrame_GetTeamMembers()
	LBmsg("Function: LeaderBotFrame_GetTeamMembers()");

	local GetTeamNotThrottled = false;
	local GetTeamThrottled = true;

	if (LBGetTeamRunning > 10) then
			print(colorCode .. "     LeaderBot:GetTeamMembers(): **WARNING** too many stacked calls (" .. LBGetTeamRunning .. ") to GetTeamMembers(). Canceling entire queue." );
			LBGetTeamRunning = 0;

	elseif (LBGetTeamRunning > 0) then
		--don't stack execution. schedule a retry to be picked up OnUpdate()
		LBGetTeamRunning = LBGetTeamRunning + 1;
		return GetTeamThrottled, nil, nil;
	end 

	LBGetTeamRunning = 1;


	local teamMembers = {};
	teamMembers["Hunter"] = {};
	teamMembers["Mage"]= {};
	teamMembers["Rogue"] = {};
	teamMembers["Paladin"] = {};
	teamMembers["Priest"] = {};
	teamMembers["Shaman"] = {};
	teamMembers["Warlock"] = {};
	teamMembers["Warrior"] = {};
	teamMembers["Druid"] = {};

	local teamCounts = {};
	teamCounts["Hunter"] = 0;
	teamCounts["Mage"]= 0;
	teamCounts["Rogue"] = 0;
	teamCounts["Paladin"] = 0;
	teamCounts["Priest"] = 0;
	teamCounts["Shaman"] = 0;
	teamCounts["Warlock"] = 0;
	teamCounts["Warrior"] = 0;
	teamCounts["Druid"] = 0;

	-- **********************************************************************
	--raids are pretty easy to walk through all the group members.
	if ( IsInRaid() ) then
		LBmsg("          We are in a raid");

		for i = 1, MAX_RAID_MEMBERS, 1 do
			local name, rank, subgroup, level, class, fileName, zone, online, 
				isDead, role, isML  = GetRaidRosterInfo(i);

			local  debugOnline = "offline";
			if (online) then
				debugOnline = "online";
			end

--TODO: (SOMETIMES) when logging in, offline raid members always return nil. But after a reloadUI, they appear in the raid but correctly as OFFLINE.
--so, what the hell?? The events that fire on login must be before the group is filled out. ><

			if (name) then

				--print(colorCode .. "     raidname" .. i .. ": " .. name .. " (" .. debugOnline .. ")" );
				teamMembers[class][name] = {};
				teamMembers[class][name]["level"] = level;
				teamMembers[class][name]["isdead"] = isDead;
				teamMembers[class][name]["role"] = role;
				teamMembers[class][name]["online"] = debugOnline;

				teamCounts[class] = teamCounts[class] + 1;
			else
				--people's position in the raid will change as you move them around groups, but so far, there
				--are never any nils in the middle of the raid. So when we hit nil, we're done.
				break;
			end
		end
	
	-- **********************************************************************
	--stupid fucking blizzard. Getting raid info gets all raid members, but getting party info excludes
	--the player. ><
	elseif (IsInGroup() ) then
		LBmsg("          We are in a group");

		--look up the player individually
		local class = UnitClass("player");
		local name = UnitName("player");

		teamMembers[class][name] = {};
		teamMembers[class][name]["level"] = UnitLevel("player");
		teamMembers[class][name]["isdead"] = UnitIsDead("player");
		teamMembers[class][name]["role"] = nil;
		teamMembers[class][name]["online"] = "online";

		teamCounts[class] = teamCounts[class] + 1;

		--now look up the party members. This is so fucking stupid. We can get the total number of members, but can only query
		--the members who aren't us.
		local numberOfMembers = GetNumGroupMembers();


		for i = 1, numberOfMembers -1  do
			local class = UnitClass("party" .. i);
			local name = UnitName("party" .. i);
			local level = UnitLevel("party" .. i);
			local isDead = UnitIsDead("party" .. i);
			local isConnected = UnitIsConnected("party" .. i);

			local  debugOnline = "offline";
			if (isConnected) then
				debugOnline = "online";
			end


			--it is technically impossible for name and class to be nil, since we already *know* there's a group member at that ID.
			--and yet, since the 12/17/19 maintenance, unitName and UnitClass sometimes return nil. So here's a defensive fix to at
			--least prevent a crash. ><
			if (name ~= nil and class ~= nil and level ~= 0) then 

				teamMembers[class][name] = {};
				teamMembers[class][name]["level"] = level;
				teamMembers[class][name]["isdead"] = isDead;
				teamMembers[class][name]["role"] = nil;
				teamMembers[class][name]["online"] = debugOnline;

				teamCounts[class] = teamCounts[class] + 1;

			else
				--Sigh. >< Either UntiClass or UnitName returned nil for a player that Blizzard has already told us *is* 
				--in this group. Make the name "unknown" and schedule a retry in OnUpdate(). Now I know why I've seen the
				--Blizzard UI add "Unknown" players to groups sometimes.

				local tempClass = "Warlock";
				local tempName = "*Unknown*";
				local tempLevel = -1;

				if (class ~= nil) then 
					tempClass = class;
				end

				if (name ~= nil) then
					tempName = name;
				end

				if (level ~= 0) then
					tempLevel = level;
				end 

				teamMembers[tempClass][tempName] = {};
				teamMembers[tempClass][tempName]["level"] = tempLevel;
				teamMembers[tempClass][tempName]["isdead"] = isDead;
				teamMembers[tempClass][tempName]["role"] = nil; 

				LBmsg("Leaderbot.GetTeamMembers(): **WARNING** Blizzard failed to respond to player lookup, the member list is probably inaccurate. Performing /reload after the raid is formed should fix this");

				LeaderBot_QueueRosterUpdate();

			end

		end
 
	-- **********************************************************************
	--probably don't need a case where I'm operating alone, except for initial debugging
	else
		--it's just me, all alone
		LBmsg("          We are solo");

		--look up the player individually
		local class = UnitClass("player");
		local name = UnitName("player");

		teamMembers[class][name] = {};
		teamMembers[class][name]["level"] = UnitLevel("player");
		teamMembers[class][name]["isdead"] = UnitIsDead("player");
		teamMembers[class][name]["role"] = nil;

		teamCounts[class] = teamCounts[class] + 1;
--[[
		if (toggleBot == 1) then
			--we've been here before, don't add anyone to simulate the fakeLock has left the group.
			toggleBot = 0
		else
			--add a fakeLock to simulate a lock joining.
			teamMembers["Warlock"]["fakeLock"] = {};
			teamMembers["Warlock"]["fakeLock"]["level"] = math.random(58,60);
			teamMembers["Warlock"]["fakeLock"]["isDead"] = false;
			teamMembers["Warlock"]["fakeLock"]["role"] = nil;

			teamCounts["Warlock"] = teamCounts["Warlock"] + 1;
			toggleBot = 1;
		end
--]]

--[[
		**** WARNING: Based on changes made 1/5/2020, using this test code to create a fake raid will cause Leaderbot
		**** to call Blizzard UnitXxx() functions with invalid player names. This, in turn, locks WoW up!

		--populate a fake raid team for testing purposes
		for i=2, 20 do
			randNumber = math.random(1,9);
			randomClass = playerClass[randNumber];
			randomName = randomClass .. i;
			LBmsg("          Random name: " .. randomName .. "(" .. randomClass .. ")" );

			teamMembers[randomClass][randomName] = {};
			teamMembers[randomClass][randomName]["level"] = math.random(58,60);
			teamMembers[randomClass][randomName]["isDead"] = false;
			teamMembers[randomClass][randomName]["role"] = nil;

			teamCounts[randomClass] = teamCounts[randomClass] + 1;
		end
--]]

	end

--[[
	--debugging code
print(colorCode .. "+++++++++++++++++++++++++");
local debugCount = LeaderBotFrame_CountGroup("tank", teamMembers, teamCounts);
for class in next, teamMembers do

	print(colorCode .. "class: " .. class .. ", count: " .. teamCounts[class] .. ", debugTankCount: " .. debugCount);
	for name, data in next, teamMembers[class] do
		print(colorCode .. "     " .. name .. "(" .. class .. ")" );
	end 
end 
print(colorCode .. "----------------------------");
--]]

	LBGetTeamRunning = 0;

	teamCounts["healer"] = LeaderBotFrame_CountGroup("healer", newTeamMembers, teamCounts);
	teamCounts["tank"] = LeaderBotFrame_CountGroup("tank", newTeamMembers, teamCounts);
	teamCounts["crowdcontroller"] = LeaderBotFrame_CountGroup("crowdcontroller", teamMembers, teamCounts);

	return GetTeamNotThrottled, teamMembers, teamCounts;

end
 

-- ***************************************************************
function LeaderBot_ToggleBot()

		if (toggleBot == 1) then
			--we've been here before, don't add anyone to simulate the fakeLock has left the group.
			toggleBot = 0
		else
			--add a fakeLock to simulate a lock joining.
			teamMembers["Warlock"]["fakeLock"] = {};
			teamMembers["Warlock"]["fakeLock"]["level"] = math.random(58,60);
			teamMembers["Warlock"]["fakeLock"]["isDead"] = false;
			teamMembers["Warlock"]["fakeLock"]["role"] = nil;
			toggleBot = 1;
		end
end 

 -- ****************************************************************************
 function LeaderBotFrame_CountClassMembers(teamMembers, class)
	LBmsg("Function: LeaderBotFrame_CountClassMembers(" .. class .. ")" );

	local count = 0;
	for name, data in next, teamMembers[class] do
		count = count + 1;
	end

	return count;

 end

-- *************************************************************************************
function LeaderBotFrame_CountGroup(groupType, teamMembers, teamCounts)
	LBmsg("Function: LeaderBotFrame_CountGroup(" .. groupType .. ")");
	local groupCount = 0;

	--walk each class defined in the group
	for indexNumber, class in next, GroupTypes[groupType] do
--		groupCount = groupCount + LeaderBotFrame_CountClassMembers(teamMembers, class);
		groupCount = groupCount + teamCounts[class];
	end

	return groupCount;
end


-- ********************************************************************************************
function LeaderBotFrame_IsMemberOfGroup(class, groupType)
	LBmsg("Funtion: LeaderBotFrame_IsMemberOfGroup(" .. class ..", " .. groupType .. ")");

	for indexNumber, memberClass in next, GroupTypes[groupType] do
		if (memberClass == class) then
			return true;
		end 
	end

	return false;

end

-- ***********************************************************************************************
function LeaderBotFrame_IsMemberOfTeam(playerName, teamMembers)
	LBmsg("Function: LeaderBotFrame_IsMemberOfTeam(" .. playerName .. ")" );
	
	for class in next, teamMembers do
		for name in next, teamMembers[class] do
			if (name == playerName) then 
				return true;
			end;
		end;
	end;

	return false;
end

--******************************************************************************************************
-- Command line parser
function LeaderBot_SlashHandler(arg1)
	LBmsg("Function: LeaderBot_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. "LeaderBot by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. "Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (LB_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. "LeaderBot debug messages disabled");
			LB_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. "LeaderBot debug messages enabled");
			LB_Debug = 1;
		end
	end

end

--******************************************************************************************************
--Takes a friendly name like {star} and converts it to the texture map string
function LeaderBot_GetTextForRaidIconName(raidIconName)
	LBmsg("Function: LeaderBot_GetTextForRaidIconName(" .. raidIconName .. ")" );
	local iconMap = {
						["{star}"] = 1,
						["{rt1}"] =1,

						["{circle}"] = 2,
						["{rt2}"] = 2,

						["{diamond}"] = 3,
						["{rt3}"] = 3,

						["{triangle}"] = 4,
						["{rt4}"] = 4,

						["{moon}"] = 5,
						["{rt5}"] = 5,

						["{square}"] = 6,
						["{rt6}"] = 6,

						["{cross}"] = 7,
						["{rt7}"] = 7,

						["{skull}"] = 8,
						["{rt8}"] = 8
					};

	--check for old data format (no longer used, should only happen on my test alts)
	if (iconMap[raidIconName] == nil) then
		print(colorCode .. "Leaderbot: unexpected raid icon name: " .. raidIconName .. "." );
		raidIconName = LeaderBotNone;
	end

	local textureString = "";
	
	--check for old data format (no longer used, should only happen on my test alts)
	if (raidIconName ~= LeaderBotNone and raidIconName ~= "-none-") then 				
		textureString = LeaderBot_GetTextForRaidIconNumber( iconMap[raidIconName]);		
	else
		textureString = raidIconName;
	end
	
	return (textureString);
end

--******************************************************************************************************
function LeaderBot_GetTextForClass(class)
	LBmsg("Function: LeaderBot_GetTextForClass(" .. class .. ")" );
	
	--only the last 4 characters are relevant, specifying the top left x,y and the bottom right x,y inside the texture map
	local classIcons =
	{
		["Warrior"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:0:64|t",
		["Mage"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:0:64|t",
		["Rogue"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:0:64|t",
		["Druid"]	= 	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:0:64|t",
		["Hunter"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:64:128|t",
		["Shaman"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:64:128|t",
		["Priest"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:64:128|t",
		["Warlock"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:64:128|t",
		["Paladin"]	=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:128:196|t",
	 }

	 return (classIcons[class] );

end

--******************************************************************************************************
-- This function takes texture maps for raid icons, applies the "tcoord" magic and converts them to text strings
function LeaderBot_GetTextForRaidIconNumber(raidIconNumber)
	LBmsg("Function: LeaderBot_GetTextForRaidIconNumber(" .. raidIconNumber .. ")" );

	--no idea what this syntax means, it's just the way it is. It apparently signals the start of a texture map being passed into a string
	local prefix = "|T"	

	local body =  "Interface\\TargetingFrame\\UI-RaidTargetingIcon_" .. raidIconNumber

	--the :0:0:0:0:256:256: never seems to change. The next two coords are the x range (0 to 256 here), and then the y range (0 to 256)
	--if the texture map has multiple textures, like teh class icons, each icon is generally 64 x 64. These dedicated raid icon files use the full 256x256 
	--for each. the |t apparently signals the end of texture definition
	local suffix = ":0:0:0:0:256:256:0:256:0:256|t";


	local returnText = prefix .. body .. suffix;
	HLmsg(returnText);
	return returnText;

end

-- ******************************************************************************
function LeaderBot_GetTextForSpellID(spellID)
	LBmsg("Function: LeaderBot_GetTextForSpellID(" .. spellID .. ")" );

	--no idea what this syntax means, it's just the way it is. It apparently signals the start of a texture map being passed into a string
	local prefix = "|T"	

	local body = GetSpellTexture(spellID);

	--the :0:0:0:0:256:256: never seems to change. The next two coords are the x range (0 to 256 here), and then the y range (0 to 256)
	--if the texture map has multiple textures, like teh class icons, each icon is generally 64 x 64. These dedicated raid icon files use the full 256x256 
	--for each. the |t apparently signals the end of texture definition
	local suffix = ":0:0:0:0:256:256:0:256:0:256|t";

	--can't make this return variable a local, or the receiving function just gets "nil". Something unique to texture maps??
	LBreturnText = prefix  .. body .. suffix;
	return LBreturnText;
end

-- ******************************************************************************
--when calling for a group, pass in groupType using lower case group name: healer, tank
--when calling for an individual class, pass in groupType using just the capitalized classname: Mage, Warlock
function LeaderBotFrame_AddColumn1(teamMembers, groupCount, groupType, panelFrame, titleText )
	LBmsg("Function: LeaderBotFrame_AddColumn1(" .. groupType .. ")"  );

	local columnBackdrop = 
	{
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true,		-- required to make Tooltip border work
		tileSize =16,	-- required to make Tooltip border work
		edgeSize = 16	
	};
		
	LeaderBotColumn1Frame = _G[groupType .. "LeaderColumn1"];
	if (nil == LeaderBotColumn1Frame) then
		LeaderBotColumn1Frame = CreateFrame("Frame", groupType .. "LeaderColumn1", panelFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	LeaderBotColumn1Frame:SetBackdrop(columnBackdrop);
	LeaderBotColumn1Frame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	LeaderBotColumn1Frame:SetPoint("TopLeft", panelFrame, "TopLeft", 2, -LeaderBotheaderHeight);

	local col1Title = _G[groupType .. "LeaderColumn1Title"];
	if (nil == col1Title) then
		col1Title = LeaderBotColumn1Frame:CreateFontString(groupType .. "LeaderColumn1Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	col1Title:SetPoint("Top", LeaderBotColumn1Frame, "Top", 0, 10);
	col1Title:SetText(groupCount .. " " .. titleText .. ":" );

	local noGroupTitle = _G[groupType .. "LeaderNoMembersTitle"];
	if (nil == noGroupTitle) then
		noGroupTitle = panelFrame:CreateFontString(groupType .. "LeaderNoMembersTitle", "ARTWORK", "GameFontNormal");
		noGroupTitle:SetPoint("Top", panelFrame, "Top", 0, -60);
		noGroupTitle:SetText("No members found");
	end

	if (groupCount > 0) then 
		noGroupTitle:Hide();
	else
		noGroupTitle:Show();
	end

	LeaderBotFrame_AdjustColumnSize(LeaderBotColumn1Frame, groupCount, column1Width);		
	
	local addIndex = 0;

	if (groupType == "healer" or groupType == "tank" or groupType == "crowdcontroller") then 
		--for healers and tanks, groupType contains a list of classes
		for class, classMember in next, teamMembers do
	
			if (LeaderBotFrame_IsMemberOfGroup(class, groupType) ) then 
				--each time we call this, we pass the index where we want to begin adding names, and the function returns
				--the index of the last slot it filled.
				addIndex = LeaderBotFrame_AddNamesToColumn1(LeaderBotColumn1Frame, teamMembers[class], groupType, addIndex);
			end
		end
	else

		--for mages, hunters, warlocks groupType *is* the class.
		addIndex = LeaderBotFrame_AddNamesToColumn1(LeaderBotColumn1Frame, teamMembers[groupType], groupType, addIndex);
	end

	LBmsg("     returning column frame");
	return LeaderBotColumn1Frame;
end

-- ******************************************************************************
--adjust the height as healers join or leave
function LeaderBotFrame_AdjustColumnSize(columnFrame, groupCount, requiredWidth)
	LBmsg("Function: LeaderBotFrame_AdjustColumnSize(Frame = " .. columnFrame:GetName() .. " count = " .. groupCount);

	local rows = groupCount;

	--Everybody calls this function to resize their columns. If there are zero members in the group, setting the height to
	--zero creates a weird, flawed, overlapping border. So, if the rowcount is zero, hide the column altogether.
	if (groupCount > 0) then 
		columnFrame:Show();
	else
		columnFrame:Hide();
	end

	if (rows > LeaderBotScrollSlots) then 
		rows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = requiredWidth +  (LeaderBotborderPadding * 2); --one col, padding on both sides
	local requiredHeight = (rows * LeaderBotplayerHeight +  (LeaderBotborderPadding * 2) ) ; --std column height + buffer on top/bottom

	columnFrame:SetSize(requiredWidth, requiredHeight);

end

-- ******************************************************************************
function LeaderBotFrame_AddNamesToColumn1(columnFrame, team, groupType, addIndex)
	LBmsg("Function: LeaderBotFrame_AddNamesToColumn1(" .. groupType .. ", addIndex: " .. addIndex .. ")");
	
	for name, data in next, team do
		addIndex = addIndex + 1;

		LBmsg("    adding name: " .. name);

		local nameFrame = _G["LeaderBot" .. groupType .."Member" .. addIndex];
		if (nil == nameFrame) then
			nameFrame = columnFrame:CreateFontString("LeaderBot" .. groupType .."Member".. addIndex, "ARTWORK", LeaderBotPlayerNameFont);
		end

		nameFrame:SetPoint("TopLeft", columnFrame, "TopLeft", 6,  (addIndex -1) * -LeaderBotplayerHeight -8 );
		local classIcon = " ";
		local specIcon = " ";

		if (groupType == "healer" or groupType == "tank" or groupType == "crowdcontroller") then
			--TODO: check priests if theyhave shadow form
			--TODO: check paladiins for ??? 
			--TODO: check druids for something, maybe bear form, cat form?

			local class = UnitClass(name);
			if (class) then 
				--class must never be nil, except during solo-play debugging.
				classIcon = LeaderBot_GetTextForClass(UnitClass(name) );
			else
				--debugging code for solo-play
				if (groupType == "healer") then
					classIcon = LeaderBot_GetTextForClass("Priest");
					specIcon = LeaderBot_GetTextForSpellID(shadowFormSpellID);
				else 
					classIcon = LeaderBot_GetTextForClass("Warrior");
					specIcon = "";
				end 
			end

			if (class == "Priest" or class == "Druid") then 
				local buffIndex = 1;
				local buffName, icon, count, debuffType, fullduration, expirationTime, source, 
						isStealable, namePlateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, 
						nameplateShowAll, timeMod = UnitBuff(name, buffIndex);
		
				while (nil ~= buffName) do

					if (string.match(buffName, "Shadowform")) then
						--well here's a another fun fact. If the target is too far away, UnitBuff still returns the buffName,
						--but other details, like source, become nil.

						specIcon = LeaderBot_GetTextForSpellID(shadowFormSpellID);
						break;

					elseif (string.match(buffName, "Bear Form") ) then
						specIcon = LeaderBot_GetTextForSpellID(bearFormSpellID);
						break;
					end

					buffIndex = buffIndex + 1;
					buffName, icon, count, debuffType, fullduration, expirationTime, source, 
							isStealable, namePlateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, 
							nameplateShowAll, timeMod = UnitBuff(name, buffIndex);
				end

			end

		end


		local level = team[name]["level"];
		nameFrame:SetText(classIcon .. "  " .. specIcon .. "  " .. name .. " (" .. level .. ")" );

		--save the player name in a custom variable for easier retrieval
		nameFrame.playerName = name;

		--color code lower-level healers
		if (level == 60) then
			nameFrame:SetTextColor(0.24, 1.0, 0.24, 1);
		else
			nameFrame:SetTextColor(1.0, 0.24, 0.24, 1);
		end

		--indicate team members that are offline
		if (data["online"] == "online") then
			-- this user may have went offline previously and came back. Or maybe they've always been online,
			--either way, remove them from the list of names that have been shows as offline.
			offlineList[name] = nil;

		elseif (data["online"] == "offline" ) then
			nameFrame:SetTextColor(0.50, 0.50, 0.50, 1);
			nameFrame:SetText(classIcon .. "  " .. specIcon .. "  " .. name .. " (OFFLINE)" );	

			if (groupType == "healer" and LeaderBotOptions["HealerLeftPopup"]) then		
				--the user wants to be notified that a healer left	
				if (offlineList[name] == nil) then 
					showOfflineAlert = true;
					offlineList[name] =  name .. " is offline.\n ";
				end 
			
			elseif (groupType == "tank" and LeaderBotOptions["TankLeftPopup"]) then
				--the user wants to be notified that a tank left	
				if (offlineList[name] == nil) then 
					showOfflineAlert = true;
					offlineList[name] =  name .. " is offline.\n ";
				end 
			
			elseif (groupType == "Warlock" and LeaderBotOptions["LockOfflinePopup"]) then		
				--the user wants to be notified that a warlock left
				if (offlineList[name] == nil) then 
					showOfflineAlert = true;		
					offlineList[name] =  name .. " is offline.\n ";
				end 
			end	

		end

		--a re-used nameFrame may be hidden from prior window resizings
		nameFrame:Show();

	end 

	LBmsg("     addIndex = " .. addIndex );

	--there may be fontstrings still visible with names of players that have left. We need to hide those.
	-- search for fontStrings until we hit nil.
	local hideIndex = addIndex + 1;
	local nameFrame = _G["LeaderBot" .. groupType .. "Member" .. hideIndex];
	while (nil ~= nameFrame) do
		nameFrame:Hide();

		hideIndex = hideIndex + 1;
		nameFrame = _G["LeaderBot" .. groupType .. "Member" .. hideIndex];
	end

	--tell the next iteration at which slot we left off (aka, at which slot to begin adding the next class)
	return addIndex;
end 

-- ********************************************************************************
function LeaderBotFrame_AddColumnRaidIcons(teamMembers, groupCount, forceDisable, 
						parentFrame, referenceColumn, 
						thisColumnWidth, columnFrameName, columnTitleText, 
						rowBaseName, dropDownBaseName,  
						assignmentArray, assignmentType,
						announceButtonName, announcementFunction, announcementTitleText,
						onClickHandler, 
						singleOrMultiSelect)
	LBmsg("Function: LeaderBotFrame_AddColumnRaidIcons");

	if (singleOrMultiSelect == nil) then 
		singleOrMultiSelect = LeaderBotSingleSelect;
	end

	local columnFrame = _G[columnFrameName];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", columnFrameName, parentFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", referenceColumn, "TopRight", 0, 0);

	LeaderBotFrame_AdjustColumnSize(columnFrame, groupCount, thisColumnWidth)

	local colTitle = _G[columnFrameName .. "Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString(columnFrameName .. "Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", columnFrame, "Top", 0, 10);
	colTitle:SetText(columnTitleText); 

	for i = 1, groupCount, 1 do

		local rowFrame = _G[rowBaseName .. i ];
		local name = rowFrame.playerName;

		-- make the drop down menu for tank targets
		local dd = _G[dropDownBaseName .. i];
		if (nil == dd) then
			dd = CreateFrame("Frame", dropDownBaseName..i, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd, thisColumnWidth - 35);
		UIDropDownMenu_EnableDropDown(dd);
		--reused element may have been hidden on a prior pass
		dd:Show();

		local topBuffer = 2;
		if (i > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (i -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = 
			function() 
				LeaderBotFrame_RaidIconDropDownInit(name, dd, onClickHandler, singleOrMultiSelect, assignmentArray, assignmentType);
			end;
		dd.playerName = name;

		local targetName;
		local iconText;

		--single select assignments are re-applied here; multi-select assignments are re-applied in the init function
		if (singleOrMultiSelect == LeaderBotSingleSelect) then 

			targetName = assignmentArray[assignmentType][name];

			--did the player previously have an assignment? If so, re-apply it.
			if (targetName == LeaderBotNone or targetName == LeaderBotLegacyNone) then
				--they were previously assigned to NONE
				iconText = LeaderBotNone;		

			elseif (targetName) then
				--they previously had an assignment; re-apply it.
				iconText = LeaderBot_GetTextForRaidIconName(targetName);
			
			else 
				--they have not had an assignment before, so assign them to RESERVE
				assignmentArray[assignmentType][name] = LeaderBotNone;	
				iconText = LeaderBotNone;		
						
			end 
		end 


		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.
 
		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[dropDownBaseName .. i .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", dropDownBaseName ..i .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[dropDownBaseName .. i .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(dropDownBaseName .. i .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);

		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		if (singleOrMultiSelect == LeaderBotMultiSelect) then 
			LeaderBot_InitNewMultiSelectPlayer(name, assignmentArray, assignmentType);
			local labelText = LeaderBot_UpdateMultiselectDropDownLabel(assignmentArray, assignmentType, name);
			ddMyLabel:SetText(whiteColorCode .. labelText);
		else
			ddMyLabel:SetText(whiteColorCode .. iconText);		
			UIDropDownMenu_SetSelectedValue(dd, iconText);
		end

--TODO: this will get removed from the healer tab

		--special, one-off case for shackle: Don't show shackle drop-downs for non-Priests
		if (columnTitleText == "Shackle" and UnitClass(name) ~= "Priest") then
			--no shackle from this guy
			UIDropDownMenu_DisableDropDown(dd);
			ddMyLabel:SetText(greyColorCode .. "N/A");
		end 

		--special cases for the multi-class Crowd Controller tab
		if(string.find(columnFrameName, "CrowdController") ) then
			if (columnTitleText ~= UnitClass(name) ) then 
				--this class can't perform the actions in this column (aka, Warlock's can't do Rogue stuff)
				UIDropDownMenu_DisableDropDown(dd);
				ddMyLabel:SetText(greyColorCode .. "N/A");
			end 
		end 

	end

	--hide any leftover elements if locks have left the raid
	local i = groupCount + 1;
	local dd = _G[dropDownBaseName .. i];
	while ( dd ~= nil) do
		dd:Hide();

		i = i  + 1;
		dd = _G[dropDownBaseName .. i];
	end

	local button = _G[announceButtonName];
	if (nil == button) then
		button = CreateFrame("Button", announceButtonName, columnFrame, "OptionsButtonTemplate");
	end

	button:SetText("Announce");
	button.tooltipText = "Announce assignments to selected chat channels";
	button:SetSize(75,30);

	local columnCenterX, columnCenterY = columnFrame:GetCenter();
	local panelCenterX, panelCenterY = parentFrame:GetCenter();
	button:SetPoint("Bottom", parentFrame, "Bottom", columnCenterX - panelCenterX, 2);
	--button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);

	if (forceDisable == true) then
		button:Disable();
	end
	button:SetScript("OnClick", function(self)
									LLmsg("Handler: " .. announceButtonName .. ":OnClick()");
									announcementFunction(announcementTitleText, assignmentType);
								end);

	if (groupCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame, assignmentArray;
end

-- ******************************************************************************
function LeaderBotFrame_RaidIconDropDownInit(playerName, dropDownMenu, onClickHandler, singleOrMultiSelect, assignmentArray, assignmentType)
	TLmsg("Function: LeaderBotFrame_RaidIconDropDownInit(" .. playerName .. ")" );
	
	local rowInfo = UIDropDownMenu_CreateInfo();

	if (singleOrMultiSelect == LeaderBotMultiSelect) then 
		rowInfo.isNotRadio = true;
		rowInfo.keepShownOnClick = true;
		
		rowInfo.text = "Multi-Select:";
		rowInfo.isTitle = true;		
		UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

		rowInfo = UIDropDownMenu_CreateInfo();

		rowInfo.isNotRadio = true;
		rowInfo.keepShownOnClick = true;
		if (LeaderBotIsAssistant) then 
			rowInfo.isTitle = false;
		else
			rowInfo.isTitle = true;
		end 	

	else 
		rowInfo.isNotRadio = false;
		rowInfo.keepShownOnClick = false;
		if (LeaderBotIsAssistant) then 
			rowInfo.isTitle = false;
		else
			rowInfo.isTitle = true;
		end 
	end 

	rowInfo.text = LeaderBotNone ;
	rowInfo.value = LeaderBotNone;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = nil;
	rowInfo.checked = false;
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBotNone;
	rowInfo.func = onClickHandler;								--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][LeaderBotNone] == nil) then 
			assignmentArray[assignmentType][playerName][LeaderBotNone] = false;
		end
		rowInfo.checked = assignmentArray[assignmentType][playerName][LeaderBotNone];
	end 
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = LeaderBot_GetTextForRaidIconNumber(1);
	rowInfo.value =  LeaderBot_GetTextForRaidIconNumber(1);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{star}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(2);
	rowInfo.value =  LeaderBot_GetTextForRaidIconNumber(2);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_2";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{circle}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(3);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(3);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_3";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{diamond}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(4);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(4);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_4";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{triangle}";								--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(5);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(5);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_5";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{moon}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(6);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(6);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_6";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{square}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(7);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(7);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_7";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{cross}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  LeaderBot_GetTextForRaidIconNumber(8);
	rowInfo.value = LeaderBot_GetTextForRaidIconNumber(8);
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_8";
	rowInfo.arg1 = playerName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = "{skull}";									--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = onClickHandler;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	if (singleOrMultiSelect == LeaderBotSingleSelect) then 
		rowInfo.checked = false;
	else 
		if (assignmentArray[assignmentType][playerName][rowInfo.arg2] == nil) then 
			assignmentArray[assignmentType][playerName][rowInfo.arg2] = false;
		end	
		rowInfo.checked = assignmentArray[assignmentType][playerName][rowInfo.arg2];
	end	
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

end

-- *******************************************************************************
-- Note1: holy christ, I would have never figured out how to parse GetChannelList() based on the lack of online docs. Lucky I found examples from others
-- Note2: When you first log in and check this upon adding loading, the channels are still nil. Have to find an event, possibly VARIABLES_LOADED
function LeaderBotFrame_GetChatChannels()
	LBmsg("Function: LeaderBotFrame_GetChatChannels");

	local rawChannels = {GetChannelList()};		--clever. I had no idea you could do this.
	local channelCount = #rawChannels / 3;
	local channelList = {};
	
	local usableChannelCount = 0;

	for i = 1, channelCount do
		
		local channelName = rawChannels[i * 3 - 2] .. ". "..rawChannels[i * 3 - 1];
		local channelType = "CHANNEL";
		local channelNumber = rawChannels[i * 3 - 2];

		--filter out global and personal channels
		if (	string.match(channelName, "LookingForGroup")
				or string.match(channelName , "General")
				or string.match(channelName, "Trade")
				or string.match(channelName, "LocalDefense")
				or string.match(channelName, "WorldDefense")
				or string.match(channelName, "pizzafactory")
				) then
				
				--do nothing; prevent user from spamming these channels.
		else
				table.insert(channelList, {rawChannels[i * 3 - 2]..". "..rawChannels[i * 3 - 1], "CHANNEL", rawChannels[i * 3 - 2]});

				usableChannelCount = usableChannelCount + 1;
				LBmsg(channelList[usableChannelCount][CHATCHANNAME] .. ": " .. channelList[usableChannelCount][CHATCHANTYPE] .. ", " .. channelList[usableChannelCount][CHATCHANNUM]);
		end

	end

	return usableChannelCount, channelList
end

--local CHATCHANNAME = 1;
--local CHATCHANTYPE = 2;
--local CHATCHANNUM = 3;

-- *****************************************************************************
function LeaderBot_RegisterColumnForScrolling(columnCount, group, columnIndex, column, widgetBaseName, widgetBaseName2)
	LBmsg("Function: LeaderBot_RegisterColumnForScrolling() ");

	if (columnCount > 0) then 
		leaderBotWidgetArray[group][columnIndex] = {column, widgetBaseName, widgetBaseName2};
		leaderBotWidgetArray[group].columnCount = columnCount;
	else 
		LBassert(colorCode .. "LeaderBot.RegisterColumnForScrolling: **ERROR** columnCount must be greater than zero");
	end

--[[
	--example code
	leaderBotWidgetArray["Warlock"][1] = {_G["WarlockLeaderColumn1"], "LeaderBotWarlockMember"};
	leaderBotWidgetArray["Warlock"][2] = {_G["ockLeaderColumn2"], "ockLeaderSSDropDown", "ockLeaderSSIcon"};
	leaderBotWidgetArray["Warlock"][3] = {_G["ockLeaderColumn3"], "ockLeaderBanishDropDown"};
	leaderBotWidgetArray["Warlock"][4] = {_G["ockLeaderColumn4"], "ockLeaderCurseDropDown"};
	leaderBotWidgetArray["Warlock"][5] = {_G["ockLeaderColumn5"], "ockLeaderSeduceDropDown"};
--]]

end

-- *****************************************************************************
function LeaderBot_SetupScrollBar(groupType)
		--LBmsg("Function: LeaderBot_SetupScrollBar(" .. GetTime() .. ")", true);
		LBmsg("Function: LeaderBot_SetupScrollBar()");

		if (groupType ~= LeaderBotOptions["currentPanel"]) then 
			--we've received a request to configure the scroll bar for a tab that's not
			--the current tab. This happens during initialization. In that case, just bail out
			return;
		end

		displayFrame = LeaderBotPanels[groupType];

		LeaderBotScrollFrame = _G["LeaderBotScrollFrame"];
		if (LeaderBotScrollFrame == nil) then
			LeaderBotScrollFrame = CreateFrame("ScrollFrame", "LeaderBotScrollFrame", LeaderBotFrameMax, "FauxScrollFrameTemplate"); 
		end 

		--the size of the scrollFrame doesn't matter. It just has to exist and draw the vertical slider. We position the rest of our elements 
		--however we want, wherever we want, in the main panel.
		LeaderBotScrollFrame:SetPoint("TopLeft", displayFrame, "TopRight", -10, -LeaderBotheaderHeight + (LeaderBotplayerHeight  / 2)  );
		LeaderBotScrollFrame:SetPoint("BottomLeft", displayFrame, "BottomRight", -10, 0)
		LeaderBotScrollFrame:SetSize( 5, (LeaderBotScrollSlots +1) * LeaderBotplayerHeight); -- +1 helps adjust position of scroll buttons

		local scrollBackground = CreateFrame("Frame", "LeaderBotScrollFrameBackground", LeaderBotScrollFrame, BackdropTemplateMixin and "BackdropTemplate");
		scrollBackground:SetPoint("TopLeft", LeaderBotScrollFrame, "TopRight", 5, 0 );
		scrollBackground:SetPoint("BottomLeft", LeaderBotScrollFrame, "BottomRight");
		scrollBackground:SetWidth(18);
		scrollBackground:SetBackdrop(labelBackdrop);
		scrollBackground:SetBackdropColor(0.1, 0.1, 0.1, 0.85);


		LeaderBotScrollFrame:SetScript("OnVerticalScroll", function(self, offset)
															FauxScrollFrame_OnVerticalScroll(self, offset, LeaderBotplayerHeight, LeaderBot_OnVerticalScroll);
													end);

		LeaderBotScrollFrame:SetScript("OnMouseWheel", function(self, arg1)
															ScrollFrameTemplate_OnMouseWheel(self, arge1, self.scrollBar);
													end);

		LeaderBotScrollFrame:SetScript("OnShow", LeaderBot_OnVerticalScroll);

		--scroll bar may have been hidden by a shorter panel that didn't need scrolling
		LeaderBotScrollFrame:Show();

		--when we resize an existing frame, we don't get an onShow event, so we have to manually initialize the
		--vertical bar 
		LeaderBot_UpdateScroll(groupType);

end 

-- *************************************************************************************
--we can't pass desired arugments directly from the SetScript FauxScrollFrame_OnVerticalScroll, so
--we have FSF_OnVerticalScroll call this function, which looks up the proper groupType (ie, the
--name of the current panel) and then do the scrolling work.
function LeaderBot_OnVerticalScroll(arg1)
	--LBmsg("Function: LeaderBot_OnVerticalScroll(" .. GetTime() .. ")", true);

	LeaderBot_UpdateScroll(LeaderBotOptions["currentPanel"]);

	--[[
	--based on prior debugging, this is what's in arg1"
	arg1 = LeaderBotScrollFrame {
 										ScrollBar = LeaderBotScrollFrameScrollBar {}
 										offset = 0
 										ScrollChildFrame = LeaderBotScrollFrameScrollChildFrame {}
 										0 = <userdata>
									}
	--]]

end

-- *************************************************************************************
function LeaderBot_UpdateScroll(groupType)
	LBmsg("Function: LeaderBot_UpdateScroll(" .. GetServerTime() .. ") ");

--[[
	--debug code
	for i = 1, leaderBotWidgetArray["Warlock"].columnCount do
		print (colorCode .. "i: " .. i);
		local column = leaderBotWidgetArray["Warlock"][i][1];
		local widgetName = leaderBotWidgetArray["Warlock"][i][2];
		print (colorCode .. "     Widgetname: "  .. widgetName);
	end
--]]

	--FauxScrollFrame has a major bug (for like 15 years). If the dataRows is less than the LeaderBotScrollSlots, instead of disabling the scroll bar,
	--the entire scrollFrame vanishes. So we have to make sure we have enough data to warrant scrolling before we let
	--FauxScrollFrame do anyting

	local teamCount = classCounts[groupType];
	if (teamCount <= LeaderBotScrollSlots) then
		--there's not enough team members to warrant scrolling. Just display elements 1-10 normally, which was already
		--handled in the _AddColumn*() functions

	else
		--we have enough data to warrant scrolling. FSF_Update() does is apparently necessary to initialize the bar, so to speak,
		--so it can do the math to determine which row we're on based on the scroll bar position. It also seems to set the "stops"
		--on the vertical scroll bar, also.
		FauxScrollFrame_Update(LeaderBotScrollFrame, teamCount, LeaderBotScrollSlots, LeaderBotplayerHeight); 

		--Call FauxScrollFrame_GetOffset(framename) to get the offset N of the bar. This just does the math to 
		--convert the scrollbar position to an integer from 1 to maxNumberItems
		local offset = FauxScrollFrame_GetOffset(LeaderBotScrollFrame);

		--hide all row elements less than the offset
		--offset begins counting at zero
		for i = 1, offset do
			--find the i widget in each column j and hide it.
			for j = 1, leaderBotWidgetArray[groupType].columnCount do
				local widgetName = leaderBotWidgetArray[groupType][j][2];
				local widget = _G[widgetName .. i];
				widget:Hide();

				--special case for soulstones, which have 2 widges per row
				widgetName = leaderBotWidgetArray[groupType][j][3];
				if (widgetName) then 
					local widget = _G[widgetName .. i];
					widget:Hide();
				end 
			end
		end

		--adjust positions of the visible items. i tells us the name of the row we want to adjust, posIndex tells us the visible slot we want to use.
		--we have to use offset+1 as offset begins at zero, but rowNames begin at 1.
		local posIndex = 1;		
		for i = offset +1, offset + LeaderBotScrollSlots do	

			--find the i widget in each column j and adjust it's position
			for j = 1, leaderBotWidgetArray[groupType].columnCount do
				local column = leaderBotWidgetArray[groupType][j][1];
				local widgetName = leaderBotWidgetArray[groupType][j][2];
				local widget = _G[widgetName .. i];

				--some crap needed to help the textstrings in col1 align vertically with dropdowns in the other columns
				local vertAlignmentBuffer = 0;
				local horizAlignmentBuffer = -5;
				if (j ==	 1) then
					vertAlignmentBuffer = 8;
					horizAlignmentBuffer = 10
				end 

				widget:SetPoint("TopLeft", column, "TopLeft", horizAlignmentBuffer,  (posIndex -1) * -LeaderBotplayerHeight -vertAlignmentBuffer );
				widget:Show();

				--note, for items that have two widges per row, like SS, we don't need to do anything to the second widget.
				--it will be hidden/shown as needed anyway, and is already setPointed against widget1, so it will move properly.
			end;

			posIndex = posIndex + 1;			
		end

		--hide all the row elements beyond the max scroll slots
		for i = offset + LeaderBotScrollSlots + 1, teamCount do	

			--find the i widget in each column j and hide it
			for j = 1, leaderBotWidgetArray[groupType].columnCount do
				local widgetName = leaderBotWidgetArray[groupType][j][2];
				local widget = _G[widgetName .. i];
				widget:Hide();

				--special case for soulstones, which have 2 widges per row
				widgetName = leaderBotWidgetArray[groupType][j][3];
				if (widgetName) then 
					local widget = _G[widgetName .. i];
					widget:Hide();
				end 

			end
		end

	end

end

-- *******************************************************
function LeaderBot_AssignmentRequestWhisper(whipserMsg, senderRealmName, senderName, suppressRaidIcons )

	local modules = {LockLeader_Assignments, HealerLeader_Assignments, TankLeader_Assignments, MageLeader_Assignments, HunterLeader_Assignments, CrowdController_Assignments};

	for index, module_Assignments in next, modules do

		for assignmentType, assignments in next, module_Assignments do

			--are we single or multiselect assignments?
			if (string.match(assignmentType, "Multi")) then 

				if (module_Assignments[assignmentType][senderName]) then 

					local msg = "";
					local comma = "";

					for assignmentName, assignmentValue in next, module_Assignments[assignmentType][senderName] do

						if (assignmentValue == true) then 
						
							local text = assignmentName;

							--TODO: Apparently this isn't needed after all, and actually breaks sendChatMessage when
							--sending multiple icons.
							--[[
							if (string.match(text, "{")) then 
									text = LeaderBot_GetTextForRaidIconName(text);
							end 
							--]] 

							msg = text .. comma .. msg;
							comma = ", ";
						end

					end 

					local response = LeaderBotOutgoingMsgPrefix .. assignmentType .. ": " .. msg;
					SendChatMessage(response, "WHISPER", nil, senderName);
				end 
		
			else --single select assignments

				for name, assignment in next, assignments do
				
					if (name == senderName) then 

						local response = LeaderBotOutgoingMsgPrefix .. assignmentType .. ": " .. assignment;
						SendChatMessage(response, "WHISPER", nil, senderName);
					end

				end 

			end 

		end 
	end
end 

-- **************************************************************
function LeaderBot_InitNewMultiSelectPlayer(playerName, assignmentArray, assignmentType)

	--Somewhere is resetting _Assignments, but MULTISELECT hasn't been added there yet.
	--I can't find where, so am forcing the issue here.
	if (not assignmentArray[assignmentType]) then
		assignmentArray[assignmentType] = {}
	end 

	--on first click for this healer, the array won't exist yet. Init it here.
	if (not assignmentArray[assignmentType][playerName]) then
		assignmentArray[assignmentType][playerName] = {};
	end 

end 

-- *************************************************************
function LeaderBot_UpdateMultiselectDropDownLabel(assignmentArray, assignmentType, playerName)
	local text = "";
	local count = 0;

	--walk all the assignments stored for thsi player
	for key, value in next, assignmentArray[assignmentType][playerName] do

		--assume this entry should be included in the display unless specifically disabled, below
		local includeInResults = true;

		--value is true/false for each row in the drop down
		if (value) then
			--we found a selected value

			--special case for Heals_MultiSelect: If a playerName is provided as a target (saved from a prior raid)
			--but that player is not in this raid, then skip them
			if (assignmentType == "Heals_MultiSelect") then 

				if (key ~= "Raid" and key ~= "Melee" and key ~= "Ranged" and string.find(key, "Group ") == nil) then 

					--this key (targetName) appears to be a player name. are they in our current raid and do they have
					--a role (main/assist tank) assigned?

					--assume this target is to be excluded unless positvely matched, below.
					includeInResults = false;

					local raidSize = GetNumGroupMembers();
					for raidIndex = 1, raidSize do

						local name, rank, subgroup, level, class, fileName, 
  										zone, online, isDead, role, isML, combatRole = GetRaidRosterInfo(raidIndex);

						if (name == key) then

							--this target is in our current raid. But do they have a role?
							if (role) then
								--ok, we found them in the current raid, and they do have a role assigned. Guess we
								--better include them.
								includeInResults = true;	
							end

							--bail out once we found the name, role or not.
							break;
						
						end

					end 

				end
			end

			if (includeInResults) then 
				count = count + 1;
				if (count == 1) then
					--if there's only one assignment then set the text to that assignment
					text = key;

					if (string.match(text, "{")) then 
						text = LeaderBot_GetTextForRaidIconName(text);
					end 

				else 
					--as soon as we know we have multiple values we can set the string to "multiple" and bail out
					text = "*multiple*";
					break;
				end

			end
	
		end
	
	end 

	return text;

end 

-- ************************************************************************************
function LeaderBot_TeamCount(team)
	return classCounts[team];
end 

-- ******************************************************************************
function LeaderBot_ChannelDropDownInit(dropDownMenu, channelCount, channelList, onClickHandler)
	HUNLmsg("Function: HunterLeader_ChannelDropDownInit()");

	--I haven't found a good event yet that reliably fires after chats have been loaded
	channelCount, channelList = LeaderBotFrame_GetChatChannels();


	if ( IsInRaid() ) then
		rowInfo = UIDropDownMenu_CreateInfo();
		rowInfo.text = "Raid";
		rowInfo.value = "Raid";
		rowInfo.owner  = dropDownMenu;
		rowInfo.icon = nil;
		rowInfo.checked = false;
		rowInfo.arg1 = "RAID";						
		rowInfo.arg2 = nil;
		rowInfo.func = onClickHandler;		
		UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to
	end

	if (IsInGroup() ) then
		rowInfo = UIDropDownMenu_CreateInfo();
		rowInfo.text = "Party";
		rowInfo.value = "Party";
		rowInfo.owner  = dropDownMenu;
		rowInfo.icon = nil;
		rowInfo.checked = false;
		rowInfo.arg1 = "PARTY";						
		rowInfo.arg2 = nil;
		rowInfo.func = onClickHandler;		
		UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	end

	for i=1, channelCount do
			--create the dropdown row

			local name = channelList[i][CHATCHANNAME];

			if (	string.match(name, "LookingForGroup")
				or string.match(name , "General")
				or string.match(name, "Trade")
				or string.match(name, "LocalDefense")
				or string.match(name, "WorldDefense")
				or string.match(name, "pizzafactory")
				) then

					-- Don't let the user spam global channels or my private channels
					--this is effectivelu a fake "continue" statement. LUA apparently has no actual continue statement.
					
			else
				rowInfo = UIDropDownMenu_CreateInfo();
				rowInfo.text = channelList[i][CHATCHANNAME];
				rowInfo.value = channelList[i][CHATCHANNAME];
				rowInfo.owner  = dropDownMenu;
				rowInfo.icon = nil;
				rowInfo.checked = false;
				rowInfo.arg1 = "CHANNEL";						
				rowInfo.arg2 = channelList[i][CHATCHANNUM];
				rowInfo.func = onClickHandler;		
				UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to	
			end

	end

end

-- ***************************
--homegrown assertion that will force a stack trace to be displayed
function LBAssert(msg)
	print(msg);
	local crashString = " " .. nonExist[56];
end 

