--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
--]]

local classicOrRetail;
local LL_Debug = 0;
local LL_loopcontrol = 0;
local version = "1.67";

LockLeaderFrameMax = nil;

local ssStatusOld = {};

local LOCKSS = "Soulstone";
local LOCKBANISH = "Banish";
local LOCKCURSES ="Curse";
local LOCKSEDUCE = "Seduce";
local LOCKSSFOUNDON = "SoulstoneFoundOn";
local LOCKASSIGNMENTCOUNT = 5;
local assignmentIndices = {LOCKSS, LOCKBANISH, LOCKCURSES, LOCKSEDUCE, LOCKSSFOUNDON};

--Global variables, to be saved.
LockLeader_Assignments = {};
ssStatusArray = {};

--each assignment index (1-5) is a table of values associated with that assignment type
LockLeader_Assignments[LOCKSS] = {};		--.warlockName, .warlockName..ssFound
LockLeader_Assignments[LOCKBANISH] = {};
LockLeader_Assignments[LOCKCURSES] = {};
LockLeader_Assignments[LOCKSEDUCE] = {};
LockLeader_Assignments[LOCKSSFOUNDON] = {};

local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local LockLeader_MsgChannel ={};

local columnWidth = 150;
local columnWidthSS = 180;
local columnWidthBanish = 100;
local columnWidthCurses = 140;
local columnWidthSeduce = 100;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CurseOfAgonyID = 11713;
local CurseOfDoomID = 603;
local CurseOfTonguesID = 11719;
local CurseOfRecklessnessID = 7658;
local CurseOfWeaknessID = 7646;
local CurseOfElementsID = 11721;
local CurseOfShadowID = 17862;
local CurseArray = {
						["Agony"] 		= CurseOfAgonyID,
						["Doom"] 		= CurseOfDoomID,
						["Tongues"] 	= CurseOfTonguesID,
						["Recklessness"]= CurseOfRecklessnessID,
						["Weakness"] 	= CurseOfWeaknessID,
						["Elements"] 	= CurseOfElementsID,
						["Shadow"] 		= CurseOfShadowID
					};

local playerClass = {"Priest", "Paladin", "Shaman", "Warrior", "Druid", "Hunter", "Mage", "Rogue", "Warlock"};

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local spellID_GreaterSoulstone = 20756;

local SOULSTONE = "Soulstone";

local colorCode = "|cffee1100 ";
local whiteColorCode = "|cffffffff";

local ssDropDownName = "LockLeaderSSDropDown";
local ssIconName = "LockLeaderSSIcon";
local banishDropDownName = "LockLeaderBanishDropDown";
local curseDropDownName = "LockLeaderCurseDropDown";
local seduceDropDownName = "LockLeaderSeduceDropDown";
local lockDropDowns = {
						[LOCKSS] = ssDropDownName,
						[LOCKBANISH] = banishDropDownName,
						[LOCKCURSES] = curseDropDownName,
						[LOCKSEDUCE] = seduceDropDownName,
						};


local soulStoneReserve = "None (Hold in reserve)";
local lockDeparted = "departed";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};
	
local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function LLmsg(msg, forceShow)
	if (LL_Debug == 1 or forceShow ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function LockLeaderFrame_OnLoad(parentFrame, currentTeamMembers, currentWarlockCount)
	LLmsg("function: LockLeaderFrame_OnLoad()");

	-- derive the frame from FauxScrollFrameTemplate
	-- assign OnVerticalScroll event to FauxScrollFrame_OnVerticalScroll(self, offset, rowHeight, myFunction)
	-- in MyFunction() query the offset and update the displayed rows accordingly

	LockLeaderFrameMax = CreateFrame("Frame", "LockLeaderFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate");								

		-- UI Commands
	SLASH_LockLeader1 = "/ll";
	SLASH_LockLeader2 = "/LL";

	SlashCmdList["LockLeader"] =LockLeader_SlashHandler;

	local monitorFrame = CreateFrame("Frame", "LockLeaderSSMonitorFrame", UIParent, BackdropTemplateMixin and "BackdropTemplate");
	monitorFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	monitorFrame:SetScript("OnEvent", LockLeaderFrame_OnEvent);

	ChatFrame1:AddMessage(colorCode .. " LockLeader Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's qLockLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	LockLeader_InitMainFrame(currentTeamMembers, currentWarlockCount);

	return(LockLeaderFrameMax);
end;

-- *************************************************************************************
function LockLeader_InitMainFrame(currentTeamMembers, currentWarlockCount)
	LLmsg("Function: LockLeader_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove warlocks

	-- ******************************************************************************************
	--each of the panels call their own OnShow() functions during initialization, so on startup
	--we find LeaderBotFrameMax's width set to the last panel that loaded last in the .toc file.
	--Hiding this frame on startup prevents that. LeaderBotFrame will adjust the width upon
	--VARIABLES_LOADED
	LockLeaderFrameMax:Hide();

	LockLeaderFrameMax:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 4);
	LockLeader_AdjustMainPanelSize(currentWarlockCount);

	LockLeaderFrameMax:SetBackdrop(mainBackdrop);
	LockLeaderFrameMax:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	LockLeaderFrameMax:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	LockLeaderFrameMax:EnableMouse(true);
	LockLeaderFrameMax:SetScript("OnShow", function(self)
					LLmsg("Function: LockLeaderFrameMax_OnShow() ");

					if (LeaderBotOptions["currentPanel"] == "Warlock") then
						LeaderBotFrameMax:SetWidth(LockLeaderFrameMax:GetWidth() );
					end


				end);
	
	local appLogo = LeaderBot_GetTextForClass("Warlock");
	local appTitle = LockLeaderFrameMax:CreateFontString("LockLeaderAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", LockLeaderFrameMax, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Lock Leader " .. version);

	local infoText = LockLeaderFrameMax:CreateFontString("LockLeaderInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	--infoText:SetText(colorCode .. "NOTE: Initial SS assignment of \"None (Hold in reserve)\" doesn't actually stick. Manually reset all SS assignments");

	columnFrames[1] = LeaderBotFrame_AddColumn1(currentTeamMembers, currentWarlockCount, "Warlock", LockLeaderFrameMax, "Warlocks");
	columnFrames[2] = LockLeader_AddColumnSoulstoneTargets(currentTeamMembers, currentWarlockCount, true);
	--columnFrames[3] = LockLeader_AddColumnBanishTargets(currentTeamMembers, currentWarlockCount, true);
	columnFrames[3], LockLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(currentTeamMembers, currentWarlockCount, true,
						LockLeaderFrameMax, columnFrames[2],
						columnWidthBanish, "LockLeaderColumn3", "Banish",
						"LeaderBotWarlockMember", banishDropDownName,
						LockLeader_Assignments, LOCKBANISH,
						"LockLeaderAnnounce3", LockLeader_AnnounceToChat, "Leader Bot: Banish Assignments", 
						LockLeader_BanishDropDownOnClick,
						LeaderBotSingleSelect) ;
	columnFrames[4] = LockLeader_AddColumnCurseAssignments(currentTeamMembers, currentWarlockCount, true);
	columnFrames[5], LockLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(currentTeamMembers, currentWarlockCount, true,
						LockLeaderFrameMax, columnFrames[4],
						columnWidthSeduce, "LockLeaderColumn5", "Seduce",
						"LeaderBotWarlockMember", seduceDropDownName,
						LockLeader_Assignments, LOCKSEDUCE,
						"LockLeaderAnnounce5", LockLeader_AnnounceToChat, "Leader Bot: Seduce Assignments", 
						LockLeader_SeduceDropDownOnClick,
						LeaderBotSingleSelect) ;


	--setup scrolling, if needed:
	if (currentWarlockCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 1, columnFrames[1], "LeaderBotWarlockMember");
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 2, columnFrames[2], ssDropDownName, ssIconName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 3, columnFrames[3], banishDropDownName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 4, columnFrames[4], curseDropDownName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 5, columnFrames[5], seduceDropDownName);

		LeaderBot_SetupScrollBar("Warlock");
	end 

	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","LockLeaderChatChannelDropDown", LockLeaderFrameMax, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );
	dd:SetPoint("BottomLeft", LockLeaderFrameMax, "BottomLeft", -15, 0);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelList, LockLeader_ChannelDropDownOnClick) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();
	UIDropDownMenu_SetText(dd, "<Select channel>");

	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end

	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);

	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end 

end

-- ******************************************************************************
function LockLeader_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	LLmsg("LockLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	LockLeader_MsgChannel[LOCKCHANTYPE] = arg1;
	LockLeader_MsgChannel[LOCKCHANLANG] = nil;
	LockLeader_MsgChannel[LOCKCHANNUMB] = arg2;

	LockLeader_AnnounceButtonEnable(1);
	
end

-- ******************************************************************************
--adjust the height as warlocks join or leave
function LockLeader_AdjustMainPanelSize(warlockCount)
	LLmsg("Function: LockLeader_AdjustMainPanelSize(count = " .. warlockCount .. ")" );

	local numberOfRows = warlockCount;
	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidth 
						+ columnWidthSS 
						+ columnWidthBanish 
						+ columnWidthCurses
						+ columnWidthSeduce
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	
	LockLeaderFrameMax:SetSize(math.max(requiredWidth, LeaderBotMinWidth), requiredHeight);

end

-- ******************************************************************************
function LockLeader_AnnounceButtonEnable(state)
	LLmsg("FUnction: LockLeader_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["LockLeaderAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then 
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["LockLeaderAnnounce" .. i];
	end 

end

-- ******************************************************************************
function LockLeader_AnnounceAlertToChat(msg)
	LLmsg("Function: LockLeader_AnnounceAlertToChat(" .. msg .. ")" );
	local chanType = LockLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = LockLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = LockLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot (Lock Leader): No chat channel selected");
		return;
	end

	--example: SendChatMessage("LockLeader: Soulstone assignments", "CHANNEL", nil, "4");

	if (chanNumber == nil) then
		--PARTY or RAID	
		SendChatMessage(msg, chanType);
	else
		--custom channels
		SendChatMessage(msg, chanType, chanLang, chanNumber)
	end

end 


-- ******************************************************************************
function LockLeader_AnnounceToChat(title, assignmentIndex)
	LLmsg("Function: LockLeader_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	local chanType = LockLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = LockLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = LockLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot (Lock Leader): No chat channel selected");
		return;
	end

	--SendChatMessage("LockLeader: Soulstone assignments", "CHANNEL", nil, "4");
	--if (chanType == "RAID") then
	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

	local footNote = "";

	for assignmentKey, assignmentValue in next, LockLeader_Assignments[assignmentIndex] do
		local warlockName, target;
		local found = "";
		local expiration = "";
		local msg = nil;

			warlockName = assignmentKey;
			targetName = assignmentValue;

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(warlockName);
			if (isInGroup) then 
				
				--but if the key is Soulstone, then pull the hidden ssFound value
				if (assignmentIndex == LOCKSS) then
					local ssAppliedToName = LockLeader_Assignments[LOCKSSFOUNDON][warlockName];

					local timer = "";
					local i = 1;
					local ssIcon = _G[ssIconName .. i];
					while (ssIcon ~= nil) do 
						if (warlockName == ssIcon.warlockName) then 
							timer = ssIcon.timerText:GetText();
							break;
						end

						i = i + 1;
						ssIcon = _G[ssIconName .. i];

					end 

					if (ssAppliedToName == targetName ) then
						-- we know the ss is up; we've found it on this target. DO we have a time estimate?
						found = "(SS Is Up " .. (timer or "") .. ")";
					else
						--we could not detect a SS on this warlock's specified target. Are we tracking a cooldown estimate?
						if (timer) then 
							timer = timer .. " CD";
						end 

						found = "(SS Not Detected " .. ( timer or "") .. ")*";
						footNote = "* = Not cast or on cooldown. All times are estimates.";
					end 
				end 

				if (string.match(targetName, lockDeparted) ) then
					--this lock has left the building. Nothing to report for them, but it's virtually impossible to remove
					--an entry from a fucking LUA table. So we have to keep them but their target was marked as "departed"
					--which tells us to skip them from the report.
					msg = nil;

				elseif (string.match(targetName, "Hold in reserve") ) then
					msg = "     " .. warlockName .. ": " .. targetName;
				else
					msg = "     " .. warlockName .. ": " .. targetName .. " " .. found;

				end

				if (msg) then 
					--if (chanType == "RAID") then
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end

					--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
					--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
					--sync everything being announced.
					if (string.find(assignmentIndex, "MultiSelect") ) then 
						for key_HealTarget, val_CheckMarkStatus in next, LockLeader_Assignments[assignmentIndex][warlockName] do
							LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", assignmentIndex, warlockName, key_TargetName, val_CheckMarkStatus);
						end
					else 
						LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", assignmentIndex, warlockName, LockLeader_Assignments[assignmentIndex][warlockName]);
					end 

				end 

			end

	end

	if (chanNumber == nil) then
		--Party, raid
		if (footNote ~= nil) then
			SendChatMessage(footNote, chanType);
		end
		SendChatMessage("-------------------------------------------", chanType);
	else
		if (footNote ~= nil) then
			SendChatMessage(footNote, chanType, chanLang, chanNumber);
		end
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
								
end


-- ******************************************************************************
function LockLeader_AddColumnSoulstoneTargets(teamMembers, warlockCount, forceDisable)
	LLmsg("Function: Lockleader_AddColumnSoulstoneTargets()" );

	local columnFrame = _G["LockLeaderColumn2"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "LockLeaderColumn2", LockLeaderFrameMax, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[1], "TopLeft", columnWidth + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, warlockCount, columnWidthSS)

	local colTitle = _G["LockLeaderColumn2Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("LockLeaderColumn2Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Soulstones");

	--get texture paths
	ssTextureEnabledPath = GetSpellTexture(spellID_GreaterSoulstone);
	ssTextureDisabledPath = GetSpellTexture(spellID_GreaterSoulstone);

	local i = 0;
	for warlockName, data in next, teamMembers["Warlock"] do
		i = i + 1;

		-- make the drop down menu for SS targets
		local dd = _G[ssDropDownName .. i];
		if (dd == nil) then
			dd = CreateFrame("Frame",ssDropDownName ..i, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd,columnWidthSS - 65);

		UIDropDownMenu_EnableDropDown(dd);
		if (LeaderBotIsAssistant == false) then 
			UIDropDownMenu_DisableDropDown(dd);
		end 

		local topBuffer = 2;
		if (i > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (i -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() LockLeader_SSDropDownInit(warlockName, dd, teamMembers) end;
		dd.warlockName = warlockName;	

		--did the lock previously have an assignment? If so, re-apply it.
		local ssTargetName = LockLeader_Assignments[LOCKSS][warlockName];
	
		if (ssTargetName == soulStoneReserve) then 
			--they were already onreserve.
			dd.ssTargetName = soulStoneReserve;

		elseif (ssTargetName) then
			--they previously had an assignment; re-apply it. but check,
			--is this target still in the raid?
			local targetClass = UnitClass(ssTargetName);
			if (targetClass) then		
				--yes the target is still here
				dd.ssTargetName = ssTargetName;
			
			else
				--no, target is not still here. reset to RESERVE				
				dd.ssTargetName = soulStoneReserve;
				ssTargetName = soulStoneReserve;

				LockLeader_Assignments[LOCKSS][warlockName] = soulStoneReserve;	
				local msg = warlockName .. " no longer has a Soulstone target as " .. ssTargetName .. " has left the raid";
			
				StaticPopupDialogs["LEADERBOT_SOULSTONE_TARGET_LEFT"] = {
						preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
						timeout = 20,
						whileDead = true,
						hideOnEscape = true,
						text = msg,
							
						button1 = "OK",
						OnAccept = function()

									end,			
					};

				if (LeaderBotOptions["LockTargetPopup"]) then
					StaticPopup_Show("LEADERBOT_SOULSTONE_TARGET_LEFT");
				end

			end

		else 		
			--they have not had an assignment before, so assign them to RESERVE
			dd.ssTargetName = soulStoneReserve;
			ssTargetName = soulStoneReserve;
			
			--TODO: When line 2257 and line 658 are both enabled, WoW locks up at random
			--points. If only one of the lines are enabled, it works fine.
		
			LockLeader_Assignments[LOCKSS][warlockName] = soulStoneReserve;			
		end 

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[ssDropDownName .. i .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", ssDropDownName ..i .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[ssDropDownName .. i .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(ssDropDownName .. i .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		ddMyLabel:SetText(whiteColorCode .. ssTargetName);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		UIDropDownMenu_SetSelectedValue(dd, ssTargetName);

		--set up the Soustone ICon to track status
		local ssIcon = _G[ssIconName .. i];
		if (ssIcon == nil) then
			ssIcon = CreateFrame("Button", ssIconName ..i, columnFrame)
			ssIcon:SetPoint("Left", dd, "Right", -3, 0);
			ssIcon:SetSize(20,20);

			ssIcon:EnableMouse(true);
			ssIcon:SetScript("OnEnter", LockLeader_OnEnterIcon);
			ssIcon:SetScript("OnLeave", LockLeader_OnLeaveIcon);

		end

		ssIcon.warlockName = warlockName;

		if(nil == ssIcon.timerText) then
			ssIcon.timerText = ssIcon:CreateFontString(ssIconName .. i .. "timerText", "ARTWORK", "GameFontNormalSmall");
			ssIcon.timerText:SetPoint("Center", ssIcon, "Center", 0, 0);
			ssIcon.timerText:SetSize(65,20);
			ssIcon.timerText:SetTextColor(0.9, 0.9, 0.9, 1);
		end 
		
		ssIcon:SetNormalTexture(ssTextureEnabledPath);
		
		ssIcon:SetDisabledTexture(ssTextureDisabledPath);
		ssIcon.disabledtexture = ssIcon:GetDisabledTexture();	--get the actual texture object
		ssIcon.disabledtexture:SetVertexColor(0.95, 0.2, 0.2);	--and grey it out
		--ssIcon:disabledtexture:SetDesaturated(true);

		--Icons may have been hidden in a prior run of this column
		dd:Show();
		ssIcon:Show();
	end

	--there may be leftover icons if we've had warlocks leave
	j = i + 1;
	local dd = _G[ssDropDownName .. j];
	while (dd ~= nil) do
		dd:Hide();

		j = j+1;
		dd = _G[ssDropDownName .. j];
	end

	k = i + 1;
	local ssIcon = _G[ssIconName .. k];
	while (ssIcon ~= nil) do
		ssIcon:Hide();

		k = k+1;
		ssIcon = _G[ssIconName .. k];
	end
		

	local button = _G["LockLeaderAnnounce2"];
	if (button == nil) then
		button = CreateFrame("Button", "LockLeaderAnnounce2", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce Soulstone assignments to selected chat channels";
	button:SetSize(75,30);

	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											LLmsg("Hander: LockLeaderAnnounce2.OnClick()");
											LockLeader_AnnounceToChat("Leader Bot: Soulstone assignments", LOCKSS);
								end);

	if (warlockCount <= 0) then
		button:Hide();
	else
		button:Show();
	end

	return columnFrame;

end

-- ******************************************************************************
function LockLeader_OnEnterIcon(self)
	GameTooltip:SetOwner(self);
	GameTooltip:SetText(self:GetName(), nil, nil, nil, nil, true);
	GameTooltip:Show()
end

-- ********************************************************************************************
--closes tooltip when you mouse away from a header button
function LockLeader_OnLeaveIcon(self)
	GameTooltip:Hide();
end

-- ******************************************************************************
function LockLeader_SSDropDownInit(warlockName, dropDownMenu, teamMembers)
	LLmsg("Function: LockLeader_SSDropDownInit(" .. warlockName .. ")" );

	local classIcons =
	{
		["WARRIOR"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:0:64|t",
		["MAGE"]=		"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:0:64|t",
		["ROGUE"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:0:64|t",
		["DRUID"]=	 	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:0:64|t",
		["HUNTER"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:64:128|t",
		["SHAMAN"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:64:128:64:128|t",
		["PRIEST"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:128:196:64:128|t",
		["WARLOCK"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:196:256:64:128|t",
		["PALADIN"]=	"|TInterface\\WorldStateFrame\\ICONS-CLASSES:0:0:0:0:256:256:0:64:128:196|t",
	 }

	--lua won't save tables in the order I define them, so I have to force numeric indices to control the display order ><
	local classOrder = 
	{
			-- display resers first
			[1] = 
			{
				["class"]		= "Priest",
				["classCap"]	= "PRIEST"
			},

			[2] =
			{
				["class"]		= "Paladin",
				["classCap"]	= "PALADIN"
			},

			[3] =
			{
				["class"]		= "Shaman",
				["classCap"]	= "SHAMAN"
			},

			 --tanks and bres next
			 [4] =
			{
				["class"]		= "Warrior",
				["classCap"]	= "WARRIOR"
			},

			[5] =
			{
				["class"]		= "Druid",
				["classCap"]	= "DRUID"
			},

			--dps last
			[6] = 
			{
				["class"]		= "Hunter",
				["classCap"]	= "HUNTER"
			},

			[7] = 
			{
				["class"]		= "Mage",
				["classCap"]	= "MAGE"
			},

			[8] = 
			{
				["class"]		= "Rogue",
				["classCap"]	= "ROGUE"
			},

			[9] =
			{
				["class"]		= "Warlock",
				["classCap"]	= "WARLOCK"
			}
	}

	--create the dropdown row
	local rowInfo = UIDropDownMenu_CreateInfo(); --returns a re-usable, empty table to minimize memory usage.
	rowInfo.text = soulStoneReserve;
	rowInfo.value = soulStoneReserve;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = LockLeader_SSDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text ="   ----    ";
	rowInfo.isTitle = true;
	UIDropDownMenu_AddButton(rowInfo);

	-- walk through each class *in the order I want*, not the random LUA storage bs
	for i =1, 9 do
		displayClass = classOrder[i].class;
		displayClassCaps = classOrder[i].classCap;
		LLmsg("         Working on: " .. displayClass .. " - " .. displayClassCaps);

		--if this class has members, iterate through them. without this check we get LUA errors sometimes whining about nil arrays
		--passed to the for loop if we happen to forget to initialize the teamMembers[<class>] = {}
		if (teamMembers[displayClass]) then

			-- track if we actually found an entry or not. Because apparently sometimes empty, but not nil, classes still get here. ><
			local found = 0;

			for name, data in next, teamMembers[displayClass] do
				--we should legit have data if we get here
				found = 88;

				--create the dropdown row
				rowInfo = UIDropDownMenu_CreateInfo();
				rowInfo.text = name;
				rowInfo.value = name;
				rowInfo.owner  = dropDownMenu;
				rowInfo.icon ="Interface\\GLUES\\CHARACTERCREATE\\UI-CHARACTERCREATE-CLASSES";
				rowInfo.tCoordLeft, rowInfo.tCoordRight, rowInfo.tCoordTop, rowInfo.tCoordBottom  = unpack(CLASS_ICON_TCOORDS[displayClassCaps]);
				rowInfo.checked = false;
				rowInfo.arg1 = warlockName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
				rowInfo.func = LockLeader_SSDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
				UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to	
			end

			--if we really found data, and we're not the last class on the display list, add a spacer after our list
			if (found == 88 and displayClass ~= "Warlock") then
				--add a spacer after each class
				rowInfo = UIDropDownMenu_CreateInfo();
				rowInfo.text ="   ----    ";
				rowInfo.isTitle = true;
				UIDropDownMenu_AddButton(rowInfo);
			end

		end --end if (teamMembers[])

	end --end for(class)

end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function LockLeader_SSDropDownOnClick(dropDownButton, warlockName, arg2, checked)
	LLmsg("Function: Lockleader_SSDropDownOnClick(" .. warlockName .. ")" );

	local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock
	LockLeader_Assignments[LOCKSS][warlockName] = dropDownButton.value;

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.warlockName = warlockName;
	dropDownFrame.ssTargetName = dropDownButton.value;

	--update our custom label
	--local coverLabel = _G[dropDownFrame:GetName() .. "label"];
	local coverLabel = dropDownFrame.ddMyLabel;
	coverLabel:SetText(whiteColorCode .. dropDownFrame.ssTargetName);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", LOCKSS, warlockName, dropDownButton.value);
end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function LockLeader_BanishDropDownOnClick(dropDownButton, warlockName, banishTargetIcon, checked)
	LLmsg("Function: Lockleader_BanishDropDownOnClick(" .. warlockName .. ")" );
	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	LockLeader_Assignments[LOCKBANISH][warlockName] = banishTargetIcon;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode ..  dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", LOCKBANISH, warlockName, banishTargetIcon);
end

-- ******************************************************************************
function LockLeader_AddColumnCurseAssignments(teamMembers, warlockCount, forceDisable)
	LLmsg("Function: Lockleader_AddColumnCurseAssignments");

	local columnFrame = _G["LockLeaderColumn4"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "LockLeaderColumn4", LockLeaderFrameMax, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[3], "TopLeft", columnWidthBanish + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, warlockCount, columnWidthCurses)

	local colTitle = _G["LockLeaderColumn4Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("LockLeaderColumn4Title", "ARTWORK", LeaderBotColumnTitleFont);
	end
	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Curses");

	i = 0;
	for name, data in next, teamMembers["Warlock"] do
		i = i + 1;

		-- make the drop down menu for SS targets
		local dd = _G[curseDropDownName ..i];
		if (nil == dd) then
			dd = CreateFrame("Frame",curseDropDownName ..i, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd,columnWidthCurses - 35);

		UIDropDownMenu_EnableDropDown(dd);
		if (LeaderBotIsAssistant == false) then 
			UIDropDownMenu_DisableDropDown(dd);
		end 

		local topBuffer = 2;
		if (i > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (i -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() LockLeader_CurseDropDownInit(name, dd) end;
		dd.warlockName = name;

		--did the lock previously have an assignment? If so, re-apply it.
		local targetName = LockLeader_Assignments[LOCKCURSES][name];
		if (targetName) then
			--they previously had an assignment; re-apply it.
			dd.targetName = targetName;

			local spellID = CurseArray[targetName];
			if (spellID) then
				dd.spellTexture = LeaderBot_GetTextForSpellID( spellID);
			else
				--put something here so we don't get more fucking "attempt to concatenate a nil value" errors from LUA. ><
				dd.spellTexture ="";
			end

		else 
			--they have not had an assignment before, so assign them to RESERVE
			dd.targetName = LeaderBotNone;
			dd.spellTexture = ""
			LockLeader_Assignments[LOCKCURSES][name] = LeaderBotNone;			
		end 

		--element may have been hidden on a prior pass
		dd:Show();

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[curseDropDownName .. i .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", curseDropDownName ..i .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[curseDropDownName .. i .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(curseDropDownName .. i .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		ddMyLabel:SetText(whiteColorCode .. dd. spellTexture .. " " .. dd.targetName );
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		UIDropDownMenu_SetSelectedValue(dd, dd.targetName);
	end

	i = i + 1;
	local dd = _G[curseDropDownName..i];
	while (dd ~= nil) do
		dd:Hide();

		i = i + 1;
		dd = _G[curseDropDownName ..i];
	end

	local button = _G["LockLeaderAnnounce4"];
	if (nil == button) then
		button = CreateFrame("Button", "LockLeaderAnnounce4", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce curse assignments to selected chat channels";
	button:SetSize(75,30);
	if (forceDisable == true) then
		button:Disable();
	end
	button:SetScript("OnClick", function(self)
											LLmsg("Hander: LockLeaderAnnounce4.OnClick()");
											LockLeader_AnnounceToChat("Leader Bot: Curse assignments", LOCKCURSES);
								end);

	if (warlockCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame;
end

-- ******************************************************************************
function LockLeader_CurseDropDownInit(warlockName, dropDownMenu)
	LLmsg("Function: LockLeader_CurseDropDownInit( " .. warlockName .. ")" );
	local rowInfo = UIDropDownMenu_CreateInfo();

	rowInfo.text = LeaderBotNone ;
	rowInfo.value = LeaderBotNone;
	--rowInfo.keepShownOnClick = true;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = nil;
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	--on or about Nov 19, 2019, Blizzard disallowed printing spellIcons in chat via sendMessage.
	--so, all the .value rows below have to change to a more mundane, text value.

	rowInfo.text =  "Agony"
	rowInfo.value = "Agony"
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =GetSpellTexture(CurseOfAgonyID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBot_GetTextForSpellID(CurseOfAgonyID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  "Doom"
	rowInfo.value = "Doom"
	--rowInfo.keepShownOnClick = true;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =GetSpellTexture(CurseOfDoomID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBot_GetTextForSpellID(CurseOfDoomID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = "Elements"
	rowInfo.value = "Elements"
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = GetSpellTexture(CurseOfElementsID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 =  LeaderBot_GetTextForSpellID(CurseOfElementsID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  "Recklessness"
	rowInfo.value = "Recklessness"
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =GetSpellTexture(CurseOfRecklessnessID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 =  LeaderBot_GetTextForSpellID(CurseOfRecklessnessID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  "Shadow";
	rowInfo.value = "Shadow";
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = GetSpellTexture(CurseOfShadowID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 =   LeaderBot_GetTextForSpellID(CurseOfShadowID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  "Tongues";
	rowInfo.value = "Tongues";
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = GetSpellTexture(CurseOfTonguesID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBot_GetTextForSpellID(CurseOfTonguesID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text =  "Weakness";
	rowInfo.value = "Weakness";
	--rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = GetSpellTexture(CurseOfWeaknessID);
	rowInfo.checked = false;
	rowInfo.arg1 = warlockName;								-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBot_GetTextForSpellID(CurseOfWeaknessID);
	rowInfo.func = LockLeader_CurseDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

end

-- ******************************************************************************
function LockLeader_CurseDropDownOnClick(dropDownButton, warlockName, spellTexture, checked)
	LLmsg("Function: LockLeader_CurseDropDownOnClick(".. warlockName .. ")" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	LockLeader_Assignments[LOCKCURSES][warlockName] = dropDownButton.value;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. (spellTexture or "") .. " " .. dropDownButton.value );

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", LOCKCURSES, warlockName, dropDownButton.value);
end

-- ******************************************************************************
function LockLeader_SeduceDropDownOnClick(dropDownButton, warlockName, targetIcon, checked)
	LLmsg("Function: LockLeader_SeduceDropDownOnClick(".. warlockName .. ")" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);
	--UIDropDownMenu_SetText(dropDownButton.owner, "|T" .. "Interface\\TargetingFrame\\UI-RaidTargetingIcon_8" .. ":0:0:0:0:256:256:0:256:0:256|t");

	--update the assignment table for this warlock
	LockLeader_Assignments[LOCKSEDUCE][warlockName] = targetIcon;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Warlock", LOCKSEDUCE, warlockName, targetIcon);
end

--******************************************************************************************************
-- Command line parser
function LockLeader_SlashHandler(arg1)
	LLmsg("Function: LockLeader_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. " LockLeader by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. " Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (LL_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. " LockLeader debug messages disabled");
			LL_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. " LockLeader debug messages enabled");
			LL_Debug = 1;
		end
	end

end

--******************************************************************************************************
-- EventHandler
function LockLeaderFrame_OnEvent(self, event, ...)
	local arg1 = select(1, ...);
	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	LLmsg("Event: " .. event .. ", " .. arg1);

	local i, whichtarget;

	-- *****************************************************************************************
	-- lookup details of the event. The lookup method varies based on the even type
	if (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		local eventInfo = {CombatLogGetCurrentEventInfo()}
		--local timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, b_or_d = select(1, ...);
	
		local timestamp = eventInfo[1];
		local event = eventInfo[2];
		local sourceGuid = eventInfo[4];
		local sourceName = eventInfo[5];
		local destGuid = eventInfo[8];
		local destName = eventInfo[9];
		local spellName = eventInfo[13];

		--print (CombatLogGetCurrentEventInfo() );
		--don't alert on SS detection if we're not in a group.
		if (event == "SPELL_CAST_SUCCESS" and string.find(spellName, "Soulstone Res") and GetNumGroupMembers() > 0 )then 
			-- we found a Soulstone cast!!
			LLmsg("I detected that " .. sourceName .. " just cast SS on " .. destName);
			LeaderBotSyncMaster_SendSSmsg(timestamp, "CAST", LeaderBotSenderName, sourceName, destName);
			LockLeader_SetSoulStoneFound(sourceName, destName, timestamp);

		elseif (event == "UNIT_DIED") then
			--Somebody died. was it a ss target?

			local found = 0;
			for warlockName in next, LockLeader_Assignments[LOCKSS] do
				if (LockLeader_Assignments[LOCKSS][warlockName] == destName) then
					--a SS target died
					found = 1;
					break;
				end 
			end 

			if (found == 1) then 
				LLmsg ("I detected that " .. destName .. " died");
				LeaderBotSyncMaster_SendSSmsg(timestamp, "DIED", LeaderBotSenderName, "n/a", destName);		
				LockLeader_SetSoulStoneLost(destName, "DIED");	
			end 

		end 


	end 

end


--********************************************************************
--called by LeaderBotFrame_OnUpdate()
function LockLeaderFrame_UpdateSoulStoneIcon(elapsed)
	LLmsg("Function: LockLeaderFrame_UpdateSoulStoneIcon()");
	local i;
	
	--check all players for an active SS buff and update timers. Save the names of any members
	--where we find a SS
	--local ssStatus = LockLeader_GetSoulStoneStatus();

	--walk throught he pulldowns
	local i = 1;
	local ssIcon = _G[ssIconName .. i];
	local ssDropDown = _G[ssDropDownName ..i];



	while (ssDropDown ~= nil) do

		--get the dd value
		local ssTargetName = ssDropDown.ssTargetName;
		local warlockName = ssDropDown.warlockName;

		--if no warlock name is set, this is a first-time pass, before the user has set any
		--value in the SS target pulldown.There's nothing to do but hide the icon.
		if (warlockName == nil) then
			ssIcon:Hide();
			return;
		end

		if (ssDropDown:IsShown() == false) then 
			--this row is hidden, if we get here it means it used to be visible and then that
			--warlock left the group. Do nothing else.

		elseif (ssTargetName == nil)	then
			--there's no target set for this warlock. Hide the icon
			ssIcon:Hide();

		elseif (string.match(ssTargetName, "None")) then
			--We selected a target value, but that value is "None"
			ssIcon:Hide();

		elseif (ssStatusArray[ssTargetName] == nil) then 
			--this happens occasionally when a new char first falls through this check
			--this is expected to happen when LockLeader_SetSoulStoneLost() is called.
			ssIcon:Show();
			ssIcon:Disable();

			--I probably should show the alert pop ups here, but I really don't think it's a common
			--enough problem. Just skip it and we'll redo it on the next go round.

		elseif (ssStatusArray[ssTargetName]["ssFound"] == "NotFound" ) then
			--A target was selected in the dropdown but SS was not found on the target
			ssIcon:Show();
			ssIcon:Disable();

			--if there was a timer running, assuming it's now the cooldown timer
			local timerString  = ssIcon.timerText:GetText();
			if (timerString) then 
				
				--text to number converision and validity check
				local minuteIndex = string.find(timerString, "m");
				if (minuteIndex) then
					minuteIndex = minuteIndex -1;
				end 
				local timerVal = string.sub(timerString, 1, minuteIndex);
				
				if (timerVal + 0> 0) then
					if (minuteIndex) then 
						timerVal = timerVal * 60;
					end 

					local expirationTime = ssStatusArray[ssTargetName].expiration;
					local remainingTime = expirationTime - GetServerTime();

					if (remainingTime > 60) then 
						ssIcon.timerText:SetText( math.floor(remainingTime / 60) +1 .. "m");
					
					elseif (remainingTime > 0) then
						--show actual seconds under 1 minute
						ssIcon.timerText:SetText(math.floor(remainingTime) );
					
					else
						--timer has expired. clean it up.
						ssIcon.timerText:SetText("");

					end

				end 
			
			end 

		else 
		
			--ssStatusArray[ssTargetName] ~= nil, meaning SS was found on that target.
			local expirationTime = ssStatusArray[ssTargetName].expiration;
			local ssFound = ssStatusArray[ssTargetName].ssFound;

			ssIcon:Show();
			ssIcon:Enable();

			--if, by some miracle, we really do receive an expiration timer, then display it
			if (nil ~= expirationTime) then
				-- this target has an active timer

				local remainingTime = expirationTime - GetServerTime();
			
				if (remainingTime > 60) then
					--show time in minutes rather than seconds
					ssIcon.timerText:SetText( math.floor(remainingTime / 60) +1 .. "m");
				
				elseif (remainingTime > 0) then
					--show actual seconds under 1 minute
					ssIcon.timerText:SetText(math.floor(remainingTime) );
				
				else
					--timer has expired. clean it up.
					ssIcon.timerText:SetText("");
				end
			else
				ssIcon.timerText:SetText("");
			end

		end

		i = i+1;
		ssIcon = _G[ssIconName .. i];
		ssDropDown = _G[ssDropDownName ..i];

	end


end

-- ************************************************
function LockLeader_DataMigration()
	LLmsg("Function: LockLeader_DataMigration()" );

	if (LeaderBotOptions["LockLeaderDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		LockLeader_Assignments = {};
		LockLeader_Assignments[LOCKSS] = {};		--.warlockName, .warlockName..ssFound
		LockLeader_Assignments[LOCKBANISH] = {};
		LockLeader_Assignments[LOCKCURSES] = {};
		LockLeader_Assignments[LOCKSEDUCE] = {};
		LockLeader_Assignments[LOCKSSFOUNDON] = {};

		LeaderBotOptions["LockLeaderDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for index, assignment in next, assignmentIndices do

		if (LockLeader_Assignments[assignment] == nil) then
			LockLeader_Assignments[assignment] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, LockLeader_Assignments[assignment] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then  	
					LockLeader_Assignments[assignment][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************
function LockLeader_RosterChanged(newTeamMembers, newWarlockCount)
	LLmsg("Function: LockLeader_RosterChanged()");

--screw all of this diff'ing the two lists. when we get a new list, just update the display with the new
--names. doesn't matter who came or went. Then walk the lock assignments. If the lock is still in the list, 
--update their assignment onscreen. (if their assignment was a name, check if that name is still in the list). 
--If not, remove them from the assignment array.

--[[
	--TODO: At first, I thought this was a good idea to wipe out assignments when a player leaves the group. However,
	--I found that upon logging in, events fire 4-6 times that call _RosterChanged(). and sometimes those events fire
	--before the group is "formed" in Blizzard's mind. So, I'd get here, see group members "were gone" and remove them. 
	--then 3 events later find them again and add them to the table with blank assignments. So I'm taking this chunk out
	--for now in an attempt to preserve assignments after re-logging.

	--walk each assignment type (LOCKSS, LOCKBANISH, etc) and remove locks that are gone
	for assignmentType in next, LockLeader_Assignments do

		--walk each warlockname (key) in the assignment type
		for warlockNameKey, targetValue in next, LockLeader_Assignments[assignmentType] do 

			--if that warlock is in the new teamMember list, then they're still with us.
			--local existingWarlockName = newTeamMembers["Warlock"][warlockNameKey];
			local playerClass = UnitClass(warlockNameKey);
			if (playerClass == nil) then 
				--this warlock has left the building
				--remove this warlock from the assignment Index		
				LockLeader_Assignments[LOCKSS][warlockNameKey] = nil;
				LockLeader_Assignments[LOCKBANISH][warlockNameKey] = nil;
				LockLeader_Assignments[LOCKCURSES][warlockNameKey] = nil;
				LockLeader_Assignments[LOCKSEDUCE][warlockNameKey] = nil;
			end 

		end --for warlockName in assignment category

	end --for assignment category in assignment array
--]]

	--update the columns to account for locks that have come and gone
	LockLeader_AdjustMainPanelSize(newWarlockCount);

	columnFrames[1] = LeaderBotFrame_AddColumn1(newTeamMembers, newWarlockCount, "Warlock", LockLeaderFrameMax, "Warlocks");
	columnFrames[2] = LockLeader_AddColumnSoulstoneTargets(newTeamMembers, newWarlockCount, false);
	columnFrames[3], LockLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newWarlockCount, false,
						LockLeaderFrameMax, columnFrames[2],
						columnWidthBanish, "LockLeaderColumn3", "Banish",
						"LeaderBotWarlockMember", banishDropDownName,
						LockLeader_Assignments, LOCKBANISH,
						"LockLeaderAnnounce3", LockLeader_AnnounceToChat, "Leader Bot: Banish Assignments", 
						LockLeader_BanishDropDownOnClick,
						LeaderBotSingleSelect) ;
	columnFrames[4] = LockLeader_AddColumnCurseAssignments(newTeamMembers, newWarlockCount, false);
	columnFrames[5], LockLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newWarlockCount, false,
						LockLeaderFrameMax, columnFrames[4],
						columnWidthSeduce, "LockLeaderColumn5", "Seduce",
						"LeaderBotWarlockMember", seduceDropDownName,
						LockLeader_Assignments, LOCKSEDUCE,
						"LockLeaderAnnounce5", LockLeader_AnnounceToChat, "Leader Bot: Seduce Assignments", 
						LockLeader_SeduceDropDownOnClick,
						LeaderBotSingleSelect) ;

	--setup scrolling, if needed:
	if (newWarlockCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 1, columnFrames[1], "LeaderBotWarlockMember");
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 2, columnFrames[2], ssDropDownName, ssIconName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 3, columnFrames[3], banishDropDownName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 4, columnFrames[4], curseDropDownName);
		LeaderBot_RegisterColumnForScrolling(5, "Warlock", 5, columnFrames[5], seduceDropDownName);

		LeaderBot_SetupScrollBar("Warlock");
	end 

	local dd = _G["LockLeaderChatChannelDropDown"];
	if (LeaderBotIsAssistant == false) then
		LockLeader_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	else
		UIDropDownMenu_EnableDropDown(dd);
	end 

end

-- **********************************************************************************************************
function LockLeader_GetSoulStoneStatus()
	LLmsg("Function: LockLeader_GetSoulStoneStatus( )");

	--The original idea was to find SS on all members, and save the expirationTime reported by UnitBuff().
	--Then, OnUpdate would watch the expiration time and clear the SS when the expirationTime had passed.
	--In classic 1.13.2, though, expirationTimes are zero for all unitIDs except "player"). So, no we have to
	--manually clear the SS status each pass and rediscover if it's up or not.

	ssStatusArray = {};

	--walk each lock assignment, find the names of ss targets
	for key, value in next, LockLeader_Assignments[LOCKSS] do 

			--I prefer to keep the key, value syntax above to help remind me of the layout of LUA tables,
			--then re-assign them to friendly names. Sue me.

			local lockName = key;
			local ssTargetName = value;

			--the lock is guilty of not casting until proven innocent (SS is found on their target)
			ssStatusArray[ssTargetName] = {};			
			ssStatusArray[ssTargetName]["ssFound"] = "NotFound";
			LockLeader_Assignments[LOCKSSFOUNDON][lockName] = "NotFound";

			if (UnitClass(lockName) == nil) then
				--this lock is in the saved assignments but not in our current raid
				--this is a fake "continue" statement 

			elseif (ssTargetName == nil) then
				-- no assignments for this lock. 
				--this is a fake "continue" statement

			elseif (string.match(ssTargetName, "reserve") ) then
			--purposeful non-assignment to this lock (hold in reserve)
			--this is a fake "continue statement"

			else 

				--walk every buff on the target and see if they have a SS up or not. 
				local buffIndex = 1;
				local buffName, icon, count, debuffType, fullduration, expirationTime, source, 
						isStealable, namePlateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, 
						nameplateShowAll, timeMod = UnitBuff(ssTargetName, buffIndex);


				while (nil ~= buffName) do		

					if (string.match(buffName, "Soulstone")) then
						--well here's a another fun fact. If the target is too far away, UnitBuff still returns the buffName,
						--but other details, like source, become nil.

						--well this is swell. in Classic 1.13.2, expiration time is always 0 except when querying my own buffs
						ssStatusArray[ssTargetName]["expiration"] = expirationTime;
						ssStatusArray[ssTargetName]["ssFound"] = ssTargetName;

						LockLeader_Assignments[LOCKSSFOUNDON][lockName] = ssTargetName;

						--(wait for onUpdate to change the icons and display the timer)
						break;
					end
				
					buffIndex = buffIndex + 1;
					buffName, icon, count, debuffType, fullduration, expirationTime, source, 
							isStealable, namePlateShowPersonal, spellID, canApplyAura, isBossDebuff, castByPlayer, 
							nameplateShowAll, timeMod = UnitBuff(ssTargetName, buffIndex);

				end				
				
			end

	end

	return ssStatusArray;

end

-- ************************************************************************
function LockLeader_SetSoulStoneFound(warlockName, targetName, castTime)
	LLmsg("Function: LockLeader_SetSoulStoneFound()");

	if (ssStatusArray[targetName] == nil) then 
		ssStatusArray[targetName] = {};
	end 

	ssStatusArray[targetName]["expiration"] = castTime + 1800; -- 30 minutes * 60 seconds
	ssStatusArray[targetName]["ssFound"] = targetName;
	LockLeader_Assignments[LOCKSSFOUNDON][warlockName] = targetName;

end

-- ************************************************************************
function LockLeader_SetSoulStoneLost(targetName, reasonCode)
	LLmsg("Function: LockLeader_SetSoulStoneLost()");

	--there seems to be a case consisting of:
	--1. ss target is assigned
	--2. lock hasn't cast it yet, for whatever reason. Possibly legit holding it for a subsequent fight
	--3. ss target dies
	--4. so ssStatusArray does not yet include this target.
	--5. Crash here!
	
	--but we don't want to always force a re-initialization here if we don't need too because we would lose the expiration timer.
	if (ssStatusArray[targetName] == nil) then 
		ssStatusArray[targetName] = {};
		ssStatusArray[targetName]["expiration"] = nil;
	end 

	ssStatusArray[targetName]["ssFound"] = "NotFound";

	for warlockName, warlockTarget in next, LockLeader_Assignments[LOCKSSFOUNDON] do 

		--check if warlockName is in the group. If not, skip them as they're just a
		--saved assignment from a prior run
		local isInGroup = UnitClass(warlockName);

		if (LockLeader_Assignments[LOCKSSFOUNDON][warlockName] == targetName and isInGroup ) then 

			LockLeader_Assignments[LOCKSSFOUNDON][warlockName] = "NotFound";

			local anncText ;
			local popupText;

			if (reasonCode == "DIED") then 
				popupText = "Soulstone Target Died\n\n" .. warlockName .. "'s Soulstone target, " .. targetName .. ", appears to have died";
				anncText = "(Leader Bot) " .. warlockName .. ", your Soulstone target, " .. targetName .. ", appears to have died";
			else 
				popupText = "Soulstone Expired\n\n" .. warlockName .. "'s Soulstone is not detected on " .. targetName;
				anncText = "(Leader Bot) " .. warlockName .. ", I'm unable to detect Soulstone on " .. targetName .. ", please double check.";
			end 

			if (LeaderBotOptions["LockSSPopup"]) then
				--since 12/2019 Blizzard no longer allows me to send these helpful messages in a custom chat channel. Only /p, /g, /r or /t 
				--without user intervention.
				StaticPopupDialogs["LEADERBOT_SOULSTONE_NOT_DETECTED"] = {
							preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
							timeout = 30,
							whileDead = true,
							hideOnEscape = true,
							text = popupText,
								
							button1 = "Announce",
							OnAccept = function()
											LockLeader_AnnounceAlertToChat(anncText);
										end,	

							button2 = "Ignore",
							OnCancel = function()

										end,		
				};


				StaticPopup_Show("LEADERBOT_SOULSTONE_NOT_DETECTED");
			end

			break;

		end 
	end 

end

-- ******************************
function crashItNow()
	local a;
	print(a[4]);
end



