--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
	All code here is inside functions. Nothing runs when the .TOC loads, other than the
	high-level variable definitions. Execution beings in LeaderBot.lua:OnLoad()

--]]

local classicOrRetail;
local ML_Debug = 0;
local version = "1.13";

MageLeaderFrame = nil;

local MAGEINTELLECT_MULTISELECT = "Intellect_MultiSelect";
local MAGESHEEP = "Sheep"
local MAGEASSIGNMENTCOUNT = 2;

--Global variables to be saved
--mage leader assignments must be global to allow saving
MageLeader_Assignments = {};
MageLeader_Assignments[MAGEINTELLECT_MULTISELECT] = {};
MageLeader_Assignments[MAGESHEEP] = {};

local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local MageLeader_MsgChannel ={};

local columnWidth = 150;
local columnWidthInt = 150;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CurseOfAgonyID = 11713;
local CurseOfDoomID = 603;
local CurseOfTonguesID = 11719;
local CurseOfRecklessnessID = 7658;
local CurseOfWeaknessID = 7646;
local CurseOfElementsID = 11721;
local CurseOfShadowID = 17862;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local variablesLoaded = false;
local addHealerIndex = 0;
local addHealerTargetIndex = 0;

local intellectDropDownName = "MageLeaderIntellectDropDown";
local sheepTargetDropDownName = "MageLeaderSheepDropDown"
local mageDropDowns = {
						[MAGEINTELLECT_MULTISELECT] = intellectDropDownName,
						[MAGESHEEP] = sheepTargetDropDownName,
						};

local mageRaid = "Raid";

local group1 = "Group 1";
local group2 = "Group 2";
local group3 = "Group 3";
local group4 = "Group 4";
local group5 = "Group 5";
local group6 = "Group 6";
local group7 = "Group 7";
local group8 = "Group 8";
local group12 = "Groups 1 & 2";
local group34 = "Groups 3 & 4";
local group56 = "Groups 5 & 6";
local group78 = "Groups 7 & 8";

local colorCode = "|cffbbaaee ";
local whiteColorCode = "|cffffffff";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};
	
local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function MLmsg(msg)
	if (ML_Debug == 1 ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function MageLeaderFrame_OnLoad(parentFrame, teamMembers, mageCount)
	MLmsg("Function: MageLeaderFrame_OnLoad()");

	MageLeaderFrame = CreateFrame("Frame", "MageLeaderFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate");

	-- UI Commands
	SLASH_MageLeader1 = "/ml";
	SLASH_MageLeader2 = "/ML";

	SlashCmdList["MageLeader"] = MageLeader_SlashHandler;

	--MageLeaderFrame:RegisterEvent("VARIABLES_LOADED");
	--MageLeaderFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	--MageLeaderFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
     --

	ChatFrame1:AddMessage(colorCode .. "MageLeader Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's mageLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	MageLeader_InitMainFrame(teamMembers, mageCount);

	return (MageLeaderFrame);
end;

-- *************************************************************************************
function MageLeader_InitMainFrame(teamMembers, mageCount)
	MLmsg("Function: MageLeader_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove mages

	MageLeaderFrame:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 2);	
	MageLeader_AdjustMainPanelSize(mageCount);	

	MageLeaderFrame:SetBackdrop(mainBackdrop);
	MageLeaderFrame:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	MageLeaderFrame:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	MageLeaderFrame:EnableMouse(true);
	MageLeaderFrame:SetScript("OnUpdate", MageLeaderFrame_OnUpdate);
	MageLeaderFrame:SetScript("OnEvent", MageLeaderFrame_OnEvent);

	MageLeaderFrame:SetScript("OnShow", function(self)
						MLmsg("Function: MageLeaderFrame_OnShow() ");
						MLmsg("current panel = " .. LeaderBotOptions["currentPanel"] );

						if (LeaderBotOptions["currentPanel"] == "Mage") then 
							local parentFrame = MageLeaderFrame:GetParent();
							parentFrame:SetWidth(MageLeaderFrame:GetWidth() );
						end					
				end);

	local appLogo = LeaderBot_GetTextForClass("Mage");
	local appTitle = MageLeaderFrame:CreateFontString("MageLeaderAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", MageLeaderFrame, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Mage Leader " .. version);

	local infoText = MageLeaderFrame:CreateFontString("MageLeaderInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	--infoText:SetText(colorCode .. " ");

	columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, mageCount, "Mage", MageLeaderFrame, "Mages");
	columnFrames[2] = MageLeader_AddColumnIntellectTargets(teamMembers, mageCount, true);
	columnFrames[3], MageLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, mageCount, true, 
							MageLeaderFrame, columnFrames[2], 
							columnWidthInt, "MageLeaderColumn3", "Sheep Target", 
							"LeaderBotMageMember", sheepTargetDropDownName, 
							MageLeader_Assignments, MAGESHEEP, 
							"MageLeaderAnnounce3", MageLeader_AnnounceToChat, "Leader Bot: Sheep Targets",
							MageLeader_SheepDropDownOnClick,
							LeaderBotSingleSelect);

	if (mageCount > LeaderBotScrollSlots) then 
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 1, columnFrames[1], "LeaderBotMageMember");
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 2, columnFrames[2], intellectDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 3, columnFrames[3], sheepTargetDropDownName);

		LeaderBot_SetupScrollBar("Mage");
	end 

	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","MageLeaderChatChannelDropDown", MageLeaderFrame, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );
	dd:SetPoint("BottomLeft", MageLeaderFrame, "BottomLeft", -15, 0);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelList, MageLeader_ChannelDropDownOnClick) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();

	UIDropDownMenu_SetText(dd, "<Select channel>");
	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end
	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);

	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end
end

-- ******************************************************************************
--adjust the height as players join or leave
function MageLeader_AdjustMainPanelSize(mageCount)
	MLmsg("Function: MageLeader_AdjustMainPanelSize(count = " .. mageCount .. ")" );

	local numberOfRows = mageCount;
	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidth 
						+ columnWidth  --padding to make the frame big enough when only one column is shown.
						+ columnWidthInt
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	
	MageLeaderFrame:SetSize(math.max(math.max(requiredWidth, LeaderBotMinWidth), LeaderBotMinWidth), requiredHeight);
end

-- ******************************************************************************
function MageLeader_AddColumnIntellectTargets(teamMembers, mageCount, forceDisable)
	MLmsg("Function: MageLeader_AddColumnIntellectTargets()" );

	local columnFrame = _G["MageLeaderColumn2"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "MageLeaderColumn2", MageLeaderFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[1], "TopLeft", columnWidthInt + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, mageCount, columnWidthInt)

	local colTitle = _G["MageLeaderColumn2Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("MageLeaderColumn2Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", 0, 10);
	colTitle:SetText("Arcane Intellect");

	local addMageIntellectIndex = MageLeader_AddIntellectTargetsToColumn(columnFrame, "Mage", teamMembers["Mage"]);

	--there may be leftover icons if we've had players leave
	addMageIntellectIndex = addMageIntellectIndex + 1;
	local dd = _G[intellectDropDownName .. addMageIntellectIndex];
	while (dd ~= nil) do
		dd:Hide();

		addMageIntellectIndex = addMageIntellectIndex+1;
		dd = _G[intellectDropDownName .. addMageIntellectIndex];
	end
		
	local button = _G["MageLeaderAnnounce2"];
	if (button == nil) then
		button = CreateFrame("Button", "MageLeaderAnnounce2", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce Arcane Intellect assignments to selected chat channels";
	button:SetSize(75,30);

	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											MLmsg("Hander: MageLeaderAnnounce2.OnClick()");
											MageLeader_AnnounceToChat("Leader Bot: Arcane Intellect assignments", MAGEINTELLECT_MULTISELECT);
								end);

	if (mageCount <= 0) then
		button:Hide();
	else
		button:Show();		
	end

	return columnFrame;

end
 -- *****************************************************************************
 function MageLeader_AddIntellectTargetsToColumn(columnFrame, class, mageTeamMembers)
	MLmsg("Function: MageLeader_AddIntellectTargetsToColumn(" .. class .. ")" )

	local addIndex = 0;
	for mageName, data in next, mageTeamMembers do
		MLmsg("     adding dropdown for mage: " .. mageName);

		addIndex = addIndex + 1;

		-- make the drop down menu for heal targets
		local dd = _G[intellectDropDownName .. addIndex];
		if (dd == nil) then
			dd = CreateFrame("Frame", intellectDropDownName .. addIndex, columnFrame, "UIDropDownMenuTemplate");
		end

		UIDropDownMenu_SetWidth(dd, columnWidthInt - 35);
		UIDropDownMenu_EnableDropDown(dd);

		local topBuffer = 2;
		if (addIndex > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (addIndex -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() MageLeader_IntellectDropDownInit(mageName, dd) end;
		dd.mageName = mageName;

		--Icons may have been hidden in a prior run of this column
		dd:Show();

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[intellectDropDownName .. addIndex .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", intellectDropDownName .. addIndex .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[intellectDropDownName .. addIndex .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(intellectDropDownName .. addIndex .. "label", "ARTWORK", LeaderBotDropDownLabelFont);
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		--ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );
		
		LeaderBot_InitNewMultiSelectPlayer(mageName, MageLeader_Assignments, MAGEINTELLECT_MULTISELECT);
		local labelText = LeaderBot_UpdateMultiselectDropDownLabel(MageLeader_Assignments, MAGEINTELLECT_MULTISELECT, mageName);
		ddMyLabel:SetText(whiteColorCode .. labelText);

		--DO NOT CALL THIS FOR MULTISELECT DROPDOWNS
		--UIDropDownMenu_SetSelectedValue(dd, dd.intellectTarget);

	end

	return addIndex;

 end

-- ******************************************************************************
function MageLeader_IntellectDropDownInit(mageName, dropDownMenu)
	MLmsg("Function: MageLeader_IntellectDropDownInit(" .. mageName .. ")" );

	--create the dropdown row
	local rowInfo = UIDropDownMenu_CreateInfo();

	rowInfo.text = "Multi-Select:";
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;
	rowInfo.isTitle = true;		
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo = UIDropDownMenu_CreateInfo();

	if (LeaderBotIsAssistant) then 
		rowInfo.isTitle = false;
	else
		rowInfo.isTitle = true;
	end 

	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][LeaderBotNone] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][LeaderBotNone] = true;
	end 

	rowInfo.text = LeaderBotNone;
	rowInfo.value = LeaderBotNone;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][LeaderBotNone];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group1] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group1] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group1;
	rowInfo.value = group1;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group1];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group2] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group2] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group2;
	rowInfo.value = group2;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group2];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group3] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group3] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group3;
	rowInfo.value = group3;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group3];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group4] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group4] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group4;
	rowInfo.value = group4;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group4];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group5] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group5] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group5;
	rowInfo.value = group5;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group5];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group6] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group6] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group6;
	rowInfo.value = group6;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group6];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group7] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group7] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group7;
	rowInfo.value = group7;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group7];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to


	if (MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group8] == nil) then
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group8] = false;
	end 

	local rowInfo = UIDropDownMenu_CreateInfo();
	rowInfo.text = group8;
	rowInfo.value = group8;
	rowInfo.isNotRadio = true;
	rowInfo.keepShownOnClick = true;	
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon =nil
	rowInfo.checked = MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][group8];
	rowInfo.arg1 = mageName;						-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.func = MageLeader_IntellectDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);					--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function MageLeader_IntellectDropDownOnClick(dropDownButton, arg1, arg2, checked)
	MLmsg("Function: MageLeader_IntellectDropDownOnClick(" .. arg1 .. ")" );

	local dropDownFrame = dropDownButton.owner;

	--mark this as the selected item in the parent dropDown list. DO NOT CALL THIS FOR MULTISELECT
	--UIDropDownMenu_SetSelectedValue(dropDownFrame, dropDownButton.value);

	--update the assignment table for this warlock
	local mageName = arg1;

	MageLeader_Assignments[MAGEINTELLECT_MULTISELECT][mageName][dropDownButton.value] = UIDropDownMenuButton_GetChecked(dropDownButton);

	--add a custom variable to the dropdown frame so we can figure out who it belongs to later.
	--Actually, the UIDropDownMenu functions are such a piece of shit and unable to get/set values properly. So just save them all 
	--in custom vars for easy access
	dropDownFrame.mageName = mageName;
	dropDownFrame.intellectTarget = dropDownButton.value;

	--dropDownFrame.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);
	local labelText = LeaderBot_UpdateMultiselectDropDownLabel(MageLeader_Assignments, MAGEINTELLECT_MULTISELECT, mageName);
	dropDownFrame.ddMyLabel:SetText(whiteColorCode .. labelText);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Mage", MAGEINTELLECT_MULTISELECT, mageName, dropDownButton.value, UIDropDownMenuButton_GetChecked(dropDownButton));	
end

-- *********************************************************************************
function MageLeader_SheepDropDownOnClick(dropDownButton, arg1, arg2, checked)
	MLmsg("Function: MageLeader_SheepDropDownOnClick(" .. arg1 .. ")" );

		--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	local mageName = arg1;
	local sheepTargetIcon = arg2;

	MageLeader_Assignments[MAGESHEEP][mageName] = sheepTargetIcon;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Mage", MAGESHEEP, mageName, sheepTargetIcon);

end

--******************************************************************************************************
-- Command line parser
function MageLeader_SlashHandler(arg1)
	MLmsg("Function: MageLeader_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. "MageLeader by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. "Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (ML_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. "MageLeader debug messages disabled");
			ML_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. "MageLeader debug messages enabled");
			ML_Debug = 1;
		end
	end

end

--******************************************************************************************************
-- EventHandler
function MageLeaderFrame_OnEvent(self, event, ...)
	local arg1 = select(1, ...);
	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	MLmsg("Function: MageLeaderFrame_OnEvent(Event: " .. event .. ", " .. arg1 .. ")" );


	-- *****************************************************************************************
	if (event == "VARIABLES_LOADED") then
			MLmsg ("     VARIABLES LOADED HANDLER");

			--used to prevent the OnShow() handler from trying to render a list before we have names
			--loaded.
			variablesLoaded = true;

			MLmsg("     preparing to run Update Display()");

	
	-- *****************************************************************************************
	-- lookup details of the event. The lookup method varies based on the even type
	elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		local timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, b_or_d = select(1, ...);
		MLmsg("event = " .. timestamp);


	-- ******************************************************************************************
	elseif (ttype == "SPELL_AURA_REMOVED" and dstGUID == UnitGUID("player") ) then
		--NightLightFrame_RemoveTimer(spellid)
	end
end

-- ************************************************
function MageLeader_DataMigration()
	MLmsg("Function: MageLeader_DataMigration()" );

	if (LeaderBotOptions["MageLeaderDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		MageLeader_Assignments = {};
		MageLeader_Assignments[MAGEINTELLECT_MULTISELECT] = {};		
		MageLeader_Assignments[MAGESHEEP] = {};

		LeaderBotOptions["MageLeaderDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for assignment = 1, MAGEASSIGNMENTCOUNT do
		if (MageLeader_Assignments[assignment] == nil) then
			MageLeader_Assignments[assignment] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, MageLeader_Assignments[assignment] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then  	
					MageLeader_Assignments[assignment][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************
function MageLeader_RosterChanged(newTeamMembers, newMageCount)
	MLmsg("Function: MageLeader_RosterChanged()");

--[[
	--TODO: At first, I thought this was a good idea to wipe out assignments when a player leaves the group. However,
	--I found that upon logging in, events fire 4-6 times that call _RosterChanged(). and sometimes those events fire
	--before the group is "formed" in Blizzard's mind. So, I'd get here, see group members "were gone" and remove them. 
	--then 3 events later find them again and add them to the table with blank assignments. So I'm taking this chunk out
	--for now in an attempt to preserve assignments after re-logging.

	-- walk each assignment type (HEAL, FORTITUDE, ETC)
	for assignmentType in next, MageLeader_Assignments do
	
		-- walk each mage in that assignment type
		for mageNameKey, targetValue in next, MageLeader_Assignments[assignmentType] do

			--local found = LeaderBotFrame_IsMemberOfTeam(mageNameKey, newTeamMembers);
			local playerClass = UnitClass(mageNameKey);
			if (playerClass == nil) then
				--this mage has left the building
				--remove their assignments
				MageLeader_Assignments[assignmentType][mageNameKey] = nil;		
			end 

		end --for mageNameKey

	end --for assignmentType
--]]

	MageLeader_AdjustMainPanelSize(newMageCount);
	columnFrames[1] = LeaderBotFrame_AddColumn1(newTeamMembers, newMageCount, "Mage", MageLeaderFrame, "Mages");
	columnFrames[2] = MageLeader_AddColumnIntellectTargets(newTeamMembers, newMageCount, false);
	columnFrames[3], MageLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newMageCount, true, 
							MageLeaderFrame, columnFrames[2], 
							columnWidthInt, "MageLeaderColumn3", "Sheep Target", 
							"LeaderBotMageMember", sheepTargetDropDownName, 
							MageLeader_Assignments, MAGESHEEP, 
							"MageLeaderAnnounce3", MageLeader_AnnounceToChat, "Leader Bot: Sheep Targets",
							MageLeader_SheepDropDownOnClick,
							LeaderBotSingleSelect);


	if (newMageCount > LeaderBotScrollSlots) then 
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 1, columnFrames[1], "LeaderBotMageMember");
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 2, columnFrames[2], intellectDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "Mage", 3, columnFrames[3], sheepTargetDropDownName);

		LeaderBot_SetupScrollBar("Mage");
	end 

	local dd = _G["MageLeaderChatChannelDropDown"];
	if (LeaderBotIsAssistant == false) then
		MageLeader_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	else
		UIDropDownMenu_EnableDropDown(dd);
	end  
end

-- ******************************************************************************
function MageLeader_AnnounceButtonEnable(state)
	MLmsg("FUnction: MageLeader_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["MageLeaderAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["MageLeaderAnnounce" .. i];
	end 

end

-- ******************************************************************************
function MageLeader_AnnounceToChat(title, assignmentIndex)
	MLmsg("Function: MageLeader_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	local chanType = MageLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = MageLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = MageLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot -Mage Leader: No chat channel selected");
		return;
	end

	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

	if (assignmentIndex == MAGEINTELLECT_MULTISELECT) then 

		for playerName, playerArray in next, MageLeader_Assignments[assignmentIndex] do

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(playerName);
			if (isInGroup) then 

				local msg = "";
				local comma = "";

				for assignmentName, assignmentValue in next, playerArray do 

					if (assignmentName ~= LeaderBotNone and assignmentValue == true) then 
						msg = assignmentName .. comma .. msg;
						comma = ", ";
					end

				end 

				if (msg ~= "") then 
					msg = playerName .. ": " .. msg;
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end

					--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
					--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
					--sync everything being announced.
					if (string.find(assignmentIndex, "MultiSelect") ) then 
						for key_TargetName, val_CheckMarkStatus in next, MageLeader_Assignments[assignmentIndex][playerName] do
							LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Mage", assignmentIndex, playerName, key_TargetName, val_CheckMarkStatus);
						end
					else 
						LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "Mage", assignmentIndex, playerName, MageLeader_Assignments[assignmentIndex][playerName]);
					end 

				end
			end

		end 
	else 

		for assignmentKey, assignmentValue in next, MageLeader_Assignments[assignmentIndex] do
			local name, target;
			local msg = nil;

			name = assignmentKey;
			targetName = assignmentValue;

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(name);
			if (isInGroup) then 

				--skip players assigned to LeaderBotNone
				if (targetName ~= LeaderBotNone ) then 
					msg = "     " .. name .. ": " .. targetName .. " " ;

					--if (chanType == "RAID") then
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end
				end 

			end 

		end -- end for loop

	end -- end if MULTISELECT

	if (chanNumber == nil) then
		--Party, raid
		SendChatMessage("-------------------------------------------", chanType);
	else
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
								
end

-- ******************************************************************************
function MageLeader_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	MLmsg("MageLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	MageLeader_MsgChannel[LOCKCHANTYPE] = arg1;
	MageLeader_MsgChannel[LOCKCHANLANG] = nil;
	MageLeader_MsgChannel[LOCKCHANNUMB] = arg2;

	MageLeader_AnnounceButtonEnable(1);
	
end

