--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
	unversioned
--]]

local LeadeBotSmallFont = "GameFontWhiteSmall"

local LeaderBotSmall = 0;
local LeaderBotColumnTitleFontSmall = "GameFontWhiteSmall";
local LeaderBotPlayerNameFontSmall  = "GameFontWhiteSmall";
local LeaderBotDropDownLabelFontSmall  = "GameFontWhiteSmall";
local LeaderBotPlayerHeightSmall  = 28;
local LeaderBotHeaderHeightSmall  = 60;
local LeaderBotFooterHeightSmall  = 35;
local LeaderBotBorderPaddingSmall  = 1;
local LeaderBotInfoPaddingSmall = -2;

local LeaderBotLarge = 1;
local LeaderBotColumnTitleFontLarge = "GameFontNormal";
local LeaderBotPlayerNameFontLarge = "GameFontNormal";
local LeaderBotDropDownLabelFontLarge = "GameFontNormal";
local LeaderBotPlayerHeightLarge = 32;
local LeaderBotHeaderHeightLarge = 80;
local LeaderBotFooterHeightLarge = 40;
local LeaderBotBorderPaddingLarge = 2;
local LeaderBotInfoPaddingLarge = -10;

--all global variables so they can a) be saved and b) be seen by the other modules.
LeaderBotColumnTitleFont = LeaderBotColumnTitleFontLarge;
LeaderBotPlayerNameFont = LeaderBotPlayerNameFontLarge;
LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontLarge;
LeaderBotplayerHeight = LeaderBotPlayerHeightLarge;
LeaderBotheaderHeight = LeaderBotHeaderHeightLarge;
LeaderBotfooterHeight = LeaderBotFooterHeightLarge;  
LeaderBotborderPadding = LeaderBotBorderPaddingLarge;
LeaderBotInfoPadding = LeaderBotInfoPaddingLarge;

LeaderBot_Footprint = LeaderBotLarge;
LeaderBotOptionsPanel = nil;

LeaderBotOptions = {
						["SyncMode"] 		= true,
						["5rows"]			= true;

						["LockLeader"] 		= true,
						["HealerLeader"]	= true,
						["TankLeader"]		= true,
						["MageLeader"] 		= true,
						["HunterLeader"]	= true,

						["LockSSWhisper"] 	= true,
						["LockSSPopup"] 	= true,
						["LockTargetPopup"] = true,

						["HunterLeftPopup"] = true;
						["TankleftPopup"] = true;

};

local O_Debug = 0;
local colorCode = "|cff999900 ";
local whiteColorCode = "|cffffffff"; 
local greyColorCode = "|cffaaaaaa";
local blueColorCode = "|cff11BBFF";

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};


 -- *****************************************************************************************************
function Omsg(msg)
	if (O_Debug == 1) then
		ChatFrame1:AddMessage( colorCode .. msg);
	end
end

--***************************************************************************************************

-- function LeaderBotOptions_SetupOptions()

local i, j;
local dd, ts, fs, cb, button;

--LeaderBotOptionsPanel = CreateFrame("FRAME","LeaderBot_OptionsFrame");  --wow classic
LeaderBotOptionsPanel = CreateFrame("FRAME", "LeaderBot_OptionsFrame", UIParent, BackdropTemplateMixin and "BackdropTemplate" );	-- tbc classic
LeaderBotOptionsPanel.name = "LeaderBot";
LeaderBotOptionsPanel:SetBackdrop(columnBackdrop);
LeaderBotOptionsPanel:SetBackdropColor(0.1, 0.1, 0.1, 1);

InterfaceOptions_AddCategory(LeaderBotOptionsPanel);

-- set the title text
ts = LeaderBotOptionsPanel:CreateFontString("LeaderBot_OptionsTitle","ARTWORK","GameFontNormalLarge");
ts:SetPoint("TOPLEFT",16,-16);
ts:SetText("Leader Bot Options");

--reloadUI button in upper right corner
local resetButton = CreateFrame("Button", "Reset", LeaderBotOptionsPanel, "OptionsButtonTemplate");
resetButton:SetPoint("TopRight", -30, -16);
resetButton:SetText("Reset");
resetButton.tooltipText = "Resets all options to defaults, clears all saved assignments and performs a " .. blueColorCode .. " reloads UI";

--Why the fuck does the "short" version work in LeaderBot.lua but not here??
resetButton:SetScript("OnEnter", LeaderBotOptionsPanel_OnEnterButton);
resetButton:SetScript("OnLeave", LeaderBotOptionsPanel_OnLeaveButton);

resetButton:SetScript("OnEnter", function(self)
									LeaderBotFrame_OnEnterButton(self);
								end);
resetButton:SetScript("OnLeave", function(self)
									LeaderBotFrame_OnLeaveButton(self);
								end);


resetButton:SetScript("OnClick", function(self)
									LeaderBotOptions["LockLeaderDataModelVersion"] = -1;
									LeaderBotOptions["HealerLeaderDataModelVersion"] = -1;
									LeaderBotOptions["TankLeaderDataModelVersion"] = -1;
									LeaderBotOptions["MageLeaderDataModelVersion"] =-1;
									LeaderBotOptions["HunterLeaderDataModelVersion"] = -1;
									LeaderBotOptions["CrowdControllerDataModelVersion"] = -1;
									LeaderBotOptions["OptionsModelVersion"] = -1;
									C_UI.Reload();
							end);

--reloadUI button in upper right corner
local button = CreateFrame("Button", "Reload UI", LeaderBotOptionsPanel, "OptionsButtonTemplate");
button:SetPoint("Top", resetButton, "Bottom", 0, -4);
button:SetText("Reload UI");
button:SetScript("OnClick", function(self)
									C_UI.Reload();
							end);

-- *****************************************************************************
-- Leader Bot Options
local ts2 = LeaderBotOptionsPanel:CreateFontString("LeaderBot_LockLeaderReloadWarning","ARTWORK","GameFontWhiteSmall");
ts2:SetPoint("TOPLEFT", ts, "BOTTOMLEFT", 0, -5);
ts2:SetText("Options in " .. blueColorCode .. "Blue " .. whiteColorCode .. "require reloading UI:");

local cbLockLeader = CreateFrame("CheckButton", "LeaderBot_LockLeaderCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbLockLeader:SetPoint("TOPLEFT", ts2, "BOTTOMLEFT", 0, -4);
getglobal(cbLockLeader:GetName() .. "Text"):SetText(blueColorCode .. "Lock Leader");
cbLockLeader.tooltipText = blueColorCode .. "Requires reloading UI";
cbLockLeader:SetScript("OnClick", function()
									if (cbLockLeader:GetChecked() ) then
										LeaderBotOptions["LockLeader"] = true;
									else
										LeaderBotOptions["LockLeader"] = false;
									end 
								end);

local cbHealerLeader = CreateFrame("CheckButton", "LeaderBot_HealerLeaderCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbHealerLeader:SetPoint("TOPLEFT",cbLockLeader,"BOTTOMLEFT", 0, 0);
getglobal(cbHealerLeader:GetName() .. "Text"):SetText(blueColorCode .. "Healer Leader");
cbHealerLeader.tooltipText = blueColorCode .. "Requires reloading UI";
cbHealerLeader:SetScript("OnClick", function()
									if (cbHealerLeader:GetChecked() ) then
										LeaderBotOptions["HealerLeader"] = true;
									else
										LeaderBotOptions["HealerLeader"] = false;
									end 
								end);

local cbTankLeader = CreateFrame("CheckButton", "LeaderBot_TankLeaderCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbTankLeader:SetPoint("LEFT",cbLockLeader,"RIGHT", 100, 0);
getglobal(cbTankLeader:GetName() .. "Text"):SetText(blueColorCode .. "Tank Leader");
cbTankLeader.tooltipText = blueColorCode .. "Requires reloading UI";
cbTankLeader:SetScript("OnClick", function()
									if (cbTankLeader:GetChecked() ) then
										LeaderBotOptions["TankLeader"] = true;
									else
										LeaderBotOptions["TankLeader"] = false;
									end 
								end);

local cbMageLeader = CreateFrame("CheckButton", "LeaderBot_MageLeaderCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbMageLeader:SetPoint("TOPLEFT", cbTankLeader,"BOTTOMLEFT", 0, 0);
getglobal(cbMageLeader:GetName() .. "Text"):SetText(blueColorCode .. "Mage Leader");
cbMageLeader.tooltipText = blueColorCode .. "Requires reloading UI";
cbMageLeader:SetScript("OnClick", function()
									if (cbMageLeader:GetChecked() ) then
										LeaderBotOptions["MageLeader"] = true;
									else
										LeaderBotOptions["MageLeader"] = false;
									end 
								end);

local cbHunterLeader = CreateFrame("CheckButton", "LeaderBot_HunterLeaderCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbHunterLeader:SetPoint("LEFT",cbTankLeader,"RIGHT", 100, 0);
getglobal(cbHunterLeader:GetName() .. "Text"):SetText(blueColorCode .. "Hunter Leader");
cbHunterLeader.tooltipText = blueColorCode .. "Requires reloading UI";
cbHunterLeader:SetScript("OnClick", function()
									if (cbHunterLeader:GetChecked() ) then
										LeaderBotOptions["HunterLeader"] = true;
									else
										LeaderBotOptions["HunterLeader"] = false;
									end 
								end);


local cbCrowdController = CreateFrame("CheckButton", "LeaderBot_CrowdControllerCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbCrowdController:SetPoint("TOPLEFT", cbHunterLeader,"BOTTOMLEFT", 0, 0);
getglobal(cbCrowdController:GetName() .. "Text"):SetText(blueColorCode .. "Crowd Controller");
cbCrowdController.tooltipText = blueColorCode .. "Requires reloading UI";
cbCrowdController:SetScript("OnClick", function()
									if (cbCrowdController:GetChecked() ) then
										LeaderBotOptions["CrowdController"] = true;
									else
										LeaderBotOptions["CrowdController"] = false;
									end 
								end);


local cbSync = CreateFrame("CheckButton", "LeaderBot_SyncCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbSync:SetPoint("TOPLEFT",cbHealerLeader,"BOTTOMLEFT", 0, -15);
getglobal(cbSync:GetName() .. "Text"):SetText("Sync mode");
cbSync.tooltipText = "Share and receive updates from other Leader Bot users in your raid";
cbSync:SetScript("OnClick", function()
									if (cbSync:GetChecked() ) then
										LeaderBotOptions["SyncMode"] = true;
										LeaderBotSyncMaster_OnInit();								
									else
										LeaderBotOptions["SyncMode"] = false;
										LeaderBotSyncMaster_Disable();	
									end 

									LeaderBot_EnableFullSyncButton(LeaderBotOptions["SyncMode"]);	

								end);


local cbSyncDebug = CreateFrame("CheckButton", "LeaderBot_SyncDebugCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbSyncDebug:SetPoint("LEFT",cbSync,"RIGHT", 105, 0);
getglobal(cbSyncDebug:GetName() .. "Text"):SetText("Sync mode debug messages");
cbSyncDebug.tooltipText = "Only click this if you know what you're doing";
cbSyncDebug:SetScript("OnClick", function()
									if (cbSyncDebug:GetChecked() ) then
										LeaderBotOptions["SyncModeDebug"] = true;
										LeaderBotSyncMaster_SetDebug(true)
									else
										LeaderBotOptions["SyncModeDebug"] = false;
										LeaderBotSyncMaster_SetDebug(false)
									end 

								end);

-- checkboxes to select between large and small footprints
local cbSmall = CreateFrame("CheckButton", "LeaderBot_FootprintCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbSmall:SetPoint("TOPLEFT", cbSync,"BOTTOMLEFT", 0, 0);
cbSmall.tooltipText = "Reduce the font size of each row" .. blueColorCode .. " Requires reloading UI";
getglobal(cbSmall:GetName() .. "Text"):SetText(blueColorCode .. "Compressed Display (requires reloading UI)");

cbSmall:SetScript("OnClick", function()
										if (cbSmall:GetChecked() ) then
											LeaderBot_Footprint = LeaderBotSmall;
											LeaderBotColumnTitleFont = LeaderBotColumnTitleFontSmall;
											LeaderBotPlayerNameFont = LeaderBotPlayerNameFontSmall;
											LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontSmall;
											LeaderBotplayerHeight = LeaderBotPlayerHeightSmall;
											LeaderBotheaderHeight = LeaderBotHeaderHeightSmall;
											LeaderBotfooterHeight = LeaderBotFooterHeightSmall;  
											LeaderBotborderPadding = LeaderBotBorderPaddingSmall;
											LeaderBotInfoPadding = LeaderBotInfoPaddingSmall;

										else 
											LeaderBot_Footprint = LeaderBotLarge;
											LeaderBotColumnTitleFont = LeaderBotColumnTitleFontLarge;
											LeaderBotPlayerNameFont = LeaderBotPlayerNameFontLarge;
											LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontLarge;
											LeaderBotplayerHeight = LeaderBotPlayerHeightLarge;
											LeaderBotheaderHeight = LeaderBotHeaderHeightLarge;
											LeaderBotfooterHeight = LeaderBotFooterHeightLarge;  
											LeaderBotborderPadding = LeaderBotBorderPaddingLarge;
											LeaderBotInfoPadding = LeaderBotInfoPaddingLarge;

										end
								end);

local cb5Rows = CreateFrame("CheckButton", "LeaderBot_5RowsCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cb5Rows:SetPoint("TOPLEFT", cbSmall,"BOTTOMLEFT", 0, 0);
getglobal(cb5Rows:GetName() .. "Text"):SetText(blueColorCode .. "Short Window Height (requires reloading UI)");
cb5Rows.tooltipText = "Display only 5 rows instead of the default 10" .. blueColorCode .. " Requires reloading UI";
cb5Rows:SetScript("OnClick", function()
									if (cb5Rows:GetChecked() ) then
										LeaderBotOptions["5rows"] = true;
										LeaderBotScrollSlots = 5;
									else
										LeaderBotOptions["5rows"] = false;
										LeaderBotScrollSlots = 10;
									end 

								end);

local cbNoLead = CreateFrame("CheckButton", "LeaderBot_noLeadCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbNoLead:SetPoint("TOPLEFT", cb5Rows,"BOTTOMLEFT", 0, 0);
getglobal(cbNoLead:GetName() .. "Text"):SetText("Display a popup when the page is read-only");
cbNoLead.tooltipText = "Display alert when the page is read-only because you are not a raid leader or assistant";
cbNoLead:SetScript("OnClick", function()
									if (cbNoLead:GetChecked() ) then
										LeaderBotOptions["noLead"] = true;
									else
										LeaderBotOptions["noLead"] = false;
									end 

								end);

-- **********************************************************
--Lock Leader Options
ts = LeaderBotOptionsPanel:CreateFontString("LeaderBot_LockLeaderOptionsTitle","ARTWORK","GameFontWhiteSmall");
ts:SetPoint("TOPLEFT", cbNoLead, 0, -50);
ts:SetText("Lock Leader Options");


local cbLockWhisper = CreateFrame("CheckButton", "LeaderBot_LockWhisperCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbLockWhisper:SetPoint("TOPLEFT",ts,"BOTTOMLEFT", 0, -4);
getglobal(cbLockWhisper:GetName() .. "Text"):SetText(greyColorCode .. "Whisper warlocks when unable to detect their Soulstone");
cbLockWhisper:SetScript("OnClick", function()
									if (cbLockWhisper:GetChecked() ) then
										LeaderBotOptions["LockSSWhisper"] = true;
									else
										LeaderBotOptions["LockSSWhisper"] = false;
									end 

								end);
cbLockWhisper:SetEnabled(false);


local cbLockOfflinePopup = CreateFrame("CheckButton", "LeaderBot_LockOfflinePopupCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbLockOfflinePopup:SetPoint("TOPLEFT",cbLockWhisper,"BOTTOMLEFT", 0, 0);
getglobal(cbLockOfflinePopup:GetName() .. "Text"):SetText("Display a popup when a warlock goes offline");
cbLockOfflinePopup.tooltipText = "Display a popup when a warlock goes offline";
cbLockOfflinePopup:SetScript("OnClick", function()
									if (cbLockOfflinePopup:GetChecked() ) then
										LeaderBotOptions["LockOfflinePopup"] = true;
									else
										LeaderBotOptions["LockOfflinePopup"] = false;
									end 

								end);


local cbLockPopup = CreateFrame("CheckButton", "LeaderBot_LockPopupCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbLockPopup:SetPoint("TOPLEFT",cbLockOfflinePopup,"BOTTOMLEFT", 0, 0);
getglobal(cbLockPopup:GetName() .. "Text"):SetText("Display a popup when unable to detect a warlock's Soulstone");
cbLockPopup:SetScript("OnClick", function()
									if (cbLockPopup:GetChecked() ) then
										LeaderBotOptions["LockSSPopup"] = true;
									else
										LeaderBotOptions["LockSSPopup"] = false;
									end 

								end);



local cbLockTargetLeftPopup = CreateFrame("CheckButton", "LeaderBot_LockTargetLeftPopupCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
cbLockTargetLeftPopup:SetPoint("TOPLEFT",cbLockPopup,"BOTTOMLEFT", 0, 0);
getglobal(cbLockTargetLeftPopup:GetName() .. "Text"):SetText("Display a popup when a Soulstone target leaves");
cbLockTargetLeftPopup:SetScript("OnClick", function()
									if (cbLockTargetLeftPopup:GetChecked() ) then
										LeaderBotOptions["LockTargetPopup"] = true;
									else
										LeaderBotOptions["LockTargetPopup"] = false;
									end 

								end);


-- ******************************************************
--Healer Leader Options
ts = LeaderBotOptionsPanel:CreateFontString("LeaderBot_HealerLeaderOptionsTitle","ARTWORK","GameFontWhiteSmall");
ts:SetPoint("TOPLEFT", cbLockTargetLeftPopup, 0, -50);
ts:SetText("Healer Leader Options");

local healerLeftPopup = CreateFrame("CheckButton", "LeaderBot_HealerLeftPopupCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
healerLeftPopup:SetPoint("TOPLEFT",ts,"BOTTOMLEFT", 0, -4);
getglobal(healerLeftPopup:GetName() .. "Text"):SetText("Show a popup when an assigned healer goes offline");
healerLeftPopup.tooltipText = "Display a popup when a healer goes offline";
healerLeftPopup:SetScript("OnClick", function()
									if (healerLeftPopup:GetChecked() ) then
										LeaderBotOptions["HealerLeftPopup"] = true;
									else
										LeaderBotOptions["HealerLeftPopup"] = false;
									end 

								end);
healerLeftPopup:SetEnabled(true);


-- ******************************************************
--Tank Leader Options
ts = LeaderBotOptionsPanel:CreateFontString("LeaderBot_TankLeaderOptionsTitle","ARTWORK","GameFontWhiteSmall");
ts:SetPoint("TOPLEFT", healerLeftPopup, 0, -50);
ts:SetText("Tank Leader Options");

local tankLeftPopup = CreateFrame("CheckButton", "LeaderBot_TankLeftPopupCheckBox", LeaderBotOptionsPanel,"OptionsCheckButtonTemplate");
tankLeftPopup:SetPoint("TOPLEFT",ts,"BOTTOMLEFT", 0, -4);
getglobal(tankLeftPopup:GetName() .. "Text"):SetText("Show a popup when an assigned tank goes offline");
tankLeftPopup.tooltipText = "Display a popup when a healer goes offline";
tankLeftPopup:SetScript("OnClick", function()
									if (tankLeftPopup:GetChecked() ) then
										LeaderBotOptions["TankLeftPopup"] = true;
									else
										LeaderBotOptions["TankLeftPopup"] = false;
									end 

								end);
tankLeftPopup:SetEnabled(true);

-- ************************************************************************************************
function LeaderBotOptionsPanel_Init()
	Omsg("Function: LeaderBotOptionsPanel_Init() ");

	LeaderBotOptionsPanel.refresh();

end

-- **********************************************************************************************
--called everytime the game options menu is opened, regardless of which addon opened it.
function LeaderBotOptionsPanel.refresh()
	Omsg("Function: LeaderBothOptionsPanel.refresh() ");

	local cbSmall = _G["LeaderBot_FootprintCheckBox"];

	if (LeaderBot_Footprint == LeaderBotSmall ) then
		cbSmall:SetChecked(true);
		LeaderBot_Footprint = LeaderBotSmall;
		LeaderBotColumnTitleFont = LeaderBotColumnTitleFontSmall;
		LeaderBotPlayerNameFont = LeaderBotPlayerNameFontSmall;
		LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontSmall;			
		LeaderBotplayerHeight = LeaderBotPlayerHeightSmall;
		LeaderBotheaderHeight = LeaderBotHeaderHeightSmall;
		LeaderBotfooterHeight = LeaderBotFooterHeightSmall;  
		LeaderBotborderPadding = LeaderBotBorderPaddingSmall;
		LeaderBotInfoPadding = LeaderBotInfoPaddingSmall;
	

	else  
		cbSmall:SetChecked(false);
		LeaderBot_Footprint = LeaderBotLarge;
		LeaderBotColumnTitleFont = LeaderBotColumnTitleFontLarge;
		LeaderBotPlayerNameFont = LeaderBotPlayerNameFontLarge;
		LeaderBotDropDownLabelFont = LeaderBotDropDownLabelFontLarge;		
		LeaderBotplayerHeight = LeaderBotPlayerHeightLarge;
		LeaderBotheaderHeight = LeaderBotHeaderHeightLarge;
		LeaderBotfooterHeight = LeaderBotFooterHeightLarge;  
		LeaderBotborderPadding = LeaderBotBorderPaddingLarge;
		LeaderBotInfoPadding = LeaderBotInfoPaddingLarge;

	end
	
	cb = _G["LeaderBot_LockLeaderCheckBox"];
	cb:SetChecked(LeaderBotOptions["LockLeader"]);

	cb = _G["LeaderBot_HealerLeaderCheckBox"];
	cb:SetChecked(LeaderBotOptions["HealerLeader"]);

	cb = _G["LeaderBot_TankLeaderCheckBox"];
	cb:SetChecked(LeaderBotOptions["TankLeader"]);

	cb = _G["LeaderBot_MageLeaderCheckBox"];
	cb:SetChecked(LeaderBotOptions["MageLeader"]);

	cb = _G["LeaderBot_HunterLeaderCheckBox"];
	cb:SetChecked(LeaderBotOptions["HunterLeader"]);

	cb = _G["LeaderBot_CrowdControllerCheckBox"];
	cb:SetChecked(LeaderBotOptions["CrowdController"]);



	local cb = _G["LeaderBot_SyncCheckBox"];
	cb:SetChecked(LeaderBotOptions["SyncMode"]);

	local cb = _G["LeaderBot_SyncDebugCheckBox"];
	cb:SetChecked(LeaderBotOptions["SyncModeDebug"]);

	cb = _G["LeaderBot_5RowsCheckBox"];
	cb:SetChecked(LeaderBotOptions["5rows"]);
	if (LeaderBotOptions["5rows"]) then 
		LeaderBotScrollSlots = 5;
	else 
		LeaderBotScrollSlots = 10;
	end 

	local cb = _G["LeaderBot_noLeadCheckBox"];
	cb:SetChecked(LeaderBotOptions["noLead"]);



	cb = _G["LeaderBot_LockWhisperCheckBox"];
	cb:SetChecked(LeaderBotOptions["LockSSWhisper"]);

	cb = _G["LeaderBot_LockOfflinePopupCheckBox"];
	cb:SetChecked(LeaderBotOptions["LockOfflinePopup"]);

	cb = _G["LeaderBot_LockPopupCheckBox"];
	cb:SetChecked(LeaderBotOptions["LockSSPopup"]);

	cb = _G["LeaderBot_LockTargetLeftPopupCheckBox"];
	cb:SetChecked(LeaderBotOptions["LockTargetPopup"]);




	cb = _G["LeaderBot_HealerLeftPopupCheckBox"];
	cb:SetChecked(LeaderBotOptions["HealerLeftPopup"]);


	cb = _G["LeaderBot_TankLeftPopupCheckBox"];
	cb:SetChecked(LeaderBotOptions["TankLeftPopup"]);



end

	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
--[[
	-- Create sub panel for aura alerts
	local panel = CreateFrame("FRAME","NightLight_AuraAlertsFrame",mainpanel, BackdropTemplateMixin and "BackdropTemplate");
	panel.name = "Aura Alerts";
	panel.parent = mainpanel.name;
	InterfaceOptions_AddCategory(panel);

	-- set the title text
	ts = panel:CreateFontString("NightLight_AuraTitle","ARTWORK","GameFontNormalLarge");
	ts:SetPoint("TOPLEFT", mainpanel, "TOPLEFT", 16,-16);
	ts:SetText("Aura Alerts");

	--make a scrolling frame to show known spells
	local sf = CreateFrame("SCROLLFRAME", "NLSpellScrollFrame", panel, "FauxScrollFrameTemplate");

	sf:SetWidth(NL_MAX_WIDTH);
	sf:SetHeight(NL_MAX_SLOTS * NL_LINE_HEIGHT + NL_LINE_HEIGHT);
	sf:SetPoint("TopLeft","NightLight_AuraTitle", "TopLeft", 5, -20);
	sf:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
                                            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", 
                                            tile = true, tileSize = 16, edgeSize = 16, 
                                            insets = { left = 5, right = 5, top = 5, bottom = 5 }});

	
	local fs1 = sf:CreateFontString("NLCol1_Title","OVERLAY", "GameTooltipText");
	fs1:SetPoint("TopLeft", sf, "TopLeft", 10, -10);
	fs1:SetText("Select:");

	local fs2 = sf:CreateFontString("NLCol2_Title","OVERLAY", "GameTooltipText");
	fs2:SetPoint("TopLeft", fs1, "TopRight", 10, 0);
	fs2:SetText("Duration");

	local fs3 = sf:CreateFontString("NLCol3_Title","OVERLAY", "GameTooltipText");
	fs3:SetPoint("TopLeft", fs2, "TopRight", 10, 0);
	fs3:SetText("Spellid");

	local fs4 = sf:CreateFontString("NLCol4_Title","OVERLAY", "GameTooltipText");
	fs4:SetPoint("TopLeft", fs3, "TopRight", 10, 0);
	fs4:SetText("Aura");

	-- initialize the scrolling entries
	
	subframe = CreateFrame("CheckButton", "NLLine0", sf, "OptionsCheckButtonTemplate");
	subframe:SetPoint("Top", "NLCol1_Title", "Bottom");
	subframe:SetScript("OnClick",function (self) NL_SpellButton(self) end );
	subframe:SetID(0);

	ts = sf:CreateFontString("NLLine0Duration","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLLine0", "Right",20, 0);

	ts =  sf:CreateFontString("NLLine0Spellid","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLLine0Duration", "Left", 65, 0);

	ts = sf:CreateFontString("NLLine0Spellname","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLLine0Spellid", "Left", 45, 0);

	for i = 1, NL_MAX_SLOTS, 1 do
		local subframe = CreateFrame("CheckButton", "NLLine" .. i, sf, "OptionsCheckButtonTemplate");
		subframe:SetPoint("Top", "NLLine" .. (i-1), "Bottom", 0, 6);
		subframe:SetScript("OnClick",function (self) NL_SpellButton(self) end );
		subframe:SetID(i);

		ts = sf:CreateFontString("NLLine"..i.."Duration","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLLine"..i, "Right",20, 0);

	 	ts = sf:CreateFontString("NLLine" .. i .. "Spellid","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLLine" .. i .. "Duration", "left", 65, 0);

	 	ts = sf:CreateFontString("NLLine" .. i .. "Spellname","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLLine" .. i .. "Spellid", "left", 45, 0);
	end

	sf:SetScript("OnVerticalScroll", function(self, offset)
							FauxScrollFrame_OnVerticalScroll(self, offset, 16, NLScrollBar_Update) 
						end);
	sf:SetScript("OnMouseWheel", function(self, arg1) 
							ScrollFrameTemplate_OnMouseWheel(self, arg1, self.scrollBar) 
						end);
	sf:SetScript("OnShow",NLScrollBar_Update);

	-- display 4 checkboxes for controlling which sounds to play

	for i=1,4 do
		cb = CreateFrame("CheckButton", "NightLight_OptionsCB"..i, panel,"OptionsCheckButtonTemplate");
		if (i == 1) then
			cb:SetPoint("TOPLEFT",sf,"BOTTOMLEFT", 0, -4);
		else 
			cb:SetPoint("TOPLEFT",dd,"BOTTOMLEFT", -8, 8);
		end

		getglobal(cb:GetName() .. "Text"):SetText(NL_CB_TEXT[i]);

		if (NL_SPELLS[NL_LastCheckedSpellID].s[i] == 1 ) then
			cb:SetChecked(true);
		else
			cb:SetChecked(false);
		end

		-- make the drop down menu for the start sound
		dd = CreateFrame("Frame","NLSoundDropDown"..i, panel, "UIDropDownMenuTemplate");
		UIDropDownMenu_SetWidth(dd,180);
		dd:SetPoint("TOPLEFT",cb, "BOTTOMLEFT", 8, 10);
		dd.initialize = function() NightLightFrame_SoundDropDownInit(i) end;


		-- if the user has already chosen this sound for this event, make sure the drop down shows it now.
		for j = 1,table.getn(NL_SOUNDS) do
			if ( NL_SPELLS[NL_LastCheckedSpellID].wav[i] == NL_SOUNDS[j].file) then
				UIDropDownMenu_SetText(getglobal("NLSoundDropDown"..i), NL_SOUNDS[j].sname);
				UIDropDownMenu_SetSelectedValue(getglobal("NLSoundDropDown"..i), NL_SOUNDS[j].file);
				j=4;
			end
		end

		-- set the OnClick handler for the CB function, and update NL_SPELLS accordingly.
		cb:SetScript("OnClick",
			function()
				if (cb:GetChecked() ) then
					NL_SPELLS[NL_LastCheckedSpellID].s[i] = 1;
				else
					NL_SPELLS[NL_LastCheckedSpellID].s[i] = 0;
				end
			end);
	end

	-- create button to remove unwanted buffs from the list
	button = CreateFrame("Button", "NLRemoveButton",panel,"UIPanelButtonTemplate");
	button:SetPoint("TopRight", NLSpellScrollFrame, "BottomRight");
	button:SetText("Remove");
	button:SetScript("OnClick", NL_RemoveButton);
	button:SetWidth(60);

	panel:Hide();
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	panel = CreateFrame("FRAME","NightLight_RareMobs",mainpanel, BackdropTemplateMixin and "BackdropTemplate");
	panel.name = "Rare Mobs";
	panel.parent = mainpanel.name;
	panel.default = NL_RaresReset;
	InterfaceOptions_AddCategory(panel);

	-- set the title text
	ts = panel:CreateFontString("NightLight_RareTitle","ARTWORK","GameFontNormalLarge");
	ts:SetPoint("TOPLEFT",16,-16);
	ts:SetText("Rare Mobs");

	sf = CreateFrame("SCROLLFRAME", "NLrScrollFrame", panel, "FauxScrollFrameTemplate");

	sf:SetWidth(NL_MAX_WIDTH);
	sf:SetHeight(NL_MAX_SLOTS * NL_LINE_HEIGHT + NL_LINE_HEIGHT);
	sf:SetPoint("TopLeft","NightLight_RareTitle", "TopLeft", 5, -20);
	sf:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
                                            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", 
                                            tile = true, tileSize = 16, edgeSize = 16, 
                                            insets = { left = 5, right = 5, top = 5, bottom = 5 }});

	
	fs1 = sf:CreateFontString("NLrCol1_Title","OVERLAY", "GameTooltipText");
	fs1:SetPoint("TopLeft", sf, "TopLeft", 10, -10);
	fs1:SetText("Select:");

	fs2 = sf:CreateFontString("NLrCol2_Title","OVERLAY", "GameTooltipText");
	fs2:SetPoint("TopLeft", fs1, "TopRight", 10, 0);
	fs2:SetText("Zone");

	fs3 = sf:CreateFontString("NLrCol3_Title","OVERLAY", "GameTooltipText");
	fs3:SetPoint("TopLeft", fs2, "TopRight", 90, 0);
	fs3:SetText("Name");

	fs4 = sf:CreateFontString("NLrCol4_Title","OVERLAY", "GameTooltipText");
	fs4:SetPoint("TopLeft", fs3, "TopRight", 90, 0);
	fs4:SetText("Level");

		-- initialize the scrolling entries
	subframe = CreateFrame("CheckButton", "NLrLine0", sf, "OptionsCheckButtonTemplate");
	subframe:SetPoint("Top", "NLrCol1_Title", "Bottom");
	subframe:SetScript("OnClick",function (self) NL_rSpellButton(self) end );
	subframe:SetID(0);

	ts = sf:CreateFontString("NLrLine0Zone","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLrLine0", "Right",20, 0);

	ts =  sf:CreateFontString("NLrLine0Name","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLrLine0Zone", "Left", 120, 0);

	ts =  sf:CreateFontString("NLrLine0Level","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLrLine0Name", "Left", 140, 0);

	for i = 1, NL_MAX_SLOTS, 1 do
		local subframe = CreateFrame("CheckButton", "NLrLine" .. i, sf, "OptionsCheckButtonTemplate");
		subframe:SetPoint("Top", "NLrLine" .. (i-1), "Bottom", 0, 6);
		subframe:SetScript("OnClick",function (self) NL_rSpellButton(self) end );
		subframe:SetID(i);

		ts = sf:CreateFontString("NLrLine"..i.."Zone","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLrLine"..i, "Right",20, 0);

	 	ts = sf:CreateFontString("NLrLine" .. i .. "Name","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLrLine" .. i .. "Zone", "left",120, 0);

	 	ts = sf:CreateFontString("NLrLine" .. i .. "Level","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLrLine" .. i .. "Name", "left",140, 0);
	end

	sf:SetScript("OnVerticalScroll", function(self, offset)
							FauxScrollFrame_OnVerticalScroll(self, offset, 16, NLrScrollBar_Update) 
						end);
	sf:SetScript("OnMouseWheel", function(self, arg1) 
							ScrollFrameTemplate_OnMouseWheel(self, arg1, self.scrollBar) 
						end);
	sf:SetScript("OnShow",NLrScrollBar_Update);
	panel:Hide();

	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	panel = CreateFrame("FRAME","NightLight_HostileAlertsFrame",mainpanel, BackdropTemplateMixin and "BackdropTemplate");
	panel.name = "Hostile Alerts";
	panel.parent = mainpanel.name;
	InterfaceOptions_AddCategory(panel);

	-- set the title text
	ts = panel:CreateFontString("NightLight_HostileTitle","ARTWORK","GameFontNormalLarge");
	ts:SetPoint("TOPLEFT",16,-16);
	ts:SetText("Hostile Alerts");

	sf = CreateFrame("SCROLLFRAME", "NLhScrollFrame", panel, "FauxScrollFrameTemplate");
	
	sf:SetWidth(NL_MAX_WIDTH);
	sf:SetHeight(NL_MAX_SLOTS * NL_LINE_HEIGHT + NL_LINE_HEIGHT);
	sf:SetPoint("TopLeft","NightLight_HostileTitle", "TopLeft", 5, -20);
	sf:SetBackdrop({bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background", 
                                            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", 
                                            tile = true, tileSize = 16, edgeSize = 16, 
                                            insets = { left = 5, right = 5, top = 5, bottom = 5 }});

	
	fs1 = sf:CreateFontString("NLhCol1_Title","OVERLAY", "GameTooltipText");
	fs1:SetPoint("TopLeft", sf, "TopLeft", 10, -10);
	fs1:SetText("Select:");

	fs2 = sf:CreateFontString("NLhCol2_Title","OVERLAY", "GameTooltipText");
	fs2:SetPoint("TopLeft", fs1, "TopRight", 10, 0);
	fs2:SetText("Class");

	fs3 = sf:CreateFontString("NLhCol3_Title","OVERLAY", "GameTooltipText");
	fs3:SetPoint("TopLeft", fs2, "TopRight", 10, 0);
	fs3:SetText("Spellid");

	fs4 = sf:CreateFontString("NLhCol4_Title","OVERLAY", "GameTooltipText");
	fs4:SetPoint("TopLeft", fs3, "TopRight", 10, 0);
	fs4:SetText("Spell");

		-- initialize the scrolling entries
	subframe = CreateFrame("CheckButton", "NLhLine0", sf, "OptionsCheckButtonTemplate");
	subframe:SetPoint("Top", "NLhCol1_Title", "Bottom");
	subframe:SetScript("OnClick",NL_HSpellButton);
	subframe:SetID(0);

	ts = sf:CreateFontString("NLhLine0Class","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLhLine0", "Right",20, 0);

	ts =  sf:CreateFontString("NLhLine0Spellid","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLhLine0Class", "Left", 65, 0);

	ts = sf:CreateFontString("NLhLine0Spellname","OVERLAY","GameTooltipTextSmall");
	ts:SetPoint("Left","NLhLine0Spellid", "Left", 45, 0);

	for i = 1, NL_MAX_SLOTS, 1 do
		local subframe = CreateFrame("CheckButton", "NLhLine" .. i, sf, "OptionsCheckButtonTemplate");
		subframe:SetPoint("Top", "NLhLine" .. (i-1), "Bottom", 0, 6);
		subframe:SetScript("OnClick",NL_HSpellButton);
		subframe:SetID(i);

		ts = sf:CreateFontString("NLhLine"..i.."Class","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLhLine"..i, "Right",20, 0);

	 	ts = sf:CreateFontString("NLhLine" .. i .. "Spellid","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLhLine" .. i .. "Class", "left", 65, 0);

	 	ts = sf:CreateFontString("NLhLine" .. i .. "Spellname","OVERLAY","GameTooltipTextSmall");
		ts:SetPoint("Left","NLhLine" .. i .. "Spellid", "left", 45, 0);
	end

	sf:SetScript("OnVerticalScroll", function(self, offset)
							FauxScrollFrame_OnVerticalScroll(self, offset, 16, NLhScrollBar_Update) 
						end);
	sf:SetScript("OnMouseWheel", function(self, arg1) 
							ScrollFrameTemplate_OnMouseWheel(self, arg1, self.scrollBar) 
						end);
	sf:SetScript("OnShow",NLhScrollBar_Update);
	panel:Hide();

	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	-- ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	local panel = CreateFrame("FRAME","NightLight_AddSoundsFrame",mainpanel, BackdropTemplateMixin and "BackdropTemplate");
	panel.name = "Add Sounds";
	panel.parent = mainpanel.name;
	InterfaceOptions_AddCategory(panel);

	-- set the title text
	local ts = panel:CreateFontString("NightLight_AddSoundTitle","ARTWORK","GameFontNormalLarge");
	ts:SetPoint("TOPLEFT",16,-16);
	ts:SetText("Add Sounds");

	ts = panel:CreateFontString("NightLight_AddSoundText","ARTWORK","GameFontNormalSmall");
	ts:SetPoint("TOPLEFT", getglobal("NightLight_AddSoundTitle"), "BOTTOMLEFT", 0, 0);
	ts:SetText(AddSoundText);

	local eb = CreateFrame("EditBox", "NightLight_AddSoundEditBox", panel, "InputBoxTemplate");
	eb:SetPoint("TOPLEFT", ts, "BOTTOMLEFT", 20, -10);
	eb:SetWidth(240);
	eb:SetHeight(14);
	eb:SetAutoFocus(false);
	
	--eb:SetScript("OnEditFocusLost",function () VetFrame_ParseString() end );
	eb:SetScript("OnEnterPressed",function () this:ClearFocus(); NightLightFrame_AddSound() end );
	--eb:SetScript("OnEscapePressed",function () this:ClearFocus(); VetFrame_ParseString() end );

	eb:SetScript("OnShow",
		function ()
			getglobal("NightLight_AddSoundEditBox"):SetCursorPosition(1);
		end);


--]]

