--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	
--]]

--[[ 

y. OnClick send assignment to raid or group
y. receive assignment sent to raid or group
y. trigger rosterUpdate


x. Join raid
x. send query "who is available to full sync?"
x. Assists respond to the channel
x. pick the first one to respond and whisper them for a full sync
x. they whisper back a full sync
x. they whisper when complete
x. we trigger a rosterUpdate


--]]

local SM_Debug = 0;
LeaderBotSyncFrame = nil;


local syncEngineVersion = 1.01;

local assignmentChangedSyncPrefix = "LBACSM";
local needFullRefreshSyncPrefix = "LBNFRSM";
local fullRefreshRowPrefix = "LBFRRSM";
local ssMessagePrefix = "LBSSSM";

local fullSyncRequested = false;

local colorCode = "|cff99aa99 ";
local whiteColorCode = "|cffffffff";

local fullRefreshRequest = false;
local fullRefershResponseInProgress = false;

local syncMasterTimers = {};

LeaderBotSyncSeqNumber = 0;

local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

    --TODO: when you click a pulldown's downarrow on a tab, clear the message icon for that tab. (Init function, I believe)
    --TODO: Only set notice flags if the recipient is modified.
    --TODO: trigger an update when a ss/heal target leaves the group and is reset to "NONE"
    --TODO: Do I need an "update log" visible to the user? Probably at least need to make the class button tooltip show who last updated


--*****************************************************************************************************
function LeaderBotSyncMaster_SetDebug(debugMode)
    SM_Debug = 0;
    if (debugMode) then
        SM_Debug = 1;
    end 
end 


--*****************************************************************************************************
-- Debugging text messages
function SMmsg(msg, forceShow)
	if (SM_Debug == 1 or forceShow ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
function LeaderBotSyncMaster_OnInit()
    SMmsg("Function: LeaderBotSyncMaster_OnInit()" );

    if (LeaderBotOptions["SyncMode"] == false) then 
        SMmsg("     Sync Mode is disabled. Returning.");    
        return LeaderBotSyncFrame;
    end

    if (LeaderBotSyncFrame == nil) then 
        LeaderBotSyncFrame = CreateFrame("Frame", "LeaderBotSyncFrame", parent, BackdropTemplateMixin and "BackdropTemplate");

        LeaderBotSyncFrame:SetScript("OnEvent", LeaderBotSyncMaster_OnEvent );
        LeaderBotSyncFrame:SetScript("OnUpdate", LeaderBotSyncMaster_OnUpdate );
 

        local rc;
        rc = C_ChatInfo.RegisterAddonMessagePrefix(assignmentChangedSyncPrefix);
        if (not rc) then 
            print(colorCode  .. "     RegosterAddonMessagePrefix failed 200");
        end 

        rc = C_ChatInfo.RegisterAddonMessagePrefix(needFullRefreshSyncPrefix);
        if (not rc) then 
            print(colorCode  .. "     RegosterAddonMessagePrefix failed 201");
        end 

        rc = C_ChatInfo.RegisterAddonMessagePrefix(fullRefreshRowPrefix);
        if (not rc) then 
            print(colorCode  .. "     RegosterAddonMessagePrefix failed 202");
        end 

        rc = C_ChatInfo.RegisterAddonMessagePrefix(ssMessagePrefix);
        if (not rc) then 
            print(colorCode  .. "     RegosterAddonMessagePrefix failed 203");
        end 

    end 

    LeaderBotSyncFrame:RegisterEvent("CHAT_MSG_ADDON", LeaderBotSyncMaster_OnEvent);

    --[[
    --debugging code 
    syncMasterTimers["testTiimer"] = {};
    syncMasterTimers["testTiimer"].time = 5.0; --sec
    syncMasterTimers["testTiimer"].handler = LeaderBotSyncMaster_RequestFullRefresh;
    --]] 

    return LeaderBotSyncFrame;

end

--******************************************************************************************************
function LeaderBotSyncMaster_Disable()
    LeaderBotSyncFrame:UnregisterEvent("CHAT_MSG_ADDON");

end 

--******************************************************************************************************
function LeaderBotSyncMaster_OnUpdate(self, elapsed)

    if (LeaderBotOptions["SyncMode"] == false) then 
        --SMmsg("     Sync Mode is disabled. Returning.");    
        return;
    end

    for key, timer in next, syncMasterTimers do
        
        if (timer.time) then 
            timer.time = timer.time - elapsed;
            if (timer.time <= 0) then 
                timer.handler();
                syncMasterTimers[key] = {};
            end

        end 
    end 

end


--******************************************************************************************************
function LeaderBotSyncMaster_OnEvent(self, event, ...)

    --CHAT_MSG_ADDON (the only event we've registered for) 
    local prefix = select(1, ...);  

    --print(colorCode .. "Event: " .. event .. "-" .. prefix);

    if (LeaderBotOptions["SyncMode"] == false) then 
        SMmsg("     Sync Mode is disabled. Returning.");    
        return;
    end

    if (prefix == assignmentChangedSyncPrefix) then
        --this message is for us
        SMmsg("Function: LeaderBotSyncMaster_OnEvent(assignmentChanged)" );

        local msg = select(2, ...); 
        LeaderBotSyncMaster_ProcessIncomingAssignment(msg);

    elseif (prefix == needFullRefreshSyncPrefix) then
        SMmsg("Function: LeaderBotSyncMaster_OnEvent(full sync requested)" );

        local msg = select(2, ...); 
        LeaderBotSyncMaster_RespondToFullRefresh(msg);

    elseif (prefix == fullRefreshRowPrefix) then 
        SMmsg("Function: LeaderBotSyncMaster_OnEvent(full refresh row)" );

       local msg = select(2, ...); 
       LeaderBotSyncMaster_ProcessUberIncomingAssignment(msg);

    elseif (prefix == ssMessagePrefix) then 
        SMmsg("Function: LeaderBotSyncMaster_OnEvent(soul stone msg)");
        
        local msg = select(2, ...);
        LeaderBotSyncMaster_ProcessSSmsg(msg);
    end

end 

--******************************************************************************************************
function LeaderBotSyncMaster_ParseMsg(msg, fieldNames)
    SMmsg("Function: LeaderBotSyncMaster_ParseMsg()" );

    -- find all the delimiters in the string
    local delimiters = {};
    local dCount = 1;
    delimiters[dCount] = string.find(msg, ":");
    while (delimiters[dCount] ~= nil) do  
        delimiters[dCount+1] = string.find(msg, ":", delimiters[dCount] + 1);
       dCount = dCount + 1;
    end 
  
    -- Did we get the correct number of fields?
    if (dCount ~= table.getn(fieldNames) ) then 
        print(colorCode .. "     Leader Bot Sync Manager **ERROR**. Found " .. dCount .. " fields but expected " .. table.getn(fieldNames));
    end 

    --extract the fields
    local fieldValues = {};
    for j = 1, dCount, 1 do 
        if (j == 1) then 
            --first pass, grab from char 1 to the delimiter
            fieldValues[ fieldNames[j] ] = string.sub(msg, 1, delimiters[j] -1);

        elseif (j == dCount) then 
            --last pass, grab from the previous delimiter (the final one) to the end of the string
            fieldValues[ fieldNames[j] ] = string.sub(msg, delimiters[j -1] + 1);

        else
            --in the middle, grab from the last delimiter to the current delimiter       
            fieldValues[ fieldNames[j] ] = string.sub(msg, delimiters[j - 1] + 1, delimiters[j] - 1 );

        end
    end

--[[
    --debugging code
    for key, value in next, fieldValues do
        SMmsg("     " .. key .. ": " .. value);
    end
--]]
    return fieldValues;

end 

--******************************************************************************************************
function LeaderBotSyncMaster_SendMsg(senderName, whisperTarget, groupName, assignmentType, assignee, assignmentValue, multiSelectState)
    SMmsg("Function: LeaderBotSyncMaster_SendMsg()" );

    if (LeaderBotOptions["SyncMode"] == false) then
        SMmsg("     Sync Mode is disabled. Returning.");    
        return;
    end 

    local chatType;

    --debugging code
    SMmsg ("     seqNumber: " .. LeaderBotSyncSeqNumber);
    SMmsg ("     SenderName: " .. senderName);
    SMmsg ("     groupName: " .. groupName);
    SMmsg ("     assignmentType " .. assignmentType);
    SMmsg ("     assignee: " .. assignee);
    SMmsg ("     assignmentValue " .. assignmentValue);


    local multiSelectStateTextVal;
    if (multiSelectState) then 
        SMmsg ("     multiSelectState: true");
        multiSelectStateTextVal = 1;
    else
        SMmsg("      multiSelectState: false");
        multiSelectStateTextVal = 0;        
    end 

    local refreshOnThisRow = "yes";

    if(whisperTarget ~= nil) then 
        SMmsg(colorCode .. "     Sending Data via Whisper to " .. whisperTarget);
        refreshOnThisRow = "no";

    elseif (IsInRaid() ) then
        SMmsg(colorCode .. "     Sending data via Raid");
        chatType = "RAID";

    elseif (IsInGroup() ) then 
        SMmsg(colorCode .. "     Sending data via Party");
        chatType = "PARTY";

    else
        --nobody to sync with
        return;
    end 


    LeaderBotSyncSeqNumber = GetServerTime();

    local msg = LeaderBotSyncSeqNumber .. ":" .. 
                syncEngineVersion .. ":" .. 
                senderName .. ":" .. 
                refreshOnThisRow .. ":".. 
                groupName .. ":" .. 
                assignmentType .. ":" .. 
                assignee .. ":" .. 
                assignmentValue .. ":" .. 
                multiSelectStateTextVal;

    local rc = C_ChatInfo.SendAddonMessage(assignmentChangedSyncPrefix, msg, chatType, whisperTarget );
    if (not rc) then
        message("Leader Bot Sync Master failed: 100");
    end 

end

--******************************************************************************************************
function LeaderBotSyncMaster_ProcessIncomingAssignment(msg)


    local fieldNames = { 
                            [1] = "seqNumber",
                            [2] = "syncEngineVersion",
                            [3] = "senderName",
                            [4] = "refreshOnThisRow",
                            [5] = "groupName",
                            [6] = "assignmentType",
                            [7] = "assignee",
                            [8] = "assignmentValue",
                            [9] = "multiSelectStateTextVal",    --optional
                        };

    local fieldValues = LeaderBotSyncMaster_ParseMsg(msg, fieldNames);

    SMmsg("     processing msg from: " .. fieldValues["senderName"] .. ": " .. 
                                                                            fieldValues["assignee"] .. ", " .. 
                                                                            fieldValues["assignmentType"]  .. 
                                                                            ")" );

   --turns out we get events firing in response to our own outgoing edits. No need to go through the whole 
    --roster update process for our own changes.
    if (string.match(fieldValues["senderName"], LeaderBotSenderName) ) then 
        SMmsg("     Recieved event for our own message, IGNORING");
        return;
    end 


    LeaderBotSyncSeqNumber = fieldValues["seqNumber"];

    if (fieldValues["multiSelectStateTextVal"] == "1") then
        fieldValues["multiSelectStateTextVal"] = true;
    else 
        fieldValues["multiSelectStateTextVal"] = false;
    end 

  
    --update the local assignment and alert notifications
   LeaderBotFrameNoticeFrame:Show();

    local assignmentArray;

    if (fieldValues["groupName"] == "Warlock") then 
        assignmentArray = LockLeader_Assignments;
		LeaderBotFrameWarlockNoticeFrame:Show();

    elseif (fieldValues["groupName"] == "healer") then 
        assignmentArray = HealerLeader_Assignments;
        LeaderBotFrameHealerNoticeFrame:Show();

    elseif (fieldValues["groupName"] == "tank") then 
        assignmentArray = TankLeader_Assignments;
        LeaderBotFrameTankNoticeFrame:Show();

    elseif (fieldValues["groupName"] == "Mage") then 
        assignmentArray = MageLeader_Assignments;
        LeaderBotFrameMageNoticeFrame:Show();

    elseif (fieldValues["groupName"] == "Hunter") then 
        assignmentArray = HunterLeader_Assignments;
        LeaderBotFrameHunterNoticeFrame:Show();
    
    elseif (fieldValues["groupName"] == "crowdcontroller") then
        assignmentArray = CrowdController_Assignments;
        LeaderBotFrameCrowdControllerNoticeFrame:Show();
    end 


    if (string.find(fieldValues["assignmentType"], "Multi")) then
        assignmentArray [ fieldValues["assignmentType"] ]
                        [ fieldValues["assignee"] ]
                        [ fieldValues["assignmentValue"] ] = fieldValues["multiSelectStateTextVal"];
    else 
        assignmentArray [ fieldValues["assignmentType"] ]
                        [ fieldValues["assignee"] ] = fieldValues["assignmentValue"];
    end 
  

    --trigger an update if this row completes the transaction, or wait it 
    --this is a batch data send.
    if (string.find(fieldValues["refreshOnThisRow"], "yes") ) then
        LeaderBot_QueueRosterUpdate();
    end


    --note: LeaderBotFrame handles clearing notice flags:
        --1. pressing leaderBotFrame maximize button
        --2. pressing any of the *_Leader buttons on LeaderBotFrameMax

end 


--******************************************************************************************************
function LeaderBotSyncMaster_RequestFullRefresh()
    SMmsg("Function: LeaderBotSyncMaster_RequestFullRefresh()" );
   
    if (LeaderBotOptions["SyncMode"] == false) then 
        SMmsg("     Sync Mode is disabled. Returning.");
        return;
    end

    if (fullSyncRequested) then 
        --we already have a request in progress.
        SMmsg("     I already have a full sync in progress");
        return;
    end

    local isPlayerLeader = UnitIsGroupLeader("player");
    if (isPlayerLeader) then
        --we are the leader already, so don't send ourselves a sync request
        SMmsg("     We are the raid leader, no need to send ourselves a sync request");
        return;
    end 

    if (IsInRaid() ) then
        SMmsg(colorCode .. "     Sending data via Raid");
        chatType = "RAID";

    elseif (IsInGroup() ) then 
        SMmsg(colorCode .. "     Sending data via PARTY");
        chatType = "PARTY";

    else
        --nobody to sync with
        SMmsg("     Not in a raid or party, nobody to sync with.");
        return;
    end 

    fullSyncRequested = true;

    --show an indication of sync in progress
    if (LeaderBotSyncMasterSyncFrame == nil) then 
        LeaderBotSyncMasterSyncFrame = CreateFrame("Frame", "LeaderBotSyncMasterSyncFrame", LeaderBotFrameMax, BackdropTemplateMixin and "BackdropTemplate");
    end 

    local currentPanel;
    for panelName, panelFrame in next, LeaderBotPanels do 
		if (LeaderBotOptions["currentPanel"] == panelName) then
			--make sure this panel shows last for rezing
			currentPanel = panelFrame;
            break;
		end 
	end 
    
    LeaderBotSyncMasterSyncFrame:SetSize(currentPanel:GetWidth(), currentPanel:GetHeight() );
 	LeaderBotSyncMasterSyncFrame:SetPoint("Center", currentPanel, "Center", 0, 0);
    LeaderBotSyncMasterSyncFrame.texture = LeaderBotSyncMasterSyncFrame:CreateTexture("SyncInProgress", "OVERLAY" );
    LeaderBotSyncMasterSyncFrame.texture:SetAllPoints(LeaderBotSyncMasterSyncFrame);
    LeaderBotSyncMasterSyncFrame.texture:SetTexture(1,1,1,1);
    LeaderBotSyncMasterSyncFrame:Show();

    local msg = LeaderBotSyncSeqNumber .. ":" .. 
                syncEngineVersion .. ":" .. 
                LeaderBotSenderName .. ":" ..
                "request";
    local rc = C_ChatInfo.SendAddonMessage(needFullRefreshSyncPrefix, msg, chatType );

    if (not rc) then 
        message("Leader Bot Sync Master failed: 101");
    end 

    syncMasterTimers["resetSyncRequestOnNoReply"] = {};
    syncMasterTimers["resetSyncRequestOnNoReply"].time = 5.5; --sec
    syncMasterTimers["resetSyncRequestOnNoReply"].handler = function() 
                                                                        SMmsg("Function: resetSyncRequestOnNoReply");

                                                                        if (fullSyncRequested) then
                                                                            --no raid officers were available to respond to this request in the allotted time.
                                                                            --so clear the request flag.
                                                                            SMmsg("     No reply to Full Sync request. Canceling request");
                                                                            fullSyncRequested = false;  

                                                                            StaticPopupDialogs["LEADERBOT_SYNC_FAILED"] = {
				                                                                        preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
				                                                                        timeout = 0,
                                                                                        whileDead = true,
                                                                                        hideOnEscape = true,
                                                                                        text = "Leader Bot sync failed, the raid leader appears not to have Leader Bot installed.",
                                                                                                    
                                                                                        button1 = "OK",
                                                                                        OnAccept = function()
                                                                                                        							
                                                                                                    end,
			
                                                                                        };

                                                                            StaticPopup_Show("LEADERBOT_SYNC_FAILED");
                                                                            
                                                                                                                                                      
                                                                        else
                                                                            SMmsg("     My Full Sync request has been completed");

                                                                            StaticPopupDialogs["LEADERBOT_SYNC_SUCCESS"] = {
				                                                                        preferredIndex = 3,  -- avoid some UI taint (not sure what that means, found it on the web. explanation link was dead)
				                                                                        timeout = 0,
                                                                                        whileDead = true,
                                                                                        hideOnEscape = true,
                                                                                        text = "Leader Bot sync completed successfully.",
                                                                                                    
                                                                                        button1 = "OK",
                                                                                        OnAccept = function()
                                                                                                        							
                                                                                                    end,
			
                                                                                        };

                                                                            StaticPopup_Show("LEADERBOT_SYNC_SUCCESS");  

                                                                        end

                                                                        LeaderBotSyncMasterSyncFrame:Hide();
                                                            end;

end 


--******************************************************************************************************
function LeaderBotSyncMaster_RespondToFullRefresh(msg)
    SMmsg("Function: LeaderBotSyncMaster_RespondToFullRefresh()" );

    local fieldNames = { 
                            [1] = "seqNumber",
                            [2] = "syncEngineVersion",
                            [3] = "senderName",
                            [4] = "messageType",        --"request" or "available" or "accepted" or completed
                        };

    local fieldValues = LeaderBotSyncMaster_ParseMsg(msg, fieldNames);

    if (string.match(fieldValues["senderName"], LeaderBotSenderName) ) then 
        SMmsg("     Recieved event for our own message, IGNORING");
        return;
    end 

    --because of race conditions when a player joins a group, they may still appear as
    --being a legit (solo) assistant here before leaderbot resets their status. So, we 
    --have to double check if someone is an assistant, are they *really* an assistant?
    if (LeaderBotIsAssistant) then 
		if (UnitIsGroupAssistant("player") == true 
			or UnitIsGroupLeader("player") == true
			or (not IsInGroup() and not IsInRaid() )
			) then 
				LeaderBotIsAssistant = true;
		else					
				LeaderBotIsAssistant = false;
		end 
    end 

    -- *****************************
    --Senders: announce AVAILABLE
    if (string.match(fieldValues["messageType"], "request") and UnitIsGroupLeader("player") == true ) then 
        --reply that I am available.
        SMmsg("     I have received a request for a full sync from " .. fieldValues["senderName"]);

        local msg = LeaderBotSyncSeqNumber .. ":" .. 
                syncEngineVersion .. ":" .. 
                LeaderBotSenderName .. ":" ..
                "available";
        local rc = C_ChatInfo.SendAddonMessage(needFullRefreshSyncPrefix, msg, "WHISPER", fieldValues["senderName"] );
        if (not rc)  then
            message("Leader Bot Sync Master failed: 102");
        end 

    -- *****************************
    --Requestor: ACCEPT the first response
    elseif (string.match(fieldValues["messageType"], "available") and fullSyncRequested) then
        --if we have a full sync request outstanding, then accept the first response we get.
        SMmsg("     " .. fieldValues["senderName"] .. " has responded to my full sync request");

        --whisper the lucky user that we want their data
        local msg = LeaderBotSyncSeqNumber .. ":" .. 
                syncEngineVersion .. ":" .. 
                LeaderBotSenderName .. ":" ..
                "accepted";
        local rc = C_ChatInfo.SendAddonMessage(needFullRefreshSyncPrefix, msg, "WHISPER", fieldValues["senderName"] );
        if (not rc)  then
            message("Leader Bot Sync Master failed: 103");
        end 

    -- *****************************
    -- Sender: Received Accepted
    elseif (string.match(fieldValues["messageType"], "accepted") ) then
        --if our offer was accepted, send them the data via whisper
        SMmsg("     " .. fieldValues["senderName"] .. " has accepted my full sync response");

       --whisper every row of all the assignments tables
        LeaderBotSyncMaster_SendFullSyncAssignments(fieldValues["senderName"]);

       --whisper a completed message
        SMmsg("     " .. fieldValues["senderName"] .. " I have completed sending data");
        
        local msg = LeaderBotSyncSeqNumber .. ":" .. 
                syncEngineVersion .. ":" .. 
                LeaderBotSenderName .. ":" ..
                "completed";
        local rc = C_ChatInfo.SendAddonMessage(needFullRefreshSyncPrefix, msg, "WHISPER", fieldValues["senderName"] );
        if (not rc) then
            message("Leader Bot Sync Master failed: 104");
        end 

    -- *****************************
    --Requestor: Received COMPLETED
    elseif (string.match(fieldValues["messageType"], "completed") and fullSyncRequested) then
        --when they tell us that they've completed the transfer, then update the display
        SMmsg("     " .. fieldValues["senderName"] .. " has completed data transfer to me");

        --initiate rosterUpdate
        LeaderBot_QueueRosterUpdate();
        fullSyncRequested = false;
    
    -- *****************************
    else 
        SMmsg("     I am not the raid leader, or this is an unknown message (" .. fieldValues["messageType"] .. ") from " .. fieldValues["senderName"] );
    end

end 

-- *******************************************************
function LeaderBotSyncMaster_SendFullSyncAssignments(whisperTarget)
    SMmsg("FUnction: LeaderBotSyncMaster_SendFullSyncAssignments(" .. whisperTarget .. ")" );
	local modules = {
            ["Warlock"] = LockLeader_Assignments, 
            ["healer"] = HealerLeader_Assignments, 
            ["tank"] = TankLeader_Assignments, 
            ["Mage"] = MageLeader_Assignments, 
            ["Hunter"] = HunterLeader_Assignments,
            ["crowdcontroller"] = CrowdController_Assignments,
            };

    LeaderBotSyncSeqNumber = LeaderBotSyncSeqNumber + 1;
    local messageHeader =   LeaderBotSyncSeqNumber .. ":" .. 
                            syncEngineVersion .. ":" .. 
                            LeaderBotSenderName .. ":" .. 
                            "no" .. ":";                    --refresh on this row == no
    local messageRows = {};
    local row;

	for groupName, module_Assignments in next, modules do

		for assignmentType, assignments in next, module_Assignments do

            if (string.find(assignmentType, "SoulstoneFoundOn") ) then 
               --fake continue statement

            else 

                --are we single or multiselect assignments?
                if (string.match(assignmentType, "Multi")) then 

                    for assignee, dataArray in next, module_Assignments[assignmentType] do 

                        --is the assignee in the current group, or just a stored value?
                        if (UnitClass(assignee) ~= nil) then 

                            local msg = "";
                            local comma = "";

                            for assignmentName, multiSelectState in next, module_Assignments[assignmentType][assignee] do

                                --convert bool to text since LUA can't fucking concat a bool into a string ><
                                local multiSelectStateTextVal = 0;
                                if (multiSelectState) then
                                    multiSelectStateTextVal = 1;
                                end 

                                row =   groupName .. ":" .. 
                                        assignmentType .. ":" .. 
                                        assignee .. ":" .. 
                                        assignmentName .. ":" .. 
                                        multiSelectStateTextVal;
                                table.insert(messageRows, row);

                                if (string.find(row, "Bearabic") ) then 
                                    --print("     mRow: " .. row);
                                end 
                            end 

                        end
                    end 
            
                else --single select assignments

          
                    for assignee, assignmentValue in next, assignments do
                        
                        --is the assignee in the current group, or just a stored value?  
                        if (UnitClass(assignee) ~= nil) then 
            
                            row =   groupName .. ":" .. 
                                        assignmentType .. ":" .. 
                                        assignee .. ":" .. 
                                        assignmentValue .. ":";
                            table.insert(messageRows, row);  

                        end 


                    end

                end 

            end -- if SoulstoneFound entry

		end 

	end

    --so now we have a header and series of rows. if we send them individually, they get throttled and lost.
    --if we send them collectively, they're limited to 255 chars. So glue them together into 250-byte
    --chunks.

--[[ 
--debugging
for index, row in next, messageRows do
    print(index .. " " .. row);
end 
--]]

    local count = 1;
    local msg = messageHeader .. ",";
    for index, row in next, messageRows do

        --have we exceeded the max message length?
        if (string.len(msg) + string.len(row) <= 250) then
            --no, we can just add this row to the message
            msg = msg .. row .. ",";
            messageRows[1] = nil;
        else
            --yes, so send the msg we've compiled so far
            
--[[            
            --debugging code
            if (string.find(msg, "Fortitude_MultiSelect")) then 
                print (msg);
                print ("length: " .. string.len(msg));
            end 
--]]

            --send the message
            local rc = C_ChatInfo.SendAddonMessage(fullRefreshRowPrefix, msg, "WHISPER", whisperTarget );
            if (not rc) then
                message("Leader Bot Sync Master failed: 105");
            else 
                SMmsg("     Send Count: " .. count);
                count = count +1;
            end 

            --start the new message with this row
            msg = messageHeader .. "," .. row .. ",";
            messageRows[1] = nil;
        end
    end 

--[[
    --debugging code
     if (string.find(msg, "Fortitude_MultiSelect")) then 
        print (msg);
        print ("length: " .. string.len(msg));
    end   
--]]

    --send the remainder message
    local rc = C_ChatInfo.SendAddonMessage(fullRefreshRowPrefix, msg, "WHISPER", whisperTarget );
    if (not rc) then
        message("Leader Bot Sync Master failed: 106");
    else 
        SMmsg("     Send Count: " .. count);
    end 

end 

-- ***********************************************************************
function LeaderBotSyncMaster_ProcessUberIncomingAssignment(msg)
    SMmsg("Function: LeaderBotSyncMaster_ProcessUberIncomingAssignment");

    --header: cut to first comma
    local firstComma = string.find(msg, ",", 1);
    local header = string.sub (msg, 1, firstComma -1);

    if (LBuberCount == nil) then 
        LBuberCount = 1;
    end 

    SMmsg("     Received full sync packet : " .. LBuberCount );
    LBuberCount = LBuberCount + 1;

--[[
    --debugging code
     if (string.find(msg, "Fortitude_MultiSelect")) then 
        print (msg);
        print ("length: " .. string.len(msg));
     end
--]]

    local oldCommaPos = firstComma;
    local commaPos = string.find(msg, ",", firstComma + 1);
    while (commaPos ~= nil) do
        local payload = string.sub(msg, oldCommaPos +1, commaPos -1);
        local packet = header .. payload;

        if (string.find(packet, "Fortitude_MultiSelect")) then 
            --print("    packet " .. packet); 
        end 

        LeaderBotSyncMaster_ProcessIncomingAssignment(packet);

        oldCommaPos = commaPos;
        commaPos = string.find(msg, ",", oldCommaPos + 1);        
    end 

end

--******************************************************************************************************
function LeaderBotSyncMaster_SendSSmsg(timestamp, msgType, senderName, caster, target)
    SMmsg("Function: LeaderBotSyncMaster_ProcessUberIncomingAssignment");

    local msg = timestamp   .. ":" .. 
                msgType     .. ":" .. 
                senderName  .. ":" ..
                caster      .. ":" .. 
                target;

    local rc = C_ChatInfo.SendAddonMessage(ssMessagePrefix , msg, chatType, whisperTarget );
    if (not rc) then
        message("Leader Bot Sync Master failed: 100");
    end 

end

-- **************************************************************************************************
function LeaderBotSyncMaster_ProcessSSmsg(msg)
    SMmsg("Function: LeaderBotSyncMaster_ProcessUberIncomingAssignment");

    local fieldNames = { 
                            [1] = "timestamp",
                            [2] = "msgType",
                            [3] = "sender",
                            [4] = "caster",
                            [5] = "target",       
                        };

    local fieldValues = LeaderBotSyncMaster_ParseMsg(msg, fieldNames);

    if (fieldValues["sender"] == LeaderBotSenderName) then 
        --ignore our own messages

    elseif (fieldValues["msgType"] == "CAST") then 
        --call LockLeader to update the SS status
        SMmsg("Received Cast message: " .. fieldValues["sender"] .. " says " .. fieldValues["caster"] .. " on " .. fieldValues["target"]);
        LockLeader_SetSoulStoneFound(fieldValues["caster"], fieldValues["target"], fieldValues["timestamp"]);

    elseif (fieldValues["msgType"] == "DIED") then 
        --call LockLeader to update the SS status
        SMmsg("Received Died message: " .. fieldValues["sender"] .. " says " .. fieldValues["target"] .. " died");
        LockLeader_SetSoulStoneLost(fieldValues["target"], fieldValues["msgType"])
    end 

end 
