--******************************************************************************************************
--[[
	(c) Black Dog Games 2019 all rights reserved
	qwarlock@gmail.com
	10/20/19  v1.0 for WoW Classic
	Manages warlock assignments during raids
	
	All code here is inside functions. Nothing runs when the .TOC loads, other than the
	high-level variable definitions. Execution beings in LeaderBot.lua:OnLoad()

--]]

local classicOrRetail;
local TL_Debug = 0;
local TL_loopcontrol = 0;
local version = "1.34";

TankLeaderFrame = nil;

local TANKS_MULTISELECT = "Tank_MultiSelect";
local TANKROT = "Tank Rotation";
local TANKASSIGNMENTCOUNT = 2;

--Global variables ot be saved
--tank leader assignments must be global to allow saving
TankLeader_Assignments = {};
TankLeader_Assignments[TANKS_MULTISELECT] = {};
TankLeader_Assignments[TANKROT] = {};

local LOCKCHANTYPE = 1;
local LOCKCHANLANG = 2;
local LOCKCHANNUMB = 3;
local TankLeader_MsgChannel ={};

local columnWidth = 150;
local columnWidthTank = 180;

local columnFrames = {};

local NAME = 1;
local CLASS = 2;
local LEVEL = 3;
local ISDEAD = 4;

local CurseOfAgonyID = 11713;
local CurseOfDoomID = 603;
local CurseOfTonguesID = 11719;
local CurseOfRecklessnessID = 7658;
local CurseOfWeaknessID = 7646;
local CurseOfElementsID = 11721;
local CurseOfShadowID = 17862;

local CHATCHANNAME = 1;
local CHATCHANTYPE = 2;
local CHATCHANNUM = 3;

local TankTypes = {"Warrior", "Paladin", "Druid" }; 

local variablesLoaded = false;
local addHealerIndex = 0;
local addHealerTargetIndex = 0;

local tankTargetDropDownName = "TankLeaderTankDropDown"
local tankRotationDropDownName = "TankLeaderTankRotationDropDown";
local tankDropDowns = { 
						[TANKS_MULTISELECT] = tankTargetDropDownName,
						[TANKROT] = tankRotationDropDownName,
					};


local colorCode = "|cffaa6699 ";
local whiteColorCode = "|cffffffff";

local mainBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\ChatFrame\\UI-DialogBox-Border",
	tile = false,		
	tileSize =0,	
	edgeSize = 2	
};

local columnBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};

local labelBackdrop = 
{
	bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
	edgeFile = nil,
	tile = true,		-- required to make Tooltip border work
	tileSize =16,	-- required to make Tooltip border work
	edgeSize = 16	
};


NL_ENEMY_MASK=bit.bor(COMBATLOG_OBJECT_REACTION_HOSTILE, 
				 COMBATLOG_OBJECT_CONTROL_PLAYER,
				 COMBATLOG_OBJECT_TYPE_PLAYER);

--*****************************************************************************************************
-- Debugging text messages
function TLmsg(msg)
	if (TL_Debug == 1 ) then
		ChatFrame1:AddMessage(colorCode .. msg);
	end
end

--******************************************************************************************************
-- Initialization
function TankLeaderFrame_OnLoad(parentFrame, teamMembers, tankCount)
	TLmsg("Function: TankLeaderFrame_OnLoad()");

	TankLeaderFrame = CreateFrame("Frame", "TankLeaderFrame", parentFrame, BackdropTemplateMixin and "BackdropTemplate");

	-- UI Commands
	SLASH_TankLeader1 = "/tl";
	SLASH_TankLeader2 = "/TL";

	SlashCmdList["TankLeader"] =TankLeader_SlashHandler;

	--HealerLeaderFrame:RegisterEvent("VARIABLES_LOADED");
	--HealerLeaderFrame:RegisterEvent("PARTY_MEMBERS_CHANGED");
	--HealerLeaderFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
     --

	ChatFrame1:AddMessage(colorCode .. "TankLeader Module for LeaderBot has been Loaded");
--	UIErrorsFrame:AddMessage("Quel's healerLeader Loaded.", 1.0, 1.0, 1.0, 1.0, UIERRORS_HOLD_TIME);

	classicOrRetail = "Classic";

	TankLeader_InitMainFrame(teamMembers, tankCount);

	return (TankLeaderFrame);

end

-- *************************************************************************************
function TankLeader_InitMainFrame(teamMembers, tankCount)
	TLmsg("Function: TankLeader_InitMainFrame");

	--Configure the main panel. This will change sizes as we add/remove columns, add/remove healers
	TankLeaderFrame:SetPoint("TopLeft", LeaderBotFrameMax, "BottomLeft", 0, 2);	
	TankLeader_AdjustMainPanelSize(tankCount);	

	TankLeaderFrame:SetBackdrop(mainBackdrop);
	TankLeaderFrame:SetBackdropColor(0.1, 0.1, 0.4, 0.85);
	TankLeaderFrame:SetBackdropBorderColor(0.0, 0.0, 0.0, 1.0);

	TankLeaderFrame:EnableMouse(true);
	TankLeaderFrame:SetScript("OnUpdate",TankLeaderFrame_OnUpdate);
	TankLeaderFrame:SetScript("OnEvent", TankLeaderFrame_OnEvent);

	TankLeaderFrame:SetScript("OnShow", function(self)
						TLmsg("Function: TankLeaderFrame_OnShow() ");
						TLmsg("current panel = " .. LeaderBotOptions["currentPanel"] );

						if (LeaderBotOptions["currentPanel"] == "tank") then 
							local parentFrame = TankLeaderFrame:GetParent();
							parentFrame:SetWidth(TankLeaderFrame:GetWidth() );
						end
				end);

	local appLogo = LeaderBot_GetTextForClass("Warrior");
	local appTitle = TankLeaderFrame:CreateFontString("TankLeaderAppTitle", "ARTWORK", "QuestFont_Large");
	appTitle:SetPoint("Top", TankLeaderFrame, "Top", 0, -7);
	appTitle:SetText(appLogo .. " Tank Leader " .. version);

	local infoText = TankLeaderFrame:CreateFontString("TankLeaderInfoText", "ARTWORK", "GameFontWhiteSmall");
	infoText:SetPoint("Top", appTitle, "Bottom", 0, LeaderBotInfoPadding);
	infoText:SetText(colorCode .. "Showing all classes that could be tanks. Assign Non-tanks to \"-None-\"\nto exclude them from assignment announcements.");

	columnFrames[1] = LeaderBotFrame_AddColumn1(teamMembers, tankCount, "tank", TankLeaderFrame, "Tanks");
	columnFrames[2], TankLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(teamMembers, tankCount, true, 
							TankLeaderFrame, columnFrames[1], 
							columnWidthTank, "TankLeaderColumn2", "Tank Target", 
							"LeaderBottankMember", tankTargetDropDownName, 
							TankLeader_Assignments, TANKS_MULTISELECT, 
							"TankLeaderAnnounce2", TankLeader_AnnounceToChat, "Leader Bot: Tanking Targets",
							TankLeader_TankTargetDropDownOnClick,
							LeaderBotMultiSelect);
	columnFrames[3] = TankLeader_AddColumnTankRotation(teamMembers, tankCount, true);

	if (tankCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(3, "tank", 1, columnFrames[1], "LeaderBottankMember");
		LeaderBot_RegisterColumnForScrolling(3, "tank", 2, columnFrames[2], tankTargetDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "tank", 3, columnFrames[3], tankRotationDropDownName);

		LeaderBot_SetupScrollBar("tank");
	end 

	-- make the drop down menu for chat channels
	local dd = CreateFrame("Frame","TankLeaderChatChannelDropDown", TankLeaderFrame, "UIDropDownMenuTemplate");
	UIDropDownMenu_SetWidth(dd,140 );
	dd:SetPoint("BottomLeft", TankLeaderFrame, "BottomLeft", -5, 5);
	dd.initialize = function() LeaderBot_ChannelDropDownInit(dd, channelCount, channelLis, TankLeader_ChannelDropDownOnClick) end;
	channelCount, channelList = LeaderBotFrame_GetChatChannels();
	UIDropDownMenu_SetText(dd, "<Select channel>");
	
	if (channelCount == 0) then
		dd.tooltipText = "No channels joined";
	else
		dd.tooltipText = "Select chat channel";
	end
	dd:SetScript("OnEnter", LeaderBotFrame_OnEnterButton);
	dd:SetScript("OnLeave", LeaderBotFrame_OnLeaveButton);

	if (LeaderBotIsAssistant) then
		UIDropDownMenu_EnableDropDown(dd);
	else 
		UIDropDownMenu_DisableDropDown(dd);
	end

end


-- ******************************************************************************
--adjust the height as players join or leave
function TankLeader_AdjustMainPanelSize(groupCount)
	TLmsg("Function: TankLeader_AdjustMainPanelSize(count = " .. groupCount .. ")" );

	local numberOfRows = groupCount;
	if (numberOfRows > LeaderBotScrollSlots) then
		numberOfRows = LeaderBotScrollSlots;
	end

	--make the main frame the size of the columns + border padding on both sides, height of the columns +padding on top and bottom
	local requiredWidth = columnWidth 
						+ columnWidthTank
						+ columnWidthTank
						+ (LeaderBotborderPadding * 12); --first col, assignment cols, padding on both sides

	local requiredHeight = LeaderBotheaderHeight 
						+ (numberOfRows * LeaderBotplayerHeight) 
						+ LeaderBotfooterHeight 
						+ (LeaderBotborderPadding * 2); --std column height + app header + buffer on top/bottom
	
	TankLeaderFrame:SetSize(math.max(requiredWidth, LeaderBotMinWidth), requiredHeight);

end


--******************************************************************************************************
-- Command line parser
function TankLeader_SlashHandler(arg1)
	TLmsg("Function: TankLeader_SlashHandler()");

	local starget;

	if (arg1 == "") then
		ChatFrame1:AddMessage(colorCode .. "TankLeader by Quel @ Black Dog Games is loaded");
		ChatFrame1:AddMessage(colorCode .. "Remember kids, 'Quel' rhymes with 'fuel' ");
		return;
	end

	if (nil ~= string.find (arg1, "config") )then
		InterfaceOptionsFrame_OpenToCategory("LeaderBot");
		return;
	end

	if (nil ~= string.find(arg1, "debug")) then
		if (TL_Debug == 1) then
			ChatFrame1:AddMessage(colorCode .. "TankLeader debug messages disabled");
			HL_Debug = 0;
		else
			ChatFrame1:AddMessage(colorCode .. "TankLeader debug messages enabled");
			TL_Debug = 1;
		end
	end

end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function TankLeader_TankTargetDropDownOnClick(dropDownButton, tankName, tankTargetIcon, checked)
	TLmsg("Function: TankLeader_TankTargetDropDownOnClick(" .. tankName .. ")" );

	--mark this as the selected item in the parent dropDown list. DO NOT CALL THIS FOR MULTISELECT
	--UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	TankLeader_Assignments[TANKS_MULTISELECT][tankName][tankTargetIcon] = UIDropDownMenuButton_GetChecked(dropDownButton);

	local labelText = LeaderBot_UpdateMultiselectDropDownLabel(TankLeader_Assignments, TANKS_MULTISELECT, tankName);
	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. labelText);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "tank", TANKS_MULTISELECT, tankName, tankTargetIcon, UIDropDownMenuButton_GetChecked(dropDownButton));
	
end

-- ******************************************************************************
function TankLeader_AddColumnTankRotation(teamMembers, tankCount, forceDisable)
	TLmsg("Function: TankLeader_AddColumnTankRotation");

	local columnFrame = _G["TankLeaderColumn3"];
	if (nil == columnFrame) then
		columnFrame = CreateFrame("Frame", "TankLeaderColumn3", TankLeaderFrame, BackdropTemplateMixin and "BackdropTemplate");
	end

	columnFrame:SetBackdrop(columnBackdrop);
	columnFrame:SetBackdropColor(0.1, 0.1, 0.1, 1);
	columnFrame:SetPoint("TopLeft", columnFrames[2], "TopLeft", columnWidthTank + (2 * LeaderBotborderPadding), 0);
	LeaderBotFrame_AdjustColumnSize(columnFrame, tankCount, columnWidthTank)

	local colTitle = _G["TankLeaderColumn3Title"];
	if (nil == colTitle) then
		colTitle = columnFrame:CreateFontString("TankLeaderColumn3Title", "ARTWORK", LeaderBotColumnTitleFont);
	end

	colTitle:SetPoint("Top", columnFrame, "Top", 0, 10);
	colTitle:SetText("Tank Rotation"); 

	for i = 1, tankCount, 1 do
		local rowFrame = _G["LeaderBottankMember".. i ];
		local name = rowFrame.playerName;

		-- make the drop down menu for tank targets
		local dd = _G[tankRotationDropDownName .. i];
		if (nil == dd) then
			dd = CreateFrame("Frame",tankRotationDropDownName..i, columnFrame, "UIDropDownMenuTemplate");
		end
		
		UIDropDownMenu_SetWidth(dd,columnWidthTank - 35);

		UIDropDownMenu_EnableDropDown(dd);
		if (LeaderBotIsAssistant == false) then 
			UIDropDownMenu_DisableDropDown(dd);
		end 

		local topBuffer = 2;
		if (addHealerTargetIndex > 1) then
			topBuffer = 0;
		end

		dd:SetPoint("TopLeft", columnFrame, "TopLeft", -5,  (i -1) * -LeaderBotplayerHeight -topBuffer );
		dd.initialize = function() TankLeader_TankRotationDropDownInit(name, dd) end;
		dd.tankName = name;
		dd.tankRotationNumber = LeaderBotNone;

		--did the tank previously have an assignment? If so, re-apply it.
		local targetName = TankLeader_Assignments[TANKROT][name];
		if (targetName == LeaderBotNone) then 
			--they were previously assigned to NONE

		elseif (targetName) then
			--they previously had an assignment; re-apply it.

		else 
			--they have not had an assignment before, so assign them to RESERVE
			TankLeader_Assignments[TANKROT][name] = LeaderBotNone;
			targetName = LeaderBotNone;		
		end 

		--reused element may have been hidden on a prior pass
		dd:Show()

		--Wow's API fails me again. ><. We can't update all these dropdowns in realtime. The function to Initialize the
		--dd takes way too long and causes WoW to freeze up when called repeatedly. So we make a label, that's fast and 
		--easy to update, and put it front of the drop down.

		--first create a black box to cover up the (non-updated) dropDown text
		local ddLabelBG = _G[tankRotationDropDownName .. i .. "labelBG"];
		if (ddLabelBG == nil) then 
			ddLabelBG = CreateFrame("Frame", tankRotationDropDownName ..i .. "labelBG", dd, BackdropTemplateMixin and "BackdropTemplate");
		end
		dd.ddLabelBG = ddLabelBG;
		ddLabelBG:SetBackdrop(labelBackdrop);
		ddLabelBG:SetBackdropColor(0.075, 0.075, 0.075, 1);
		ddLabelBG:SetPoint("Center", dd, "Center", -7, 2);
		ddLabelBG:SetSize(dd:GetWidth() - 60, dd:GetHeight() -16 );

		--then overlay a fontstring to show the user what value they've selected. 
		local ddMyLabel = _G[tankRotationDropDownName .. i .. "label"];
		if (ddMyLabel == nil) then 
			ddMyLabel = ddLabelBG:CreateFontString(tankRotationDropDownName .. i .. "label", "ARTWORK", "GameFontNormalSmall");
		end 
		dd.ddMyLabel = ddMyLabel;
		ddMyLabel:SetPoint("Center", dd, "Center", -8, 0);
		ddMyLabel:SetText(whiteColorCode .. targetName);
		local fontHeightPerRow = ddMyLabel:GetStringHeight();	
		ddMyLabel:SetSize(ddLabelBG:GetWidth() -5, fontHeightPerRow );

		UIDropDownMenu_SetSelectedValue(dd, targetName);
	end

	--hide any leftover elements if locks have left the raid
	local i = tankCount + 1;
	local dd = _G[tankRotationDropDownName .. i];
	while ( dd ~= nil) do
		dd:Hide();

		i = i  + 1;
		dd = _G[tankRotationDropDownName .. i];
	end

	local button = _G["TankLeaderAnnounce3"];
	if (nil == button) then
		button = CreateFrame("Button", "TankLeaderAnnounce3", columnFrame, "OptionsButtonTemplate");
	end

	button:SetPoint("Top",  columnFrame, "Bottom", 0, LeaderBotInfoPadding);
	button:SetText("Announce");
	button.tooltipText = "Announce tanking assignments to selected chat channels";
	button:SetSize(75,30);
	if (forceDisable == true) then
		button:Disable();
	end

	button:SetScript("OnClick", function(self)
											LLmsg("Hander: TankLeaderAnnounce3.OnClick()");
											TankLeader_AnnounceToChat("Leader Bot: Tanking Rotation", TANKROT);
								end);
	if (tankCount <= 0) then
		button:Hide();
	else
		button:Show();
	end 

	return columnFrame;
end

-- ******************************************************************************
function TankLeader_TankRotationDropDownInit(tankName, dropDownMenu)
	TLmsg("Function: TankLeader_TankRotationDropDownInit(" .. tankName .. ")" );
	local rowInfo = UIDropDownMenu_CreateInfo();

	rowInfo.text = LeaderBotNone ;
	rowInfo.value = LeaderBotNone;
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	rowInfo.icon = nil;
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = LeaderBotNone;
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = "1st";
	rowInfo.value =  "1st";
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = 1;											--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = "2nd";
	rowInfo.value =  "2nd";
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = 2;											--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = "3rd";
	rowInfo.value =  "3rd";
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = 3;											--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to
	
	rowInfo.text = "4th";
	rowInfo.value =  "4th";
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = 4;											--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to

	rowInfo.text = "5th";
	rowInfo.value =  "5th";
	rowInfo.isNotRadio = true;
	rowInfo.owner  = dropDownMenu;
	--rowInfo.icon = "Interface\\TargetingFrame\\UI-RaidTargetingIcon_1";
	rowInfo.checked = false;
	rowInfo.arg1 = tankName;									-- the click handler needs to know which checkbox this dropdown pertains to, start,early,late or end.
	rowInfo.arg2 = 5;											--this is used at SendChatMessage() time to force the target icon into chat
	rowInfo.func = TankLeader_TankRotationDropDownOnClick;		--set the click handler for this SPECIFIC BUTTON in this drop down.
	UIDropDownMenu_AddButton(rowInfo);							--add the button. UI knows which dropDown you've clicked on, so it knows who to add this to
	
	
	end

--******************************************************************************************************
-- DropDown click handler. Note: this is a pain in the ass because you don't get a pointer to the dropDown, you get a pointer to the 
--specific button in the list.
function TankLeader_TankRotationDropDownOnClick(dropDownButton, arg1, arg2, checked)
	TLmsg("Function: TankLeader_TankRotationDropDownOnClick(" .. arg1 .. ")" );
	
	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	--update the assignment table for this warlock
	local tankName = arg1;
	local tankRotationNumber = dropDownButton.value;
	TankLeader_Assignments[TANKROT][tankName] = tankRotationNumber;

	dropDownButton.owner.ddMyLabel:SetText(whiteColorCode .. dropDownButton.value);

	LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "tank", TANKROT, tankName, dropDownButton.value);
end

--******************************************************************************************************
-- EventHandler
function TankLeaderFrame_OnEvent(self, event, ...)
	local arg1 = select(1, ...);
	if (arg1 == nil) then
		arg1 = "<n/a>";
	end
	TLmsg("Function: TankLeaderFrame_OnEvent(Event: " .. event .. ", " .. arg1 .. ")" );

	-- *****************************************************************************************
	if (event == "VARIABLES_LOADED") then
			TLmsg ("     VARIABLES LOADED HANDLER");

			--used to prevent the OnShow() handler from trying to render a list before we have names
			--loaded.
			variablesLoaded = true;


	-- *****************************************************************************************
	-- lookup details of the event. The lookup method varies based on the even type
	elseif (event == "COMBAT_LOG_EVENT_UNFILTERED") then
		local timestamp, event, hideCaster, srcGUID, srcName, srcFlags, srcFlags2, dstGUID, dstName, dstFlags, spellid, spellname, spellschool, b_or_d = select(1, ...);
		TLmsg("event = " .. timestamp);


	-- ******************************************************************************************
	elseif (ttype == "SPELL_AURA_REMOVED" and dstGUID == UnitGUID("player") ) then
		--NightLightFrame_RemoveTimer(spellid)
	end
end

--[[ 
--********************************************************************
-- arg1 = seconds since last OnUpdate()
function TankLeaderFrame_OnUpdate(self, elapsed)
	--ChatFrame1:AddMessage("     ****LockleaderFrameMax_OnUpdate(" .. elapsed .. ")****");
	local i;
	
	HL_loopcontrol = HL_loopcontrol + elapsed;
	if (HL_loopcontrol < 1.0) then
		return;
	end
	HL_loopcontrol = 0;

end
--]]

-- ************************************************
function TankLeader_DataMigration()
	TLmsg("Function: TankLeader_DataMigration()" );

	if (LeaderBotOptions["TankLeaderDataModelVersion"] ~= LeaderBotCurrentDataModel) then
		TankLeader_Assignments = {};
		TankLeader_Assignments[TANKS_MULTISELECT] = {};
		TankLeader_Assignments[TANKROT] = {};

		LeaderBotOptions["TankLeaderDataModelVersion"] = LeaderBotCurrentDataModel;
	end

	--if the _assignments array was saved and then a column is added, it doesn't matter if we
	--initialize the new column or not. After VARIABLES_LOADED, the _assignments array will be reset
	--to the saved version, which is missing the column and results in various nil errors. So,
	--upon variables loaded, walk all the assignments and make sure they're actually initialized.
	for assignment = 1, TANKASSIGNMENTCOUNT do
		if (TankLeader_Assignments[assignment] == nil) then
			TankLeader_Assignments[assignment] = {};
		else
			--replace all instances of legacy "-none-" with "-None-"
			for key, value in next, TankLeader_Assignments[assignment] do
				if (value == "-none-" or value == " -none-" or value == " -None-") then 	
					TankLeader_Assignments[assignment][key] = LeaderBotNone;
				end 
			end 
		end
	end 
end

-- ************************************************
function TankLeader_RosterChanged(newTeamMembers, newTankCount)
	TLmsg("Function: TankLeader_RosterChanged()");

--[[
	--TODO: At first, I thought this was a good idea to wipe out assignments when a player leaves the group. However,
	--I found that upon logging in, events fire 4-6 times that call _RosterChanged(). and sometimes those events fire
	--before the group is "formed" in Blizzard's mind. So, I'd get here, see group members "were gone" and remove them. 
	--then 3 events later find them again and add them to the table with blank assignments. So I'm taking this chunk out
	--for now in an attempt to preserve assignments after re-logging.

	--walk each assignmentType TANKS, TANKROT, etc
	for assignmentType in next, TankLeader_Assignments do

		--walk each tank in that assignment type
		for tankNameKey, targetValue in next, TankLeader_Assignments[assignmentType] do

			--see if this tank is still in the raid as you can only query other team members
			--directly by name
			local playerClass = UnitClass(tankNameKey);

			if (playerClass == nil) then
				--this tank has left the building
				TankLeader_Assignments[assignmentType][tankNameKey] = nil;
			end 

		end -- for tankNameKey 

	end --for assignmentType
--]]

	TankLeader_AdjustMainPanelSize(newTankCount);
	columnFrames[1] = LeaderBotFrame_AddColumn1(newTeamMembers, newTankCount, "tank", TankLeaderFrame, "Tanks");
	columnFrames[2] , TankLeader_Assignments = LeaderBotFrame_AddColumnRaidIcons(newTeamMembers, newTankCount, true, 
							TankLeaderFrame, columnFrames[1], 
							columnWidthTank, "TankLeaderColumn2", "Tank Target", 
							"LeaderBottankMember", tankTargetDropDownName, 
							TankLeader_Assignments, TANKS_MULTISELECT, 
							"TankLeaderAnnounce2", TankLeader_AnnounceToChat, "Leader Bot: Tanking Targets",
							TankLeader_TankTargetDropDownOnClick, 
							LeaderBotMultiSelect);
	
	columnFrames[3] = TankLeader_AddColumnTankRotation(newTeamMembers, newTankCount, false);

	if (newTankCount > LeaderBotScrollSlots) then
		LeaderBot_RegisterColumnForScrolling(3, "tank", 1, columnFrames[1], "LeaderBottankMember");
		LeaderBot_RegisterColumnForScrolling(3, "tank", 2, columnFrames[2], tankTargetDropDownName);
		LeaderBot_RegisterColumnForScrolling(3, "tank", 3, columnFrames[3], tankRotationDropDownName);

		LeaderBot_SetupScrollBar("tank");
	end 
	
	local dd = _G["TankLeaderChatChannelDropDown"];
	if (LeaderBotIsAssistant == false) then
		TankLeader_AnnounceButtonEnable(0);
		UIDropDownMenu_DisableDropDown(dd);
	else
		UIDropDownMenu_EnableDropDown(dd);
	end 

end

-- ******************************************************************************
function TankLeader_AnnounceButtonEnable(state)
	TLmsg("Function: TankLeader_AnnounceButtonEnable(" .. state .. ")");

	--button names start with 2 because they begin in column2
	local i = 2;
	local button = _G["TankLeaderAnnounce" .. i];

	while (button) do 
		if (state == 0) then
			button:Disable();
		elseif (LeaderBotIsAssistant) then
			button:Enable();
			button:Show();
		end

		i = i + 1;
		button = _G["TankLeaderAnnounce" .. i];
	end 

end

-- ******************************************************************************
function TankLeader_AnnounceToChat(title, assignmentIndex)
	TLmsg("Function: TankLeader_AnnounceToChat(" .. title .. ", type = " .. assignmentIndex);

	--this array is populated by the drop-down OnClick handler.
	local chanType = TankLeader_MsgChannel[LOCKCHANTYPE];
	local chanLang = TankLeader_MsgChannel[LOCKCHANLANG];
	local chanNumber = TankLeader_MsgChannel[LOCKCHANNUMB];

	if (chanType == nil) then
		ChatFrame1:AddMessage("LeaderBot - Healer Leader: No chat channel selected");
		return;
	end

	if (chanNumber == nil) then
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType);
		SendChatMessage(title, chanType);
	else
		SendChatMessage("++++++++++++++++++++++++++++++++++", chanType, chanLang, chanNumber);
		SendChatMessage(title, chanType, chanLang, chanNumber)
	end

	if (assignmentIndex == TANKS_MULTISELECT) then 
		for playerName, playerArray in next, TankLeader_Assignments[assignmentIndex] do

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(playerName);
			if (isInGroup) then 

				local msg = "";
				local comma = "";

				for assignmentName, assignmentValue in next, playerArray do 

					if (assignmentName ~= LeaderBotNone and assignmentValue == true) then 
						msg = assignmentName .. comma .. msg;
						comma = ", ";
					end

				end 

				if (msg ~= "") then 
					msg = playerName .. ": " .. msg;
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end

					--if the raid leader isn't using leaderbot and a team member joins after the class leader made assignments, then the new team
					--member doesn't receive a sync of the previously made changes because the raid leader never responds. This is a work around to 
					--sync everything being announced.
					if (string.find(assignmentIndex, "MultiSelect") ) then 
						for key_TargetName, val_CheckMarkStatus in next, TankLeader_Assignments[assignmentIndex][playerName] do
							LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "tank", assignmentIndex, playerName, key_TargetName, val_CheckMarkStatus);
						end
					else 
						LeaderBotSyncMaster_SendMsg(LeaderBotSenderName, nil, "tank", assignmentIndex, playerName, TankLeader_Assignments[assignmentIndex][playerName]);
					end 

				end 

			end

		end 

	else 
		for assignmentKey, assignmentValue in next, TankLeader_Assignments[assignmentIndex] do
			local name, target;
			local msg = nil;

			name = assignmentKey;
			targetName = assignmentValue;

			--leader assignments save assignments for previous players, so only announce assignments for folks in the raid
			local isInGroup = UnitClass(name);
			if (isInGroup) then 

				--skip players assigned to LeaderBotNone. These are probably tank-able classes playing DPS
				--or heals
				if (targetName ~= LeaderBotNone) then 
					msg = "     " .. name .. ": " .. targetName;

					--if (chanType == "RAID") then
					if (chanNumber == nil) then
						--Party, raid
						SendChatMessage(msg, chanType);
					else
						SendChatMessage(msg, chanType, chanLang, chanNumber);
					end

				end

			end
		end
	end 

	if (chanNumber == nil) then
		--Party, raid
		SendChatMessage("-------------------------------------------", chanType);
	else
		SendChatMessage("--------------------------------------------", chanType, chanLang, chanNumber);
	end
								
end

-- ******************************************************************************
function TankLeader_ChannelDropDownOnClick(dropDownButton, arg1, arg2, checked)
	TLmsg("TankLeader_ChannelDropDownOnClick()" );

	--mark this as the selected item in the parent dropDown list.
	UIDropDownMenu_SetSelectedValue(dropDownButton.owner, dropDownButton.value);

	TankLeader_MsgChannel[LOCKCHANTYPE] = arg1;
	TankLeader_MsgChannel[LOCKCHANLANG] = nil;
	TankLeader_MsgChannel[LOCKCHANNUMB] = arg2;

	TankLeader_AnnounceButtonEnable(1);
	
end



